#include "SmageDataProcessor.h"
#include "SmageSettings.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <iostream>

#include <zlib.h>

#define SMAGE_CHECK_ERR(err, msg) { if (err != Z_OK) {_SmageLogFile.lprintf("%s error: %d\n", msg, err); exit(1); }}

using namespace std;

// Structure's constructor
SmageData::SmageData()
{
  byteStream.clear();
  subID = 0;
}

SmagePropsStruct::SmagePropsStruct()
{
  props = 0;
  link.clear();
}

// Constructor
SmageDataProcessor::SmageDataProcessor()
{
  _currentDataSize = 0;
  _propsStreamSize = 0;
  _compressedDataSize = 0;
  _decompressedDataSize = 0;
  _dataStream = NULL;
  _propsStream = NULL;
  _compressedDataStream = NULL;
  _decompressedDataStream = NULL;
}

// Destructor
SmageDataProcessor::~SmageDataProcessor()
{
  if (_currentDataSize > 0)
    delete _dataStream;

  if (_propsStreamSize > 0)
    delete _propsStream;

  if (_compressedDataSize > 0)
    delete _compressedDataStream;

  if (_decompressedDataSize > 0)
    delete _decompressedDataStream;
}

// Set class data to NULL
void SmageDataProcessor::resetDataStream()
{
  // Freeing allocated memory and setting size of data streams
  // to NULL

  if (_currentDataSize > 0)
    delete _dataStream;

  if (_compressedDataSize > 0)
    delete _compressedDataStream;

  if (_decompressedDataSize > 0)
    delete _decompressedDataStream;

  _currentDataSize = 0; _compressedDataSize = 0; _decompressedDataSize = 0;
  _dataStream = NULL; _compressedDataStream = NULL; _decompressedDataStream = NULL;
}

// Set stream holding the properties to NULL
void SmageDataProcessor::resetPropsStream()
{
  // Freeing allocated memory and setting size of data streams
  // to NULL

  if (_propsStreamSize > 0)
    delete _propsStream;

  _propsStreamSize = 0;
  _propsStream = NULL;
}

// Return the original data stream
const char *SmageDataProcessor::getDataStream(unsigned long &dataStreamSize)
{
  dataStreamSize = _currentDataSize;
  return _dataStream;
}

// Return the compressed data stream    
const char *SmageDataProcessor::getCompressedDataStream(unsigned long &comprDataStreamSize)
{
  comprDataStreamSize = _compressedDataSize;
  return _compressedDataStream;
}

// Return the decompressed data stream
const char *SmageDataProcessor::getDecompressedDataStream(unsigned long &decomprDataStreamSize)
{
  decomprDataStreamSize = _decompressedDataSize;
  return _decompressedDataStream;
}

void SmageDataProcessor::convertToChars(const unsigned long &src, char dest[4])
{
  // Converting a long ("src", 4 bytes) to an array of characters ("dest");
  // "dest"[0] = 8 lowest bits of "src"
  // "dest"[3] = 8 highest bits of "src"

  unsigned long srcCopy = src;

  dest[0] = 0; dest[0] = (srcCopy & 255); srcCopy = (srcCopy >> 8);
  dest[1] = 0; dest[1] = (srcCopy & 255); srcCopy = (srcCopy >> 8);
  dest[2] = 0; dest[2] = (srcCopy & 255); srcCopy = (srcCopy >> 8);
  dest[3] = 0; dest[3] = (srcCopy & 255);
}

void SmageDataProcessor::convertToLong(const char src[4], unsigned long &dest)
{
  // Converting an array of characters ("src") to a long "dest"
  // 8 lowest bits of "dest" = "src"[0]
  // 8 highest bits of "dest" = "src"[3]

  dest = 0;

  dest |= ((src[3] & 255) << 24);
  dest |= ((src[2] & 255) << 16);
  dest |= ((src[1] & 255) << 8);
  dest |= (src[0] & 255);
}

void SmageDataProcessor::convertEnumToChar(const DataType &dataTypeSrc, char &dataTypeDest)
{
  dataTypeDest = 0;

  switch (dataTypeSrc)
  {
    case SMAGE_TEXT: dataTypeDest = 0; break;
    case SMAGE_FORM_TEXT: dataTypeDest = 1; break;
    case SMAGE_IMG: dataTypeDest = 2; break;
    case SMAGE_COLOR_TABLE: dataTypeDest = 3; break;
    case SMAGE_FILE: dataTypeDest = 4; break;
    case SMAGE_GAME: dataTypeDest = 5; break;
    case SMAGE_DICT: dataTypeDest = 6; break;
    default: 
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::convertEnumToChar): unknown code type ***\n");
#endif
      break;
    }
  }
}

void SmageDataProcessor::convertCharToEnum(const char &dataTypeSrc, DataType &dataTypeDest)
{
  switch (dataTypeSrc)
  {
    case 0: dataTypeDest = SMAGE_TEXT; break;
    case 1: dataTypeDest = SMAGE_FORM_TEXT; break;
    case 2: dataTypeDest = SMAGE_IMG; break;
    case 3: dataTypeDest = SMAGE_COLOR_TABLE; break;
    case 4: dataTypeDest = SMAGE_FILE; break;
    case 5: dataTypeDest = SMAGE_GAME; break;
    case 6: dataTypeDest = SMAGE_DICT; break;
    default: 
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::convertCharToEnum): unknown code type ***\n");
#endif
      break;
    }
  }
}

void SmageDataProcessor::convertEnumToInt(const CodePattern &codePatternSrc, int &codePatternDest)
{
  codePatternDest = 0;

  switch (codePatternSrc)
  {
    case SMAGE_AUTO: codePatternDest = 0; break;
    case SMAGE_LSB: codePatternDest = 1; break;
    case SMAGE_BITS_3: codePatternDest = 2; break;
    case SMAGE_BITS_4: codePatternDest = 3; break;
    case SMAGE_LS_2B: codePatternDest = 4; break;
    case SMAGE_BITS_6: codePatternDest = 5; break;
    case SMAGE_BITS_7: codePatternDest = 6; break;
    case SMAGE_LS_3B: codePatternDest = 7; break;
    case SMAGE_BITS_9: codePatternDest = 8; break;
    case SMAGE_BITS_10: codePatternDest = 9; break;
    case SMAGE_LS_4B: codePatternDest = 10; break;
    case SMAGE_BITS_12: codePatternDest = 11; break;
    case SMAGE_BITS_13: codePatternDest = 12; break;
    case SMAGE_LS_5B: codePatternDest = 13; break;
    case SMAGE_LS_8B: codePatternDest = 14; break;
    case SMAGE_MAX: codePatternDest = 15; break;
    default: 
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::convertEnumToInt): unknown code pattern ***\n");
#endif
      break;
    }
  }
}

void SmageDataProcessor::convertIntToEnum(const int &codePatternSrc, CodePattern &codePatternDest)
{
  switch (codePatternSrc)
  {
    case 0: codePatternDest = SMAGE_AUTO; break;
    case 1: codePatternDest = SMAGE_LSB; break;
    case 2: codePatternDest = SMAGE_BITS_3; break;
    case 3: codePatternDest = SMAGE_BITS_4; break;
    case 4: codePatternDest = SMAGE_LS_2B; break;
    case 5: codePatternDest = SMAGE_BITS_6; break;
    case 6: codePatternDest = SMAGE_BITS_7; break;
    case 7: codePatternDest = SMAGE_LS_3B; break;
    case 8: codePatternDest = SMAGE_BITS_9; break;
    case 9: codePatternDest = SMAGE_BITS_10; break;
    case 10: codePatternDest = SMAGE_LS_4B; break;
    case 11: codePatternDest = SMAGE_BITS_12; break;
    case 12: codePatternDest = SMAGE_BITS_13; break;
    case 13: codePatternDest = SMAGE_LS_5B; break;
    case 14: codePatternDest = SMAGE_LS_8B; break;
    case 15: codePatternDest = SMAGE_MAX; break;
    default:
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::convertIntToEnum): unknown code pattern ***\n");
#endif
      break;
    }
  }
}

// Encode data type and sub-ID to data stream
void SmageDataProcessor::encodeType_ID(const DataType &dataType, const char &subID, char &dataType_ID)
{
  // Code "dataType" in first 4 bits of "dataType_ID" and
  // "subID" in last 4 bits of "dataType_ID";

  dataType_ID = 0;
  char type;
  convertEnumToChar(dataType, type);

  dataType_ID |= ((subID & 15) << 4);
  dataType_ID |= (type & 15);
}

// Decode data type and sub-ID from data stream
void SmageDataProcessor::decodeType_ID(const char &dataType_ID, DataType &dataType, char &subID)
{
  // Decode "dataType" from first 4 bits of "dataType_ID" and
  // "subID" from last 4 bits of "dataType_ID";

  subID = 0;
  char type = 0;;

  subID |= ((dataType_ID & 240) >> 4);
  type |= (dataType_ID & 15);

  convertCharToEnum(type, dataType);
}

void SmageDataProcessor::setProps(const SmagePropsStruct &propsStruct)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::setProps(SmagePropsStruct)\n");
#endif

  // Are properties set?
  if (propsStruct.props == 0)
    return;

  char *propsData = NULL;

  unsigned long size = 0;
  unsigned int linkSize = propsStruct.link.size();
  // "+ 2" because: 2 characters for coding the link size
  size += (unsigned long)(linkSize) + 2;

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::setProps(): Link size: %d\n", linkSize);
  _SmageLogFile.lprintf("SmageDataProcessor::setProps(): Properties stream size: %lu\n", size);
#endif

  propsData = new char[size];

  if (propsData == NULL)
    return;

  char *destPointer = propsData;

  if (propsStruct.props & SMAGE_LINK)
  {
    *destPointer = char(linkSize & (255 << 8));
    destPointer++;
    *destPointer = char(linkSize & 255);
    destPointer++;

    for (unsigned int i=0; i<propsStruct.link.size(); i++)
    {
      *destPointer = propsStruct.link[i];
      destPointer++;
    }
  }

  setProps(propsStruct.props, propsData, size);
}

// Set properties
bool SmageDataProcessor::setProps(const char &props, const char *propsData, const unsigned long &size)
{
  // The first character of "_propsStream" will be set to "props";
  // After it, copying "propsData" into "_propsStream" and storing its size ("size") + 1 ("props") in "_propsStreamSize";
  // data is reset before so that possibly stored data gets lost

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::setProps(...)\n");
#endif

  resetPropsStream();

  _propsStream = new char[size+1];
  if (_propsStream == NULL)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::setProps): out of memory ***\n");
#endif
    return 0;
  }

  const char *srcPos = propsData;
  char *destPos = _propsStream;

  // Setting the properties ID
  *destPos = props;
  destPos++;

  unsigned long propsStreamSize = size;

  while (propsStreamSize > 0)
  {
    *destPos = *srcPos;
    srcPos++; destPos++;
    propsStreamSize--;
  }
  
  _propsStreamSize = size + 1;
  
  return 1;
}

// Copying "_propsStreamSize" and "_propsStream" to "_dataStream";
//  "_dataStream" must have been initialized
void SmageDataProcessor::copyPropsToDataStream()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::copyPropsToDataStream()\n");
#endif

  // the resulting "_dataStream" will look like:
  // [ _dataStream, _propsStreamSize" (4 characters), "_propsStream"]

  if (_dataStream != NULL && _currentDataSize > 0)
  {
    unsigned long propsStreamSize = 1;
    if (_propsStreamSize > 0)
      propsStreamSize = _propsStreamSize;

#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("SmageDataProcessor::copyPropsToDataStream(): Properties size: %lu\n", propsStreamSize);
#endif

    unsigned long tmpDataSize = _currentDataSize + (4 + propsStreamSize);

#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("SmageDataProcessor::copyPropsToDataStream(): Resulting Stream size: %lu\n", tmpDataSize);
#endif

    // Copy "_dataStream" into a temporary byteStream "tmpDataStream" and
    // allocate new memory for "_dataStream" according to "tmpDataSize" which 
    // will be the resulting size of "_dataStream"

    char *tmpDataStream = (char *)calloc(_currentDataSize, 1);
    memcpy(tmpDataStream, _dataStream, (unsigned int)_currentDataSize);

    if (tmpDataStream != NULL)
    {
      delete _dataStream;
      _dataStream = new char[tmpDataSize];

      if (_dataStream == NULL)
      {
#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::copyPropsToDataStream): out of memory ***\n");
#endif
        return;
      }
      else
      {
        char *dataStreamPointer = NULL;
        char *curSrcPos = NULL;

        // Copying old "_dataStream" into actual "_dataStream"
        unsigned long oldCurrentDataSize = _currentDataSize;
        _currentDataSize = tmpDataSize;

        dataStreamPointer = _dataStream;
        curSrcPos = tmpDataStream;

        while (oldCurrentDataSize > 0)
        {
          *dataStreamPointer = *curSrcPos;
          curSrcPos++; dataStreamPointer++;
          oldCurrentDataSize--;
        }

        // Appending the properties
        
        // Assigning "_propsStreamSize" to "_dataStream"
        int counter = 4;
        char sizeChar[4];
        convertToChars(propsStreamSize, sizeChar);

        while (counter > 0)
        {
          *dataStreamPointer = sizeChar[counter-1];
          dataStreamPointer++; counter--;
        }
      
        // Appending "_propsStream" to "_dataStream"
        if (propsStreamSize == 1)
        {
#ifdef SMAGE_PRINT_DEBUG
          _SmageLogFile.lprintf("SmageDataProcessor::copyPropsToDataStream(): No properties specified.\n");
#endif
          *dataStreamPointer = 0;
          dataStreamPointer++;
        }
        else
        {      
          const char *curDataPos = _propsStream;

#ifdef SMAGE_PRINT_DEBUG
          _SmageLogFile.lprintf("SmageDataProcessor::copyPropsToDataStream(): Properties: ");
#endif

          while (propsStreamSize > 0)
          {
            *dataStreamPointer = *curDataPos;

#ifdef SMAGE_PRINT_DEBUG
            _SmageLogFile.lprintf("%c", *dataStreamPointer);
#endif

            curDataPos++; dataStreamPointer++;
            propsStreamSize--;
          }

#ifdef SMAGE_PRINT_DEBUG
          _SmageLogFile.lprintf("\n");
#endif
        }
      }
    }
  }
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::copyPropsToDataStream): Data stream not initialized ***\n");
#endif
  }
}

void SmageDataProcessor::decodeProps(unsigned long &curDataSize, const char **curDataPos, SmagePropsStruct &propsStruct)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::decodeProps()\n");
#endif

  propsStruct.props = **curDataPos;
  (*curDataPos)++;
  curDataSize--;

  if (propsStruct.props & SMAGE_LINK)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("SmageDataProcessor::decodeProps(): A link has been coded.\n");
#endif

    if (curDataSize-2 > 0)
    {
      // A link location has been coded; its size is coded in a short int (16 bit);
      // Bits 9-16 are coded in the first character, Bits 1-8 in the following character
      short int linkSize = 0;
      linkSize = (unsigned int)(**curDataPos);
      linkSize = linkSize << 8;
      (*curDataPos)++;
      linkSize |= (unsigned int)(**curDataPos);
      (*curDataPos)++;
      curDataSize -= 2;

#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("SmageDataProcessor::decodeProps(): Link size: %d\n", int(linkSize));
#endif

#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("SmageDataProcessor::decodeProps(): The link: ");
#endif

      while (linkSize > 0)
      {
        propsStruct.link.push_back(**curDataPos);

#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("%c", **curDataPos);
#endif

        (*curDataPos)++;
        linkSize--; curDataSize--;
      }

#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("\n");
#endif
    }
  }
}

// Set data (character field)
bool SmageDataProcessor::setData(const char *dataStream, const unsigned long &size)
{
  // Copying "dataStream" into "_dataStream" and storing its size in "_currentDataSize";
  // data is reset before so that possibly stored data gets lost

  resetDataStream();

  _dataStream = new char[size];
  if (_dataStream == NULL)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::setData): out of memory ***\n");
#endif
    return 0;
  }

  _currentDataSize = size;

  const char *srcPos = dataStream;
  char *destPos = _dataStream;
  
  while (_currentDataSize > 0)
  {
    *destPos = *srcPos;
    srcPos++; destPos++;
    _currentDataSize--;
  }

  _currentDataSize = size;

  return 1;
}

// Set data (character vector)
bool SmageDataProcessor::setData(const std::vector<char> &dataStream)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::setData()\n");
#endif
  resetDataStream();

  _dataStream = new char[dataStream.size()];
  if (_dataStream == NULL)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::setData): out of memory ***\n");
#endif
    return 0;
  }

  _currentDataSize = dataStream.size();

  vector<char>::const_iterator srcPos;
  char *destPos = _dataStream;
  
  for (srcPos=dataStream.begin(); srcPos!=dataStream.end(); srcPos++)
  {
    *destPos = *srcPos;
    destPos++;
  }

  return 1;
}

// Add data (character vector)
bool SmageDataProcessor::addData(const std::vector<char> &dataStream, const DataType &dataType, 
         const char &subID)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::addData()\n");
#endif

  char *data = new char[dataStream.size()];
  if (data == NULL)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::addData): out of memory ***\n");
#endif
    return 0;
  }

  vector<char>::const_iterator dataIter;
  char *destPointer = data;

  for (dataIter=dataStream.begin(); dataIter!=dataStream.end(); dataIter++)
  {
    *destPointer = *dataIter;
    destPointer++;
  }

  bool result = addData(data, dataStream.size(), dataType, subID);

  delete[] data;

  return result;
}

// Add data (character field)
bool SmageDataProcessor::addData(const char *dataStream, const unsigned long &size,
         const DataType &dataType, const char &subID)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::addData()\n");
#endif

  // Adding "dataStream" to "_dataStream";
  // if size of "_dataStream" is greater than 0, "_dataStream" is first copied
  // before the new "dataStream" is appended;
  // "_currentDataSize" holds the actual size of "_dataStream";
  // the resulting "_dataStream" will look like:
  // [ old _dataStream, "size" (4 characters), "dataType"+"subID" (1 character), "dataStream" ("size" characters) ]

  unsigned long tmpDataSize = _currentDataSize + (4 + 1 + size);
  char *curSrcPos = NULL;
  char *curDestPos = NULL;

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::addData(): Resulting stream size: %lu\n", tmpDataSize);
#endif

  if (_currentDataSize > 0)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("SmageDataProcessor::addData(): Data stream is added to already existing data stream.\n");
#endif
    // Copy "_dataStream" into a temporary byteStream "tmpDataStream" and
    // allocate new memory for "_dataStream" according to "tmpDataSize" which 
    // will be the resulting size of "_dataStream"

    char *tmpDataStream = (char *)calloc(_currentDataSize, 1);
    memcpy(tmpDataStream, _dataStream, (unsigned int)_currentDataSize);

    if (tmpDataStream != NULL)
    {
      delete _dataStream;
      _dataStream = new char[tmpDataSize];

      if (_dataStream == NULL)
      {
#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::addData): out of memory ***\n");
#endif
        return 0;
      }
      else
      {
        unsigned long oldCurrentDataSize = _currentDataSize;
        _currentDataSize = tmpDataSize;

        curDestPos = _dataStream;
        curSrcPos = tmpDataStream;

        while (oldCurrentDataSize > 0)
        {
          *curDestPos = *curSrcPos;
          curSrcPos++; curDestPos++;
          oldCurrentDataSize--;
        }
      }
    }
  }
  else
  {
    // "_dataStream" is NULL and memory has to be allocated initially

    _dataStream = new char[tmpDataSize];
    if (_dataStream == NULL)
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::addData): out of memory ***\n");
#endif
      return 0;
    }

    _currentDataSize = tmpDataSize;
    curDestPos = _dataStream;
  }

  // Converting "size" which is of type long into an array of 4 characters
  // and appending it to "_dataStream"

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::addData(): Size of data stream: %lu\n", size);
#endif

  int counter = 4;
  char sizeChar[4];
  convertToChars(size, sizeChar);

  while (counter > 0)
  {
    *curDestPos = sizeChar[counter-1];  
    curDestPos++; counter--;
  }

  // Appending "dataType" and ID to "_dataStream"
  char dataType_ID;
  encodeType_ID(dataType, subID, dataType_ID);
  *curDestPos = dataType_ID;
  curDestPos++;

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::addData(): Data type and ID: %d\n", int(dataType_ID));
#endif

  // Appending "dataStream" to "_dataStream"
  const char *curDataPos = dataStream;
  tmpDataSize = size;

#ifdef SMAGE_PRINT_DEBUG
  if (size < 50)
    _SmageLogFile.lprintf("SmageDataProcessor::addData(): Data stream: ");
#endif

  while (tmpDataSize > 0)
  {
    *curDestPos = *curDataPos;

#ifdef SMAGE_PRINT_DEBUG
    if (size < 50)
      _SmageLogFile.lprintf("%c", *curDestPos);
#endif

    curDataPos++; curDestPos++;
    tmpDataSize--;
  }

#ifdef SMAGE_PRINT_DEBUG
  if (size < 50)
    _SmageLogFile.lprintf("\n");
#endif

  return 1;
}

// Get a compressed byte stream from a vector of data items
void SmageDataProcessor::getPreparedDataStream(const vector<SmageData> &srcData, vector<char> &destData)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::getPreparedDataStream()\n");
#endif

  destData.clear();
  resetDataStream();

  // Add data to class
  vector<SmageData>::const_iterator srcDataIter;
  for (srcDataIter=srcData.begin(); srcDataIter!=srcData.end(); srcDataIter++)
    addData((*srcDataIter).byteStream, (*srcDataIter).dataType, (*srcDataIter).subID);

  copyPropsToDataStream();

  // Compress data
  compressData();
  
  // Assign compressed data stream to "destData"
  char *srcChar = _compressedDataStream;
  for (unsigned int j=0; j<_compressedDataSize; j++)
  {
    destData.push_back(*srcChar);
    srcChar++;
  }
}

// Unsplit the data stream (already assigned to the class) and extract the single data items
void SmageDataProcessor::unsplitDataStream(vector<SmageData> &streamData, 
             SmagePropsStruct &props, const bool &decompress)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::unsplitDataStream()\n");
#endif

  // Extracting the single data stream items out of "_dataStream" and storing
  // them in a "SmageData" array;
  // "decompress" -> should data be decompressed before extracting the data items?

  if (decompress)
    if (!decompressData())
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::unsplitDataStream): could not decompress data ***\n");
#endif
      return;
    }

  unsigned long currentDataSize;
  if (decompress)
    currentDataSize = _decompressedDataSize;
  else
    currentDataSize = _currentDataSize;

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::unsplitDataStream(): Decompressed stream size: %lu\n", currentDataSize);
#endif

  if (currentDataSize != 0)
  {
    streamData.clear();
      
    unsigned long size = 0;
    char c_curDataSize[4];
    unsigned long l_curDataSize;
    const char *curDataPos;

    if (decompress)
      curDataPos = _decompressedDataStream;
    else
      curDataPos = _dataStream;

    // Assign the single data items to "SmagaData" items
    while (size < currentDataSize)
    {
      c_curDataSize[3] = *curDataPos; curDataPos++;
      c_curDataSize[2] = *curDataPos; curDataPos++;
      c_curDataSize[1] = *curDataPos; curDataPos++;
      c_curDataSize[0] = *curDataPos; curDataPos++;
      size += 4;

      convertToLong(c_curDataSize, l_curDataSize);

#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("SmageDataProcessor::unsplitDataStream(): Data size: %lu\n", l_curDataSize);
#endif
    
      if ((size + l_curDataSize) < currentDataSize)
      {
        SmageData curStreamData;

        // Data type and sub-ID
        decodeType_ID(*curDataPos, curStreamData.dataType, curStreamData.subID);

#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("SmageDataProcessor::unsplitDataStream(): Data type and ID: %d\n", int(*curDataPos));
#endif
        curDataPos++; size++;

        size += l_curDataSize;

        // The data
#ifdef SMAGE_PRINT_DEBUG
	bool printData = 0;
        if (l_curDataSize < 50)
	  {
	    printData = 1;
	    _SmageLogFile.lprintf("SmageDataProcessor::unsplitDataStream(): Data stream: ");
	  }
#endif

        while (l_curDataSize > 0)
        {
          curStreamData.byteStream.push_back(*curDataPos);

#ifdef SMAGE_PRINT_DEBUG
          if (printData)
            _SmageLogFile.lprintf("%c", *curDataPos);
#endif
          curDataPos++;
          l_curDataSize--;
        }

#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("\n");
#endif

        streamData.push_back(curStreamData);
      }
      else if ((size + l_curDataSize) == currentDataSize)
      {        
        size += l_curDataSize;
        decodeProps(l_curDataSize, &curDataPos, props);
      }
    }
  }
}

// Get the number of data items in a data stream

int SmageDataProcessor::getNumDataObjs(const char *dataStream, const unsigned long &dataSize)
{
  // How many data objects are included in "_dataStream"

  unsigned long size = 0;
  char c_curDataSize[4];
  unsigned long l_curDataSize;
  const char *curDataPos = dataStream;
  int counter = 0;

  while (size < dataSize)
  {
    counter++;

    c_curDataSize[3] = *curDataPos; curDataPos++;
    c_curDataSize[2] = *curDataPos; curDataPos++;
    c_curDataSize[1] = *curDataPos; curDataPos++;
    c_curDataSize[0] = *curDataPos; curDataPos++;
    size += 4;
    size++; curDataPos++; // skipping dataType_ID

    convertToLong(c_curDataSize, l_curDataSize);

    size += l_curDataSize;
    curDataPos += l_curDataSize;
  }

  return counter;
}

// Compress data
bool SmageDataProcessor::compressData()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::compressData()\n");
#endif

  // Compressing "_dataStream" and storing it in "_compressedDataStream";
  // "_compressedDataSize" holds the resulting size

  if (_currentDataSize > 5000000)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::compressData): data stream size is too large ***\n");
#endif
    return 0;
  }

  if (_currentDataSize != 0 && _compressedDataSize == 0)
  {
    _compressedDataSize = _currentDataSize + 10000;

    _compressedDataStream = new char[_compressedDataSize];
    if (_compressedDataStream == NULL)
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::compressData): out of memory ***\n");
#endif
      return 0;
    }

    // compression stream
    z_stream c_stream;
    int err;
      
    // Initializing the stream
    c_stream.zalloc = (alloc_func)0;
    c_stream.zfree = (free_func)0;
    c_stream.opaque = (voidpf)0;

    c_stream.next_in  = (Bytef*)_dataStream;
    c_stream.avail_in = (uInt)_currentDataSize;
    c_stream.next_out = (Bytef*)_compressedDataStream;
    c_stream.avail_out = (uInt)_compressedDataSize;

    err = deflateInit(&c_stream, Z_BEST_COMPRESSION);
#ifdef SMAGE_PRINT_DEBUG
    SMAGE_CHECK_ERR(err, "deflateInit");
#endif
      
    err = deflate(&c_stream, Z_FINISH);
    if (err != Z_STREAM_END)
    {
      deflateEnd(&c_stream);
#ifdef SMAGE_PRINT_DEBUG
      SMAGE_CHECK_ERR(err, "deflate");
#endif
    }
    
    _compressedDataSize = c_stream.total_out;

    err = deflateEnd(&c_stream);
#ifdef SMAGE_PRINT_DEBUG
    SMAGE_CHECK_ERR(err, "deflateEnd");
#endif
    return 1;
  }
  
  return 0;
}

// Decompress data
bool SmageDataProcessor::decompressData()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageDataProcessor::decompressData()\n");
#endif

  // Decompressing "_dataStream" and storing it in "_decompressedDataStream";
  // "_decompressedDataSize" holds the resulting size

  if (_currentDataSize > 5000000)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::decompressData): data stream size is too large ***\n");
#endif
    return 0;
  }

  if (_currentDataSize != 0 && _decompressedDataSize == 0)
  {
    _decompressedDataSize = 6000000;
    _decompressedDataStream = new char[_decompressedDataSize];
    if (_decompressedDataStream == NULL)
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageDataProcessor::decompressData): out of memory ***\n");
#endif
      return 0;
    }

    // decompression stream
    z_stream d_stream;
    int err;

    // Initializing the stream
    d_stream.zalloc = (alloc_func)0;
    d_stream.zfree = (free_func)0;
    d_stream.opaque = (voidpf)0;

    d_stream.next_in  = (Bytef*)_dataStream;
    d_stream.avail_in = (uInt)_currentDataSize;
    d_stream.next_out = (Bytef*)_decompressedDataStream;
    d_stream.avail_out = (uInt)_decompressedDataSize;

    err = inflateInit(&d_stream);
#ifdef SMAGE_PRINT_DEBUG
    SMAGE_CHECK_ERR(err, "inflateInit");
#endif

    err = inflate(&d_stream, Z_FINISH);
    if (err != Z_STREAM_END)
    {
      inflateEnd(&d_stream);
#ifdef SMAGE_PRINT_DEBUG
      SMAGE_CHECK_ERR(err, "inflate");
#endif
    }
    
    _decompressedDataSize = d_stream.total_out;

    err = inflateEnd(&d_stream);
#ifdef SMAGE_PRINT_DEBUG
    SMAGE_CHECK_ERR(err, "inflateEnd");
#endif
    return 1;
  }

  return 0;
}
