#ifndef  _SMAGE_DATA_PROCESSOR_H
#define  _SMAGE_DATA_PROCESSOR_H

#include <vector>

/** Data types that may be encoded.
  *
  * @ingroup LibSmage
  *
  */
enum DataType { 
  SMAGE_TEXT,
  SMAGE_FORM_TEXT,
  SMAGE_IMG, 
  SMAGE_COLOR_TABLE, 
  SMAGE_FILE,
  SMAGE_GAME,
  SMAGE_DICT
};

/** Properties for a Smage item.
  *
  * @ingroup LibSmage
  *
  */
enum SmageProps {
  SMAGE_LINK      = 1,                // Smage item is a link
  SMAGE_INVISIBLE = 2,                // Smage item should not be highlighted on request
  SMAGE_MISC      = 4
};

/** Structure that holds the properties for a single data item.
  *
  * @ingroup LibSmage
  *
  */
struct SmagePropsStruct {
  SmagePropsStruct();                 // Constructor
  char props;                         // Properties ID
  std::vector<char> link;             // If "SMAGE_LINK" is set, "link" holds the location
};

/** Code patterns; how many bits per pixel are coded
  *
  * @ingroup LibSmage
  *
  */
enum CodePattern {
  SMAGE_AUTO, SMAGE_LSB, SMAGE_BITS_3, SMAGE_BITS_4, 
  SMAGE_LS_2B, SMAGE_BITS_6, SMAGE_BITS_7, SMAGE_LS_3B, 
  SMAGE_BITS_9, SMAGE_BITS_10, SMAGE_LS_4B, SMAGE_BITS_12,
  SMAGE_BITS_13, SMAGE_LS_5B, SMAGE_LS_8B, SMAGE_MAX
};

#define SMAGE_MAX_BITS_PER_PIXEL 23

/** Structure that holds the necessary information for a single data item.
  *
  * @ingroup LibSmage
  *
  */
struct SmageData {
  SmageData();                   // Constructor
  int idColor[3];                // Color this data item is associated with in the ID-Buffer
  std::vector<char> byteStream;  // The actual data
  DataType dataType;             // Type of data
  char subID;                    // Data item's sub-ID (different data items may belong to different regions
                                 // in the ID-Buffer, but those regions may belong to one object with its ID)
};

/** Class SmageDataProcessor for generating and compressing a byte stream.
  *
  * @ingroup LibSmage
  *
  * A generated byte stream is composed of the single data items' 
  * size, object IDs, the type of data, and the actual data.
  * After generating, it is compressed using the "zlib".
  *
  * Usually, the class is used as follows:
  * <ol>
  *  <li> Get a class instance ("SmageDataProcessor()")
  *  <li> Set or add data ("setdData()" or "addData()")
  *  <li> Compress or decompress the data ("compressData()" or 
  *     "decompressData()")
  *  <li> Get the compressed or decompressed data stream
  *     ("getCompressedDataStream()" or "getDecompressedDataStream()")
  *  </ol>
  */
class SmageDataProcessor
{
  public:

    /** Constructor.
      */
    SmageDataProcessor();

    /** Destructor.
      */
    ~SmageDataProcessor();

    /** Method for adding data that shall be processed.
      */
    bool addData(const char *dataStream, const unsigned long &size, const DataType &dataType, const char &subID);

    /** Method for adding data that shall be processed.
      */
    bool addData(const std::vector<char> &dataStream, const DataType &dataType, const char &subID);

    /** Method for setting data that shall be processed.
      */
    bool setData(const char *dataStream, const unsigned long &size);

    /** Method for setting data that shall be processed.
      */
    bool setData(const std::vector<char> &dataStream);

    /** 
      */
    bool setProps(const char &props, const char *propsData, const unsigned long &size);

    /** 
      */
    void setProps(const SmagePropsStruct &props);

    /** Get a compressed byte stream from a vector of data items
      */
    void getPreparedDataStream(const std::vector<SmageData> &srcData, std::vector<char> &destData);

    /** Set class data to NULL.
      */
    void resetDataStream();

    /** Set class data to NULL.
      */
    void resetPropsStream();

    /** Get original data stream.
      */
    const char *getDataStream(unsigned long &dataStreamSize);

    /** Get compressed data stream.
      */
    const char *getCompressedDataStream(unsigned long &comprDataStreamSize);

    /** Get decompressed data stream.
      */
    const char *getDecompressedDataStream(unsigned long &decomprDataStreamSize);

    /** Unsplit the data stream (already assigned to the class) and extract the single data items
      */
    void unsplitDataStream(std::vector<SmageData> &dataStream, SmagePropsStruct &props, const bool &decompress);
      
    /** Converts enum to char.
      */
    void convertEnumToChar(const DataType &dataTypeSrc, char &dataTypeDest);

    /** Converts char to enum.
      */
    void convertCharToEnum(const char &dataTypeSrc, DataType &dataTypeDest);

    /** Converts enum to int.
      */
    void convertEnumToInt(const CodePattern &codePatternSrc, int &codePatternDest);

    /** Converts int to enum.
      */
    void convertIntToEnum(const int &codePatternSrc, CodePattern &codePatternDest);

    /** Compresses data and return if succeeded.
      */
    bool compressData();

    /** Decompresses data and return if succeeded.
      */
    bool decompressData();

  private:

    /** Converts to chars.
      */
    void convertToChars(const unsigned long &src, char dest[4]);
    
    /** Converts to long.
      */
    void convertToLong(const char src[4], unsigned long &dest);

    /** Encode and extract data type and sub-ID to/from data stream.
      */
    void encodeType_ID(const DataType &dataType, const char &subID, char &dataType_ID);

    /** Encode and extract data type and sub-ID to/from data stream.
      */
    void decodeType_ID(const char &dataType_ID, DataType &dataType, char &subID);

    /** Get the number of data items in a data stream
      */
    int getNumDataObjs(const char *dataStream, const unsigned long &dataSize);

    /** Copy "_propsStreamSize" and "_propsStream" to the beginning of "_dataStream"
      */
    void copyPropsToDataStream();

    /** Decode the coded properties
      */
    void decodeProps(unsigned long &curDataSize, const char **curDataPos, SmagePropsStruct &propsStruct);

    /** Size of the "_dataStream".
      */
    unsigned long _currentDataSize;

    /** Size of the "_propsStream".
      */
    unsigned long _propsStreamSize;

    /** Size of the "_compressedDataStream".
      */
    unsigned long _compressedDataSize;

    /** Size of the "_decompressedDataStream".
      */
    unsigned long _decompressedDataSize;

    /** Size of ?.
      */
    unsigned int _streamDataLength;

    /** holds only data.
      */
    char *_dataStream;

    /** holds a character specifying "SmageProps" + the properties data
      */
    char *_propsStream;

    /** compressed stream.
      */
    char *_compressedDataStream;

    /** decompressed stream.
      */
    char *_decompressedDataStream;
};

#endif
