#include "SmageIdBuffer.h"
#include "SmageSettings.h"

#include <iostream>
#include <algorithm>

using namespace std;

// Structure's constructor
SmageIdSubRegion::SmageIdSubRegion()
{
  subID = 0;
  numOfInnerPixels = 0;
  seedPointPos = QPoint(-1,-1);
}

// Structure's constructor
SmageIdRegion::SmageIdRegion()
{
  subRegions.clear();
  numOfAllInnerPixels = 0;
  numOfAllOutlinePixels = 0;
}

// Constructor; arguments are the size of the ID-Buffer that will be generated and if an
//  ID-Buffer does exist; if there is no ID-Buffer available ("fillBuffer" is TRUE), an ID-Buffer
//  with one main region, that extends over the wohle image, is created.
SmageIdBuffer::SmageIdBuffer(const int &imgWidth, const int &imgHeight,
           const bool &fillBuffer) : SmageIdBufferGenerator(imgWidth, imgHeight)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBuffer::SmageIdBuffer()\n");
#endif

  _regionMap.clear();  

  if (fillBuffer)
    fillCommonIdBuffer();
}

// Constructor with no arguments
SmageIdBuffer::SmageIdBuffer() : SmageIdBufferGenerator(0, 0)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBuffer::SmageIdBuffer()\n");
#endif
}

// Destructor
SmageIdBuffer::~SmageIdBuffer()
{
}

// If no ID-Buffer has been specified, a common ID-Buffer is generated; a common ID-Buffer
//  has only one region of the image's size and its outline is drawn along the image's border
void SmageIdBuffer::fillCommonIdBuffer()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBuffer::fillCommonIdBuffer()\n");
#endif
  
  // Append fill and outline color to the already existing ID-Buffer's color table
  _idBuffer.setNumColors(4);
  _idBuffer.setColor(COL_FILL_DEFAULT_INDEX, qRgb(COL_FILL_DEFAULT));
  _idBuffer.setColor(COL_BOUNDARY_DEFAULT_INDEX, qRgb(COL_BOUNDARY_DEFAULT));

  // Fill the ID-Buffer along the border (outline) with the according table index
  createCommonIdBuffer();

  // Seed point position is clear in a common ID-Buffer
  QPoint seedPoint(1,1);
  _idBuffer.setPixel(1, 1, COL_FILL_DEFAULT_INDEX);

  // Insert region data into "_regionMap"
  SmageIdSubRegion idSubRegion;
  idSubRegion.subID = 0;
  idSubRegion.seedPointPos = seedPoint;

  SmageIdRegion idRegion;
  idRegion.subRegions.push_back(idSubRegion);
  _regionMap[DEFAULT_MAP_POS] = idRegion;

  // Fill "_idColorMap" that holds for each object-ID the associated colors
  vector<SmageIdColorStruct> colors;
  SmageIdColorStruct colorStruct;
  colorStruct.subID = DEFAULT_MAP_POS;
  colors.push_back(colorStruct);
  _idColorMap[DEFAULT_MAP_POS] = colors;
  setIdColors(DEFAULT_MAP_POS, DEFAULT_MAP_POS, COL_FILL_DEFAULT_INDEX, qRgb(COL_FILL_DEFAULT),
        COL_BOUNDARY_DEFAULT_INDEX, qRgb(COL_BOUNDARY_DEFAULT));

  // Fill the region bounded by the outline and start from the seed point position
  fillSeedPointRegions();
}

//   Draw the outline in the common ID-Buffer
void SmageIdBuffer::createCommonIdBuffer()
{
  int k=0;
  int l=0;

  // Two vertical lines
  while (k<_idBuffer.height())
  {
    for (l=1; l<_idBuffer.width()-1; l++)
      *(_idBuffer.scanLine(k) + l) = COL_BOUNDARY_INDEX;

    k += _idBuffer.height()-1;
  }

  // Two horizontal lines
  for (k=1; k<_idBuffer.height()-1; k++)
  {
    l=0;
    while (l<_idBuffer.width())
    {
      *(_idBuffer.scanLine(k) + l) = COL_BOUNDARY_INDEX;
      l += _idBuffer.width()-1;
    }
  }
}

// Seed point positions and their outlines are given; fill those regions starting from the seed 
// points.
void SmageIdBuffer::fillSeedPointRegions()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBuffer::fillSeedPointRegions()\n");
#endif

  int currentColIndex;
  vector<SmageIdSubRegion>::iterator subRegionIter;
  vector<int> seedPointColorIndices;
  seedPointColorIndices.clear();

  unsigned long numAllPixels;

  // Seed point positions are extracted from the "_regionMap"
  map<unsigned int, SmageIdRegion>::iterator mapIter;
  for (mapIter=_regionMap.begin(); mapIter!=_regionMap.end(); mapIter++)
  {
    numAllPixels = 0;

    for (subRegionIter=mapIter->second.subRegions.begin(); 
      subRegionIter!=mapIter->second.subRegions.end(); subRegionIter++)
    {
      // The fill color is taken from the seed point's color
      currentColIndex = _idBuffer.pixelIndex(subRegionIter->seedPointPos.x(), subRegionIter->seedPointPos.y());

      // The seed point's color is set to background such that the fill operation can be done exactly
      _idBuffer.setPixel(subRegionIter->seedPointPos.x(), subRegionIter->seedPointPos.y(), COL_BACKGROUND_INDEX);

      // Seed point color indices are stored; they will be used during outline revision
      if (!findValue(seedPointColorIndices, currentColIndex))
        seedPointColorIndices.push_back(currentColIndex);

      // Fill the region that belongs to the current seed point and store the number of pixels that have been 
      // filled
      if (currentColIndex != COL_BACKGROUND_INDEX)
      {
        unsigned long numPixels = fillRegion(_idBuffer, subRegionIter->seedPointPos, COL_BACKGROUND_INDEX, 
               COL_BOUNDARY_INDEX, currentColIndex);
        numPixels -= getReservedPixelNum(subRegionIter->seedPointPos, COL_BOUNDARY_INDEX);
        subRegionIter->numOfInnerPixels = numPixels;

        numAllPixels += numPixels;
      }
      else
      {
#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("*** Warning (SmageIdBuffer::fillSeedPointRegions): seed point pixel color is the same as the background color ***\n");
#endif
      }
    }

    mapIter->second.numOfAllInnerPixels = numAllPixels;
  }

  // Revise the outlines and store the number of outline pixels
  map<unsigned int, unsigned long> outlinePixelMap;
  reviseOutlines(seedPointColorIndices, COL_BACKGROUND_INDEX, COL_BOUNDARY_INDEX, outlinePixelMap);
  assignNumOutlinePixels(outlinePixelMap);
}

// Only outlines are given; fill those regions, search for seed points and fill the "_regionMap"
bool SmageIdBuffer::fillOutlineRegions()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBuffer::fillOutlineRegions()\n");
#endif

  // Each outline must have its own color; if two outlines have the same colors,
  // the first outline's inner region is filled but second is ignored

  // (1) Find a seed point and verify it
  // (2) Fill a region starting from the seed point

  // "_tmpBuffer" is also filled during the fill operation

  int width = _idBuffer.width();
  int height = _idBuffer.height();
  
  int currentColIndex;
  
  QPoint possibleSeedPoint;
  bool havePossibleSeedPoint;

  // Colors of already filled outlines
  vector<int> filledOutlines;
  filledOutlines.clear();

  for (int y=0; y<height; y++)  
  {
    havePossibleSeedPoint = FALSE;

    for (int x=0; x<width; x++)
    {
      currentColIndex = _idBuffer.pixelIndex(x, y);
  
      // Current color index should be different from background color;
      // to verify that it is an outline pixel, it is also compared with
      // "_tmpBuffer" which is filled during the fill operation (outline
      // pixels are not filled here)
      if (currentColIndex != COL_BACKGROUND_INDEX &&
        _tmpBuffer.pixelIndex(x, y) != COL_TMP_FILL_INDEX)
      {
        // Found an outline pixel

        if (!findValue(filledOutlines, currentColIndex))
        {
          if (!havePossibleSeedPoint)
          {
            // Position (x+1,y) might be a seed point position

            if (_idBuffer.valid(x+1, y) && _idBuffer.valid(x-1, y))
            {
              if (_idBuffer.pixelIndex(x+1, y) == COL_BACKGROUND_INDEX &&
                !isVerticalPeak(QPoint(x,y), currentColIndex))
              {
                possibleSeedPoint = QPoint(x+1, y);
                havePossibleSeedPoint = TRUE;
              }
            }
            else if (_idBuffer.valid(x+1, y))
            {
              if (_idBuffer.pixelIndex(x+1, y) == COL_BACKGROUND_INDEX)
              {
                possibleSeedPoint = QPoint(x+1, y);
                havePossibleSeedPoint = TRUE;
              }
            }
          }
          else
          {
            // Have another outline pixel which should be the second one of
            // a pair of outline pixels on a scan line; otherwise outline
            // pixel would be a peak

            if (_idBuffer.valid(x+1, y) && _idBuffer.valid(x-1, y))
            {
              if (_idBuffer.pixelIndex(x-1, y) == COL_BACKGROUND_INDEX && 
                !isVerticalPeak(QPoint(x,y), currentColIndex))
              {            
                // Have a possible seed point ... verify it
                if (isSeedPoint(possibleSeedPoint, currentColIndex))
                {
                  // "possibleSeedPoint" is a seed point; start filling its region
                  unsigned long numPixels = 
                     fillRegion(_idBuffer, possibleSeedPoint, COL_BACKGROUND_INDEX, 
                     currentColIndex, currentColIndex-1);
                  numPixels -= getReservedPixelNum(possibleSeedPoint, currentColIndex);

                  havePossibleSeedPoint = FALSE;
                  filledOutlines.push_back(currentColIndex);

#ifdef SMAGE_PRINT_INFOS
                  cout << "Found a seed point at: " << possibleSeedPoint.x() << " " 
                  << possibleSeedPoint.y() << endl;
#endif

                  // Update "_regionMap"
                  SmageIdSubRegion idSubRegion;
                  idSubRegion.seedPointPos = possibleSeedPoint;
                  idSubRegion.numOfInnerPixels = numPixels;

                  int objId, objSubId;
                  if (getObjectId(_idBuffer.color(currentColIndex-1), objId, objSubId))
                  {
                    map<unsigned int, SmageIdRegion>::iterator regionMapIter = _regionMap.find(objId);
                    if (regionMapIter != _regionMap.end())
                    {
                      regionMapIter->second.subRegions.push_back(idSubRegion);
                      regionMapIter->second.numOfAllInnerPixels += numPixels;
                    }
                    else
                    {
                      SmageIdRegion idRegion;
                      idRegion.subRegions.push_back(idSubRegion);
                      idRegion.numOfAllInnerPixels = numPixels;
                      _regionMap[objId] = idRegion;
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  // Revise the outlines and store the number of outline pixels
  map<unsigned int, unsigned long> outlinePixelMap;
  reviseOutlines(COL_BACKGROUND_INDEX, outlinePixelMap);
  assignNumOutlinePixels(outlinePixelMap);

  return TRUE;
}

// Connected filled regions are given; generate their outlines, search for seed points and 
//  fill the "_regionMap"
void SmageIdBuffer::generateOutlines()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBuffer::generateOutlines()\n");
#endif

  // Given an ID-Buffer with filled regions;
  // - generating outlines of those regions
  // - searching for seed points for each region

  unsigned char *curValueLine, *nextValueLine, *prevValueLine;
  unsigned char curLineCurCol, curLinePrevCol, curLineNextCol, prevLineCurCol, nextLineCurCol;

  int width = _idBuffer.width();
  int height = _idBuffer.height();
  
  // Temporary buffer that will later be used (during the search for seed points) to 
  // identify a region that belongs to a certain seed point; only outline pixels
  // are set
  QImage tmpFillIdBuffer = QImage(width, height, 8, 3);
  tmpFillIdBuffer.setColor(COL_BACKGROUND_INDEX, qRgb(COL_BACKGROUND));
  tmpFillIdBuffer.setColor(COL_BOUNDARY_INDEX, qRgb(COL_BOUNDARY));
  tmpFillIdBuffer.setColor(COL_FILL_DEFAULT_INDEX, qRgb(COL_FILL_DEFAULT));
  tmpFillIdBuffer.fill(COL_BACKGROUND_INDEX);

  // Pixels at the image border are considered separately

  for (int y=1; y<height-1; y++)
  {
    prevValueLine = _idBuffer.scanLine(y-1);
    curValueLine = _idBuffer.scanLine(y);
    nextValueLine = _idBuffer.scanLine(y+1);

    for (int x=1; x<width-1; x++)
    {
      curLineCurCol = curValueLine[x];

      if (curLineCurCol != COL_BACKGROUND_INDEX)
      {
        // 4 neighborhood of current pixel
        curLinePrevCol = curValueLine[x-1];
        curLineNextCol = curValueLine[x+1];
        prevLineCurCol = prevValueLine[x];
        nextLineCurCol = nextValueLine[x];

        // Test if one of the current pixel's neighbors has a different color than the current
        // color; if so, current pixel is an outline pixel
        if ((curLinePrevCol != curLineCurCol && (curLinePrevCol != (curLineCurCol+1))) || 
           (curLineNextCol != curLineCurCol && (curLineNextCol != (curLineCurCol+1))) ||
           (prevLineCurCol != curLineCurCol && (prevLineCurCol != (curLineCurCol+1))) || 
           (nextLineCurCol != curLineCurCol && (nextLineCurCol != (curLineCurCol+1))))
        {
          _idBuffer.setPixel(x, y, curLineCurCol+1);
          tmpFillIdBuffer.setPixel(x, y, COL_BOUNDARY_INDEX);
        }
        else
        {
          _tmpBuffer.setPixel(x, y, COL_TMP_FILL_INDEX);
        }
      }
    }
  }

  // Find outline pixels at the image boundary

  int k=0;
  int l=0;

  // First and last line
  while (k < height)
  {
    curValueLine = _idBuffer.scanLine(k);

    for (l=0; l<width; l++)
    {
      curLineCurCol = curValueLine[l];

      if (curLineCurCol != COL_BACKGROUND_INDEX)
      {
        _idBuffer.setPixel(l, k, curLineCurCol+1);
        tmpFillIdBuffer.setPixel(l, k, COL_BOUNDARY_INDEX);
      }
    }

    k += height-1;
  }

  // First and last column
  for (k=1; k<height-1; k++)
  {
    curValueLine = _idBuffer.scanLine(k);

    l=0;
    while (l<width)
    {
      curLineCurCol = curValueLine[l];

      if (curLineCurCol != COL_BACKGROUND_INDEX)
      {
        _idBuffer.setPixel(l, k, curLineCurCol+1);
        tmpFillIdBuffer.setPixel(l, k, COL_BOUNDARY_INDEX);
      }

      l += width-1;
    }
  }

  // Revise the outlines and store the number of outline pixels
  map<unsigned int, unsigned long> outlinePixelMap;
  reviseOutlines(COL_BACKGROUND_INDEX, outlinePixelMap);

  // Search for seed points
  searchForSeedpoints(tmpFillIdBuffer);

  assignNumOutlinePixels(outlinePixelMap);
}

//  Connected filled regions and their outlines are given; search for seed points and 
//  fill the "_regionMap"
void SmageIdBuffer::analyzeIdBuffer()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBuffer::analyzeIdBuffer()\n");
#endif

  // Only "_tmpBuffer" and "tmpFillIdBuffer" are filled since they will be
  // used during seed point search

  int width = _idBuffer.width();
  int height = _idBuffer.height();

  // Temporary buffer that will later be used (during the search for seed points) to 
  // identify a region that belongs to a certain seed point
  QImage tmpFillIdBuffer = QImage(width, height, 8, 3);
  tmpFillIdBuffer.setColor(COL_BACKGROUND_INDEX, qRgb(COL_BACKGROUND));
  tmpFillIdBuffer.setColor(COL_BOUNDARY_INDEX, qRgb(COL_BOUNDARY));
  tmpFillIdBuffer.setColor(COL_FILL_DEFAULT_INDEX, qRgb(COL_FILL_DEFAULT));
  tmpFillIdBuffer.fill(COL_BACKGROUND_INDEX);

  int currentColIndex;

  for (int y=0; y<height; y++)
    for (int x=0; x<width; x++)
    {
      currentColIndex = _idBuffer.pixelIndex(x, y) - NUM_DEFAULT_TABLE_ENTRIES;
    
      // If current color index is even, a fill color has been used and
      // thus, "_tmpBuffer" is filled; otherwise "tmpFillIdBuffer" is filled

      if (currentColIndex >= 0)
      {
        if (isEven(currentColIndex-NUM_DEFAULT_TABLE_ENTRIES))
          _tmpBuffer.setPixel(x, y, COL_TMP_FILL_INDEX);
        else
          tmpFillIdBuffer.setPixel(x, y, COL_BOUNDARY_INDEX);
      }
    }

  // Revise the outlines and store the number of outline pixels
  map<unsigned int, unsigned long> outlinePixelMap;
  reviseOutlines(COL_BACKGROUND_INDEX, outlinePixelMap);

  // Search for seed points
  searchForSeedpoints(tmpFillIdBuffer);

  assignNumOutlinePixels(outlinePixelMap);  
}

// Search for seed points in an ID-Buffer whose single regions are already outlined
void SmageIdBuffer::searchForSeedpoints(QImage &tmpIdBuffer)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBuffer::searchForSeedpoints()\n");
#endif

  // Seed point positions are searched and the "_regionMap" is updated

  int width = _idBuffer.width();
  int height = _idBuffer.height();
  
  int curIdBufferIndex;

  for (int y=1; y<height-1; y++)
    for (int x=1; x<width-1; x++)
    {
      // Seed point must be located in a filled region
      if (_tmpBuffer.pixelIndex(x, y) == COL_TMP_FILL_INDEX)
      {
        // Since "tmpIdBuffer" was not yet filled within an outline, it can
        // be used to verify that this region does not already have a seed point
        if (tmpIdBuffer.pixelIndex(x, y) != COL_FILL_DEFAULT_INDEX)
        {
          curIdBufferIndex = _idBuffer.pixelIndex(x, y);

          // If seed point position is (x,y), pixels at (x-1,y) and (x,y-1) must be
          // outline pixels
          if (_idBuffer.pixelIndex(x-1, y) == (curIdBufferIndex+1) &&
              _idBuffer.pixelIndex(x, y-1) == (curIdBufferIndex+1))
          {
            // Fill the region in "tmpIdBuffer"

            unsigned long numPixels =
               fillRegion(tmpIdBuffer, QPoint(x,y), COL_BACKGROUND_INDEX, COL_BOUNDARY_INDEX, 
               COL_FILL_DEFAULT_INDEX);
            numPixels -= getReservedPixelNum(QPoint(x, y), curIdBufferIndex+1);

#ifdef SMAGE_PRINT_INFOS
            cout << "Found a seed point at: " << x << " " << y << endl;
#endif

            // Update "_regionMap"
            SmageIdSubRegion idSubRegion;
            idSubRegion.seedPointPos = QPoint(x, y);
            idSubRegion.numOfInnerPixels = numPixels;

            int objId, objSubId;
            if (getObjectId(_idBuffer.color(curIdBufferIndex), objId, objSubId))
            {
              map<unsigned int, SmageIdRegion>::iterator regionMapIter = _regionMap.find(objId);
              if (regionMapIter != _regionMap.end())
              {
                regionMapIter->second.subRegions.push_back(idSubRegion);
                regionMapIter->second.numOfAllInnerPixels += numPixels;
              }
              else
              {
                SmageIdRegion idRegion;
                idRegion.subRegions.push_back(idSubRegion);
                idRegion.numOfAllInnerPixels = numPixels;
                _regionMap[objId] = idRegion;
              }
            }
          }
        }
      }
    }
}

// Does "point" fullfill the criteria of a seed point:
//  - there is only one seed point for each image region
//  - the first pixel (at position (x,y), scanning the image top-down) within an 
//    outline is a seed point if:
//    pixels at (x-1,y) and at (x,y-1) are outline pixels
bool SmageIdBuffer::isSeedPoint(const QPoint &point, const int &colIndex)
{
  // "colIndex" is an outline pixel's color index

  if (_idBuffer.valid(point.x(), point.y()-1) && _idBuffer.valid(point.x()-1, point.y()))
  {
    return ((_idBuffer.pixelIndex(point.x(), point.y()) == COL_BACKGROUND_INDEX) &&
        (_idBuffer.pixelIndex(point.x()-1, point.y()) == colIndex) &&
        (_idBuffer.pixelIndex(point.x(), point.y()-1) == colIndex));
  }
  
  return FALSE;
}

// During revision of outlines, outline pixels are counted and assigned to each "SmageIdRegion"
//  as part of the "_regionMap"
void SmageIdBuffer::assignNumOutlinePixels(const map<unsigned int, unsigned long> &outlineCounterMap)
{
  map<unsigned int, unsigned long>::const_iterator counterMapIter;
  map<unsigned int, SmageIdRegion>::iterator regionMapIter;
  for (counterMapIter=outlineCounterMap.begin(); counterMapIter!=outlineCounterMap.end(); counterMapIter++)
  {
    regionMapIter = _regionMap.find(counterMapIter->first);
    if (regionMapIter != _regionMap.end())
    {

#ifdef SMAGE_PRINT_INFOS
      cout << "Number of outline pixels for object-ID: " << counterMapIter->first << " is " 
           << counterMapIter->second << endl;
#endif
    
      // At least two of the pixels of each outline are reserved for seed point identification
      // and thus, they are not available for encoding data
      regionMapIter->second.numOfAllOutlinePixels = counterMapIter->second - 2;
    }
    else
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.
      lprintf("*** Warning (SmageIdBuffer::assignNumOutlinePixels): could not find object-ID in \"_regionMap\" ***\n");
#endif
    }
  }
}

// The identification pixels around a seed point are subtracted from "numOfAllInnerPixels" 
//  in "SmageIdRegion" since they cannot be used for information encoding
int SmageIdBuffer::getReservedPixelNum(const QPoint &seedPointPos, const int &outlineIndex)
{
  // Seed point position is the first reserved pixel position within an outline
  int counter = 1;

  if (_idBuffer.valid(seedPointPos.x()+1, seedPointPos.y()))
    if (_idBuffer.pixelIndex(seedPointPos.x()+1, seedPointPos.y()) != outlineIndex)
      counter++;

  if (_idBuffer.valid(seedPointPos.x(), seedPointPos.y()+1))
    if (_idBuffer.pixelIndex(seedPointPos.x(), seedPointPos.y()+1) != outlineIndex)
      counter++;

  return counter;
}
