#ifndef  _SMAGE_ID_BUFFER_H
#define  _SMAGE_ID_BUFFER_H

#include "SmageIdBufferGenerator.h"

/** Each "SmageIdRegion" may be composed of single sub-regions that belong to
  * the same region. Each "SmageIdSubRegion" has a sub-ID, a seed point, and
  * a number of pixels which belong to that region.
  *
  * @ingroup LibSmage
  *
  */
struct SmageIdSubRegion {
  SmageIdSubRegion();              // Constructor
  char subID;                      // ID for that sub-region
  unsigned long numOfInnerPixels;  // Number of pixels within the outlines sub-region
  QPoint seedPointPos;             // Position of the sub-region's seed point
};

/** Structure for all the regions that belong to a specific object-ID.
  *
  * @ingroup LibSmage
  *
  */
struct SmageIdRegion {
  SmageIdRegion();                           // Constructor
  std::vector<SmageIdSubRegion> subRegions;  // Sub-regions that belong to that main region
  unsigned long numOfAllInnerPixels;         // Number of all the region's inner pixels
  unsigned long numOfAllOutlinePixels;       // Number of all the region's outline pixels
};

/** Class SmageIdBuffer that processes the ID-Buffer image. 
  *
  * @ingroup LibSmage
  *
  * Class is derived from "SmageIdBufferGenerator".
  * Once an ID-Buffer image has been set and converted into 
  * an indexed image according the defined structure, outlines
  * are calculated or the area within outlines is filled and
  * seed point positions are searched.
  */
class SmageIdBuffer : public SmageIdBufferGenerator
{
  public:

    /** Constructor.
      * arguments are the size of the ID-Buffer that will be generated and if an
      * ID-Buffer does exist; if there is no ID-Buffer available ("fillBuffer" is TRUE), an ID-Buffer
      * with one main region, that extends over the wohle image, is created.
      */
    SmageIdBuffer(const int &width, const int &height, const bool &fillBuffer);
    SmageIdBuffer();

    /** Destructor.
      */
    ~SmageIdBuffer();

    /** Set and get a region map that holds for each object-ID the associated image regions with their seed points.
      */
    const std::map<unsigned int, SmageIdRegion> &getRegionMap() const { return _regionMap; }
    void setRegionMap(const std::map<unsigned int, SmageIdRegion> &regionMap) { _regionMap = regionMap; }

    /** Seed point positions and their outlines are given; fill those regions starting from the seed points.
      */
    void fillSeedPointRegions();

    /** Only outlines are given.
      * fill those regions, search for seed points and fill the "_regionMap"
      */
    bool fillOutlineRegions();
      
    /** Connected filled regions are given; generate their outlines, search for seed points and fill the "_regionMap"
      */
    void generateOutlines();

    /** Connected filled regions and their outlines are given; search for seed points and fill the "_regionMap".
      */
    void analyzeIdBuffer();

  private:

    /** If no ID-Buffer has been specified, a common ID-Buffer is generated; a common ID-Buffer
      * has only one region of the image's size and its outline is drawn along the image's border
      */
    void fillCommonIdBuffer();
    void createCommonIdBuffer();

    /** Search for seed points in an ID-Buffer whose single regions are already outlined
      */
    void searchForSeedpoints(QImage &tmpIdBuffer);

    /** During revision of outlines, outline pixels are counted and assigned to each "SmageIdRegion"
      * as part of the "_regionMap"
      */
    void assignNumOutlinePixels(const std::map<unsigned int, unsigned long> &outlineCounterMap);

    /** The identification pixels around a seed point are subtracted from "numOfAllInnerPixels" 
      * in "SmageIdRegion" since they cannot be used for information encoding
      */
    int getReservedPixelNum(const QPoint &seedPointPos, const int &outlineIndex);

    /** Does "point" fullfill the criteria of a seed point
      */
    bool isSeedPoint(const QPoint &point, const int &colIndex);

    /** Map that holds for each object-ID the associated seed point; since a seed point is 
      * always located within such a region, all the region's pixels can simply be determined
      * through a fill operation
      */
    std::map<unsigned int, SmageIdRegion> _regionMap;
};

#endif
