#include "SmageIdBufferGenerator.h"
#include "SmageSettings.h"

#include <iostream>
#include <math.h>

#include <algorithm>

using namespace std;

// Structure's constructor

SmageIdBufferStruct::SmageIdBufferStruct()
{
  idBuffer = QImage();
  backgroundColor = QColor(COL_BACKGROUND);
  bufferType = SMAGE_FILLED_REGIONS;
}

// Structure's constructor
SmageIdColorStruct::SmageIdColorStruct()
{
  subID = 0;
  fillColor = qRgb(COL_FILL_DEFAULT);
  outlineColor = qRgb(COL_BOUNDARY);
  fillColorIndex = COL_FILL_DEFAULT_INDEX;
  outlineColorIndex = COL_BOUNDARY_INDEX;
}

// Constructor; ID-Buffer is initialized as 8 bits per pixel image
SmageIdBufferGenerator::SmageIdBufferGenerator(const int &width, const int &height)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBufferGenerator::SmageIdBufferGenerator(): width: %d  height: %d\n", width, height);
#endif

  if (width > 0 && height > 0)
  {
    _idBuffer = QImage(width, height, 8);  

    initializeIdBuffer();

    // Temporary ID-Buffer; has only two colors;
    // is filled as reference during the fill operation
    _tmpBuffer = QImage(width, height, 8, 2);
    _tmpBuffer.setColor(COL_BACKGROUND_INDEX, qRgb(COL_BACKGROUND));
    _tmpBuffer.setColor(COL_TMP_FILL_INDEX, qRgb(COL_FILL_DEFAULT));
    _tmpBuffer.fill(COL_BACKGROUND_INDEX);      
  }
  else
    _idBuffer = QImage();  

  _idColorMap.clear();
}

//  Destructor
SmageIdBufferGenerator::~SmageIdBufferGenerator()
{
}

// Color comparison
bool SmageIdBufferGenerator::areEqualColors(const QRgb &c1, const QRgb &c2)
{
  return (qRed(c1)==qRed(c2) && qGreen(c1)==qGreen(c2) && qBlue(c1)==qBlue(c2));
}

bool SmageIdBufferGenerator::areEqualColors(const QColor &c1, const QColor &c2)
{
  return (c1.red()==c2.red() && c1.green()==c2.green() && c1.blue()==c2.blue());
}

bool SmageIdBufferGenerator::areEqualColors(const QRgb &c1, const QColor &c2)
{
  return (qRed(c1)==c2.red() && qGreen(c1)==c2.green() && qBlue(c1)==c2.blue());
}


//   Is the number even
bool SmageIdBufferGenerator::isEven(const int &number)
{
  if (fmod(double(number), double(2.0)) == 0) return TRUE;
  else return FALSE;
}

//   Is specified "value" included in the vector
bool SmageIdBufferGenerator::findValue(const vector<int> &values, const int &value)
{
  return (!(find(values.begin(), values.end(), value) == values.end()));
}

bool SmageIdBufferGenerator::findValue(const vector<QColor> &values, const QColor &value)
{
  vector<QColor>::const_iterator colIter;

  for (colIter=values.begin(); colIter!=values.end(); colIter++)
  {
    if (areEqualColors(*colIter, value))
      return TRUE;
  }

  return FALSE;
}

//  Is the specified point an upper or a lower peak in the outline;
//  is the point an extreme vertex whose right and left neighbors do neither
//  belong to the outline nor to the outline's inner pixels
//
//         #  <- Upper peak
//  # #
//       #   #
//     #      #
bool SmageIdBufferGenerator::isVerticalPeak(const QPoint &point, const int &colIndex)
{
  bool result = FALSE;

  int x = point.x();
  int y = point.y();

  // Have right and left neighbor pixels same color
  if (_idBuffer.pixelIndex(x-1, y) != colIndex && _idBuffer.pixelIndex(x+1, y) != colIndex)
    result = TRUE;

  if (result)
  {
    // Is upper peak?
    if (_idBuffer.valid(x, y-1))
    {
      if (_idBuffer.pixelIndex(x-1, y-1) == colIndex ||
          _idBuffer.pixelIndex(x, y-1) == colIndex ||
          _idBuffer.pixelIndex(x+1, y-1) == colIndex)
        result = FALSE;
    }

    // Is lower peak?
    if (_idBuffer.valid(x, y+1) && !result)
    {
      if (!(_idBuffer.pixelIndex(x-1, y+1) == colIndex ||
            _idBuffer.pixelIndex(x, y+1) == colIndex ||
            _idBuffer.pixelIndex(x+1, y+1) == colIndex))
        result = TRUE;
    }
  }
  
  return result;
}

//   Initialize the ID-Buffer with 2 colors and fill it with "COL_BACKGROUND"
void SmageIdBufferGenerator::initializeIdBuffer()
{
  _idBuffer.setNumColors(2);
  _idBuffer.setColor(COL_BACKGROUND_INDEX, qRgb(COL_BACKGROUND));
  _idBuffer.setColor(COL_BOUNDARY_INDEX, qRgb(COL_BOUNDARY));

  _idBuffer.fill(COL_BACKGROUND_INDEX);
}

//  Revise outlines and get the number of outline pixels;
//  here all outlines have the same color index "boundaryColorIndex";
void SmageIdBufferGenerator::reviseOutlines(const vector<int> &seedPointColorIndices, const int &backgroundColorIndex, 
              const int &boundaryColorIndex, map<unsigned int, unsigned long> &outlinePixelMap)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBufferGenerator::reviseOutlines\n");
#endif

  // There may be outline pixels which are not adjacent to their enclosed region;
  // that means none of their 4 direct neighbors has the color of the enclosed region
  // (1) Find those useless outline pixels

  // "seedPointColorIndices" is a vector of all the seed points' color indices,
  // which are the same colors as the enclosed regions' colors
  // (2) Since all outline pixels are of the same color, a new color (which is just the color 
  //     index of the seed point + 1) according to the associated seed point
  //     color is assigned to the respective outline pixels

  int currentColIndex;
  bool setNewOutline;
  int objectId, objectSubId;
  map<unsigned int, unsigned long>::iterator outlinePixelMapIter;

  for (int y=0; y<_idBuffer.height(); y++)
    for (int x=0; x<_idBuffer.width(); x++)
    {
      currentColIndex = _idBuffer.pixelIndex(x, y);

      if (currentColIndex == boundaryColorIndex)
      {
        // Current pixel is an outline pixel

        setNewOutline = FALSE;

        // Trying to find a neighbor pixel whose color is among the
        // "seedPointColorIndices"; if none such pixel is found, the current
        // outline pixel's color is set to background
        if (_idBuffer.valid(x, y-1))
        {
          if (findValue(seedPointColorIndices, _idBuffer.pixelIndex(x, y-1)))
          {
            _idBuffer.setPixel(x, y, _idBuffer.pixelIndex(x, y-1)+1);
            setNewOutline = TRUE;
          }
        }
        if (_idBuffer.valid(x+1, y) && !setNewOutline)
        {
        if (findValue(seedPointColorIndices, _idBuffer.pixelIndex(x+1, y)))
        {
          _idBuffer.setPixel(x, y, _idBuffer.pixelIndex(x+1, y)+1);
          setNewOutline = TRUE;
        }
      }
      if (_idBuffer.valid(x, y+1) && !setNewOutline)
      {
        if (findValue(seedPointColorIndices, _idBuffer.pixelIndex(x, y+1)))
        {
          _idBuffer.setPixel(x, y, _idBuffer.pixelIndex(x, y+1)+1);
          setNewOutline = TRUE;
        }
      }
      if (_idBuffer.valid(x-1, y) && !setNewOutline)
      {
        if (findValue(seedPointColorIndices, _idBuffer.pixelIndex(x-1, y)))
        {
          _idBuffer.setPixel(x, y, _idBuffer.pixelIndex(x-1, y)+1);
          setNewOutline = TRUE;
        }
      }

      if (!setNewOutline)
        _idBuffer.setPixel(x, y, backgroundColorIndex);
      else
      {
        // Update outline pixel counter in the "outlinePixelMap"
        if (getObjectId(_idBuffer.color(_idBuffer.pixelIndex(x, y)-1), objectId, objectSubId))
        {
          outlinePixelMapIter = outlinePixelMap.find(objectId);
          if (outlinePixelMapIter != outlinePixelMap.end())
            outlinePixelMapIter->second++;
          else
          {
            unsigned long counter = 1;
            outlinePixelMap[objectId] = counter;
          }
        }
        else
        {
#ifdef SMAGE_PRINT_DEBUG
          _SmageLogFile.lprintf("*** Warning (SmageIdBufferGenerator::reviseOutlines): could not find object ID ***\n");
#endif
        }  
      }
    }
  }
}

// Revise outlines and get the number of outline pixels;
//  here, different outlines have different colors
void SmageIdBufferGenerator::reviseOutlines(const int &backgroundColorIndex,
              map<unsigned int, unsigned long> &outlinePixelMap)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBufferGenerator::reviseOutlines\n");
#endif

  // There may be outline pixels which are not adjacent to their enclosed region;
  // that means none of their 4 direct neighbors has the color of the enclosed region;
  // those useless outline pixels are set to the background color;
  // "_tmpBuffer" is used since it has all the regions that have previously been filled

  bool pixelDone;
  int objectId, objectSubId;
  map<unsigned int, unsigned long>::iterator outlinePixelMapIter;

  for (int y=0; y<_idBuffer.height(); y++)
    for (int x=0; x<_idBuffer.width(); x++)
    {
      pixelDone = FALSE;
  
      if (_tmpBuffer.pixelIndex(x, y) != COL_TMP_FILL_INDEX)
      {
        // If current pixel is not filled in "_tmpBuffer", it may be an outline
        // or a background pixel;
        // if none of current pixel's neighbor pixels is filled in "_tmpBuffer",
        // current pixel is set to background

        if (_idBuffer.valid(x-1, y))
        {
          if (_tmpBuffer.pixelIndex(x-1, y) == COL_TMP_FILL_INDEX)
            pixelDone = TRUE;
        }
        if (_idBuffer.valid(x, y-1) && !pixelDone)
        {
          if (_tmpBuffer.pixelIndex(x, y-1) == COL_TMP_FILL_INDEX)
            pixelDone = TRUE;
        }
        if (_idBuffer.valid(x+1, y) && !pixelDone)
        {
          if (_tmpBuffer.pixelIndex(x+1, y) == COL_TMP_FILL_INDEX)
            pixelDone = TRUE;
        }
        if (_idBuffer.valid(x, y+1) && !pixelDone)
        {
          if (_tmpBuffer.pixelIndex(x, y+1) == COL_TMP_FILL_INDEX)
            pixelDone = TRUE;
        }

        if (!pixelDone)
          _idBuffer.setPixel(x, y, backgroundColorIndex);
        else
        {    
          // Update outline pixel counter in the "outlinePixelMap"
          if (getObjectId(_idBuffer.color(_idBuffer.pixelIndex(x, y)-1), objectId, objectSubId))
          {
            outlinePixelMapIter = outlinePixelMap.find(objectId);
            if (outlinePixelMapIter != outlinePixelMap.end())
              outlinePixelMapIter->second++;
            else
            {
              unsigned long counter = 1;
              outlinePixelMap[objectId] = counter;
            }
          }
          else
          {
#ifdef SMAGE_PRINT_DEBUG
            _SmageLogFile.lprintf("*** Warning (SmageIdBufferGenerator::reviseOutlines): could not find object ID ***\n");
#endif
          }
        }
      }
    }
}

//  Fill the region within an outline;
//  the number of pixels within the outlined region is returned;
unsigned long SmageIdBufferGenerator::fillRegion(QImage &idBuffer, const QPoint &seedPointPos, 
             const int &backgroundColorIndex, const int &boundaryIndex, 
             const int &fillIndex)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBufferGenerator::fillRegion() for seed point at: %d %d\n", seedPointPos.x(), seedPointPos.y());
#endif

  // Fill a region bounded by an already defined boundary in "idBuffer"
  // starting at "seedPointPos";
  // Scanline-Algorithm - for each scanline a focus point is defined from 
  // which scanning is started (focus points are stored on two stacks, 
  // one for going upwards and one for going downwards)

  unsigned long numOfAvailablePixel = 0;

  if (idBuffer.valid(seedPointPos.x(), seedPointPos.y()))
  {
    stack<QPoint> stackUp;
    stack<QPoint> stackDown;
    bool proceedWithStacks = TRUE;
    QPoint focusPoint;

    updateStacks(idBuffer, seedPointPos, backgroundColorIndex, fillIndex, boundaryIndex, 
      stackUp, stackDown, numOfAvailablePixel);
      
    while (proceedWithStacks)
    {
      // Process stack with focus points for the upper part
      while (!stackUp.empty())
      {
        focusPoint = stackUp.top();
        stackUp.pop();
        updateStacks(idBuffer, focusPoint, backgroundColorIndex, fillIndex, boundaryIndex, 
          stackUp, stackDown, numOfAvailablePixel);
      }

      // Process stack with focus points for the lower part
      while (!stackDown.empty())
      {
        focusPoint = stackDown.top();
        stackDown.pop();
        updateStacks(idBuffer, focusPoint, backgroundColorIndex, fillIndex, boundaryIndex,
          stackUp, stackDown, numOfAvailablePixel);
      }

      // If both stacks are empty, fill operation is finished
      if (stackUp.empty() && stackDown.empty())
        proceedWithStacks = FALSE;
    }
  }
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageIdBufferGenerator::fillRegion): seed point position is invalid ***\n");
#endif
  }
  
  return numOfAvailablePixel;
}

// Fill the pixels along a scan line and 
//  search for new focus points and put them on the respective stack
void SmageIdBufferGenerator::updateStacks(QImage &idBuffer, const QPoint &focusPoint, 
            const int &backgroundColorIndex, const int &fillColorIndex, 
            const int &boundaryIndex,
            stack<QPoint> &stackUp, stack<QPoint> &stackDown,
            unsigned long &numOfAvailablePixel)
{
  // Fill current scan line (including "focusPoint") with "fillColorIndex" and
  // find further focus points

  int currentColIndex, neighborColIndex1, neighborColIndex2;
  bool putOnStack;
  int currentPos_X = focusPoint.x();

  if (idBuffer.valid(currentPos_X, focusPoint.y()-1))
    if (idBuffer.pixelIndex(currentPos_X, focusPoint.y()-1) == backgroundColorIndex)
      stackUp.push(QPoint(currentPos_X, focusPoint.y()-1));

  if (idBuffer.valid(currentPos_X, focusPoint.y()+1))
    if (idBuffer.pixelIndex(currentPos_X, focusPoint.y()+1) == backgroundColorIndex)
      stackDown.push(QPoint(currentPos_X, focusPoint.y()+1));

  // Go to the right
  currentColIndex = idBuffer.pixelIndex(currentPos_X, focusPoint.y());
  while (currentColIndex != boundaryIndex && currentColIndex != fillColorIndex &&
   idBuffer.valid(currentPos_X, focusPoint.y()))
  {
    idBuffer.setPixel(currentPos_X, focusPoint.y(), fillColorIndex);
    numOfAvailablePixel++;

    _tmpBuffer.setPixel(currentPos_X, focusPoint.y(), COL_TMP_FILL_INDEX);

    // In the previous and next scan line new focus points are searched
    if (idBuffer.valid(currentPos_X, focusPoint.y()-1))
    {
      if (idBuffer.pixelIndex(currentPos_X, focusPoint.y()-1) == backgroundColorIndex)
      {
        putOnStack = FALSE;

        if (idBuffer.valid(currentPos_X+1, focusPoint.y()-1))
        {
          neighborColIndex1 = idBuffer.pixelIndex(currentPos_X+1, focusPoint.y()-1);
          if (neighborColIndex1 == boundaryIndex || neighborColIndex1 == fillColorIndex)
            putOnStack = TRUE;
        }
        if (idBuffer.valid(currentPos_X-1, focusPoint.y()-1))
        {
          neighborColIndex2 = idBuffer.pixelIndex(currentPos_X-1, focusPoint.y()-1);
          if (neighborColIndex2 == boundaryIndex || neighborColIndex2 == fillColorIndex)
            putOnStack = TRUE;
        }
        
        if (putOnStack)
          stackUp.push(QPoint(currentPos_X, focusPoint.y()-1));
      }
    }
      
    if (idBuffer.valid(currentPos_X, focusPoint.y()+1))
    {
      if (idBuffer.pixelIndex(currentPos_X, focusPoint.y()+1) == backgroundColorIndex)
      {
        putOnStack = FALSE;

        if (idBuffer.valid(currentPos_X+1, focusPoint.y()+1))
        {
          neighborColIndex1 = idBuffer.pixelIndex(currentPos_X+1, focusPoint.y()+1);
          if (neighborColIndex1 == boundaryIndex || neighborColIndex1 == fillColorIndex)
            putOnStack = TRUE;
        }
        if (idBuffer.valid(currentPos_X-1, focusPoint.y()+1))
        {
          neighborColIndex2 = idBuffer.pixelIndex(currentPos_X-1, focusPoint.y()+1);
          if (neighborColIndex2 == boundaryIndex || neighborColIndex2 == fillColorIndex)
            putOnStack = TRUE;
        }
        
        if (putOnStack)
          stackDown.push(QPoint(currentPos_X, focusPoint.y()+1));
      }
    }

    currentPos_X++;
    if (idBuffer.valid(currentPos_X, focusPoint.y()))
      currentColIndex = idBuffer.pixelIndex(currentPos_X, focusPoint.y());
  }

  // Go to the left
  currentPos_X = focusPoint.x() - 1;
  if (idBuffer.valid(currentPos_X, focusPoint.y()))
    currentColIndex = idBuffer.pixelIndex(currentPos_X, focusPoint.y());

  while (currentColIndex != boundaryIndex && currentColIndex != fillColorIndex &&
    idBuffer.valid(currentPos_X, focusPoint.y()))
  {
    idBuffer.setPixel(currentPos_X, focusPoint.y(), fillColorIndex);
    numOfAvailablePixel++;

    _tmpBuffer.setPixel(currentPos_X, focusPoint.y(), COL_TMP_FILL_INDEX);

    // In the previous and next scan line new focus points are searched
    if (idBuffer.valid(currentPos_X, focusPoint.y()-1))
    {
      if (idBuffer.pixelIndex(currentPos_X, focusPoint.y()-1) == backgroundColorIndex)
      {
        putOnStack = FALSE;
        
        if (idBuffer.valid(currentPos_X+1, focusPoint.y()-1))
        {
          neighborColIndex1 = idBuffer.pixelIndex(currentPos_X+1, focusPoint.y()-1);
          if (neighborColIndex1 == boundaryIndex || neighborColIndex1 == fillColorIndex)
            putOnStack = TRUE;
        }
        if (idBuffer.valid(currentPos_X-1, focusPoint.y()-1))
        {
          neighborColIndex2 = idBuffer.pixelIndex(currentPos_X-1, focusPoint.y()-1);
          if (neighborColIndex2 == boundaryIndex || neighborColIndex2 == fillColorIndex)
            putOnStack = TRUE;
        }
        
        if (putOnStack)
          stackUp.push(QPoint(currentPos_X, focusPoint.y()-1));
      }
    }
      
    if (idBuffer.valid(currentPos_X, focusPoint.y()+1))
    {
      if (idBuffer.pixelIndex(currentPos_X, focusPoint.y()+1) == backgroundColorIndex)
      {
        putOnStack = FALSE;

        if (idBuffer.valid(currentPos_X+1, focusPoint.y()+1))
        {
          neighborColIndex1 = idBuffer.pixelIndex(currentPos_X+1, focusPoint.y()+1);
          if (neighborColIndex1 == boundaryIndex || neighborColIndex1 == fillColorIndex)
            putOnStack = TRUE;
        }
        if (idBuffer.valid(currentPos_X-1, focusPoint.y()+1))
        {
          neighborColIndex2 = idBuffer.pixelIndex(currentPos_X-1, focusPoint.y()+1);
          if (neighborColIndex2 == boundaryIndex || neighborColIndex2 == fillColorIndex)
            putOnStack = TRUE;
        }
        
        if (putOnStack)
          stackDown.push(QPoint(currentPos_X, focusPoint.y()+1));
      }
    }

    currentPos_X--;
    if (idBuffer.valid(currentPos_X, focusPoint.y()))
      currentColIndex = idBuffer.pixelIndex(currentPos_X, focusPoint.y());
  }
}

//  Get color information for an object ID/sub-ID
bool SmageIdBufferGenerator::getFillColorIndices(const int &id, const int &subId, int &fill, int &boundary)
{
  if (id >= 0)
  {
    map<unsigned int, vector<SmageIdColorStruct> >::const_iterator idColorMapIter = 
      _idColorMap.find((unsigned int)(id));
    if (idColorMapIter != _idColorMap.end())
    {
      vector<SmageIdColorStruct>::const_iterator idColorVecIter;
      for (idColorVecIter=idColorMapIter->second.begin(); 
           idColorVecIter!=idColorMapIter->second.end(); idColorVecIter++)
      {    
        if (int(idColorVecIter->subID) == subId)
        {
          fill = idColorVecIter->fillColorIndex;
          boundary = idColorVecIter->outlineColorIndex;
     
          return TRUE;
        }
      }
    }
  }

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("*** Warning (SmageIdBufferGenerator::getFillColorIndices): could not find color ***\n");
#endif

  return FALSE;
}

//  Get color information for an object ID/sub-ID
bool SmageIdBufferGenerator::getFillColor(const int &id, const int &subId, QColor &fill, QColor &boundary)
{
  if (id >= 0)
  {
    map<unsigned int, vector<SmageIdColorStruct> >::const_iterator idColorMapIter = 
      _idColorMap.find((unsigned int)(id));
    if (idColorMapIter != _idColorMap.end())
    {
      vector<SmageIdColorStruct>::const_iterator idColorVecIter;
      for (idColorVecIter=idColorMapIter->second.begin(); 
           idColorVecIter!=idColorMapIter->second.end(); idColorVecIter++)
      {
        if (int(idColorVecIter->subID) == subId)
        {
          fill = QColor(idColorVecIter->fillColor);
          boundary = QColor(idColorVecIter->outlineColor);

          return TRUE;
        }
      }
    }
  }

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("*** Warning (SmageIdBufferGenerator::getFillColor): could not find color ***\n");
#endif

  return FALSE;
}

//  Test if specified colors are associated with "id"
bool SmageIdBufferGenerator::isAmongFillIndices(const int &id, const int &colorIndex)
{
  if (id >= 0)
  {
    map<unsigned int, vector<SmageIdColorStruct> >::const_iterator idColorMapIter = 
      _idColorMap.find((unsigned int)(id));
    if (idColorMapIter != _idColorMap.end())
    {
      vector<SmageIdColorStruct>::const_iterator idColorVecIter;
      for (idColorVecIter=idColorMapIter->second.begin(); 
           idColorVecIter!=idColorMapIter->second.end(); idColorVecIter++)
      {
        if (idColorVecIter->fillColorIndex == colorIndex)
          return TRUE;
      }
    }
  }

  return FALSE;
}

bool SmageIdBufferGenerator::isAmongColors(const int &color, const vector<SmageIdColorStruct> &colorValues)
{
  vector<SmageIdColorStruct>::const_iterator idColorVecIter;
  for (idColorVecIter=colorValues.begin(); idColorVecIter!=colorValues.end(); idColorVecIter++)
    {
      if (idColorVecIter->fillColorIndex == color || idColorVecIter->outlineColorIndex == color)
	return TRUE;
    }

  return FALSE;
}

//   Test if specified colors are associated with "id"
bool SmageIdBufferGenerator::isAmongOutlineIndices(const int &id, const int &colorIndex)
{
  if (id >= 0)
  {
    map<unsigned int, vector<SmageIdColorStruct> >::const_iterator idColorMapIter = 
      _idColorMap.find((unsigned int)(id));
    if (idColorMapIter != _idColorMap.end())
    {
      vector<SmageIdColorStruct>::const_iterator idColorVecIter;
      for (idColorVecIter=idColorMapIter->second.begin(); 
           idColorVecIter!=idColorMapIter->second.end(); idColorVecIter++)
      {
        if (idColorVecIter->outlineColorIndex == colorIndex)
          return TRUE;
      }
    }
  }

  return FALSE;
}

//   Is specified color among all fill colors
bool SmageIdBufferGenerator::isAmongFillColors(const QRgb &color)
{
  map<unsigned int, vector<SmageIdColorStruct> >::const_iterator idColorMapIter;

  for (idColorMapIter=_idColorMap.begin(); idColorMapIter!=_idColorMap.end(); idColorMapIter++)
  {
    vector<SmageIdColorStruct>::const_iterator idColorVecIter;
    for (idColorVecIter=idColorMapIter->second.begin(); 
         idColorVecIter!=idColorMapIter->second.end(); idColorVecIter++)
    {
      if (areEqualColors(idColorVecIter->fillColor, color))
        return TRUE;
    }
  }

  return FALSE;
}

//  Get the color table index for the specified color or return -1 if not included
int SmageIdBufferGenerator::getColorTableIndex(const QRgb &color)
{
  QRgb *colors = _idBuffer.colorTable();
  for (int i=0; i<_idBuffer.numColors(); i++)
  {
    if (areEqualColors(colors[i], color))
    return i;
  }

  return -1;
}

//   Get object ID/sub-ID associated with the specified color
bool SmageIdBufferGenerator::getObjectId(const QColor &color, int &id, int &subId)
{
  id = subId = -1;

  map<unsigned int, vector<SmageIdColorStruct> >::const_iterator idColorMapIter;
  vector<SmageIdColorStruct>::const_iterator idColorVecIter;
  
  for (idColorMapIter=_idColorMap.begin(); idColorMapIter!=_idColorMap.end(); idColorMapIter++)
  {
    for (idColorVecIter=idColorMapIter->second.begin(); 
         idColorVecIter!=idColorMapIter->second.end(); idColorVecIter++)
    {
      if (areEqualColors(idColorVecIter->fillColor, color) || 
          areEqualColors(idColorVecIter->outlineColor, color))
      {
        id = idColorMapIter->first;
        subId = int(idColorVecIter->subID);

        return TRUE;
      }
    }
  }

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("*** Warning (SmageIdBufferGenerator::getObjectId): could not find object ID ***\n");
#endif

  return FALSE;
}

//  Set colors associated with specified "id" in "_idColorMap"
void SmageIdBufferGenerator::setIdColors(const int &id, const int &subId,
           const int &fillIndex, const QRgb &fillColor,
           const int &outlineIndex, const QRgb &outlineColor)
{
  map<unsigned int, vector<SmageIdColorStruct> >::iterator idColorMapIter = _idColorMap.find((unsigned int)(id));
  if (idColorMapIter != _idColorMap.end())
  {
    vector<SmageIdColorStruct>::iterator idColorVecIter;
    for (idColorVecIter=idColorMapIter->second.begin();
         idColorVecIter!=idColorMapIter->second.end(); idColorVecIter++)
    {
      if (int(idColorVecIter->subID) == subId)
      {
        idColorVecIter->fillColor = fillColor;
        idColorVecIter->outlineColor = outlineColor;
        idColorVecIter->fillColorIndex = fillIndex;
        idColorVecIter->outlineColorIndex = outlineIndex;
      }

      break;
    }
  }
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageIdBufferGenerator::setIdColors): could not find ID in map ***\n");
#endif
  }
}

//  Set ID-Buffer image data
void SmageIdBufferGenerator::setIdBuffer(const QImage &idBuffer, const QColor &backgroundColor,
           const IdBufferType &bufferType)
{
  // If "idBuffer" is an RGB image, the class intern ID-Buffer must be generated from "idBuffer";
  // ID-Buffer's color table is filled with the colors in "idBuffer"

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBufferGenerator::setIdBuffer()\n");
#endif

  // If "idBuffer" is already an indexed image, its color table must be of the defined structure
  if (idBuffer.depth() == 8)
  {
    _idBuffer = idBuffer.copy();

    // still TODO
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageIdBufferGenerator::setIdBuffer): set ID-Buffer is already an indexed image ... \"_idColorMap\" still need to be filled ... not yet implemented ***\n");
#endif
  }

  else if (idBuffer.depth() > 8)
  {
    if (bufferType == SMAGE_OUTLINED_REGIONS)
      setOutlinedIdBuffer(idBuffer, backgroundColor);

    else
    {
      QColor curColor;
      int currentFillIndex;
    
      vector<QColor> colorValues;
      vector<QColor> refusedColors;
      colorValues.clear();
      refusedColors.clear();

      for (int y=0; y<idBuffer.height(); y++)
        for (int x=0; x<idBuffer.width(); x++)
        {
          curColor = QColor(idBuffer.pixel(x, y));
          currentFillIndex = -1;

          if (!areEqualColors(curColor, backgroundColor))
          {
            if (!findValue(refusedColors, curColor))      
            {
              // Is the current color already included in the ID-Buffer's color table
              for (unsigned int i=0; i<colorValues.size(); i++)
                if (areEqualColors(colorValues[i], curColor))
                {
                  currentFillIndex = i; break;
                }
    
              if (currentFillIndex != -1)
                currentFillIndex = 2*currentFillIndex+NUM_DEFAULT_TABLE_ENTRIES;
              else
              {
                // Color is not yet included in ID-Buffer's color table
                // Get object IDs according to current color
                int objId, subObjId;
                if (getObjectId(curColor, objId, subObjId))
                {
                  colorValues.push_back(curColor);
                  // Append the two new colors (fill and outline color) to ID-Buffer's
                  // color table;
                  // outline color or fill color has to be generated ("tmpColor") 
                  // depending on ID-Buffer type
                  _idBuffer.setNumColors(_idBuffer.numColors()+2);
                  QColor tmpColor;
                  if (curColor.blue() > 0)
                    tmpColor = QColor(curColor.red(), curColor.green(), curColor.blue()-1);
                  else
                    tmpColor = QColor(curColor.red(), curColor.green(), curColor.blue()+1);
        
                  currentFillIndex = _idBuffer.numColors()-2;
        
                  if (bufferType == SMAGE_OUTLINES)
                  {
                    // "tmpColor" is fill color
                    _idBuffer.setColor(_idBuffer.numColors()-2, tmpColor.rgb());
                    _idBuffer.setColor(_idBuffer.numColors()-1, curColor.rgb());
            
                    setIdColors(objId, subObjId, currentFillIndex, tmpColor.rgb(), 
                    currentFillIndex+1, curColor.rgb());
                  }
                  else
                  {
                    // "tmpColor" is outline color
                    _idBuffer.setColor(_idBuffer.numColors()-2, curColor.rgb());
                    _idBuffer.setColor(_idBuffer.numColors()-1, tmpColor.rgb());
        
                    setIdColors(objId, subObjId, currentFillIndex, curColor.rgb(), 
                    currentFillIndex+1, tmpColor.rgb());
                  }
                }
                else
                {
                  currentFillIndex = COL_BACKGROUND_INDEX;
                  refusedColors.push_back(curColor);
                }
              }

              if (bufferType == SMAGE_OUTLINES && currentFillIndex != COL_BACKGROUND_INDEX)
                currentFillIndex++;
              _idBuffer.setPixel(x, y, currentFillIndex);
            }
          }
        }
    }
  }  
}

//  Set an ID-Buffer that has already outlined regions
void SmageIdBufferGenerator::setOutlinedIdBuffer(const QImage &idBuffer, const QColor &backgroundColor)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageIdBufferGenerator::setOutlinedIdBuffer()\n");
#endif

  // As soon as a color differs from the background color and is not yet included in
  // the color table, it is assumed that this color must be an outline color;
  // when an outline color has been found, its associated fill color is searched;
  //  => not until both colors are found, they are inserted into the color table

  QColor curColor;
  int currentFillIndex;
        
  for (int y=0; y<idBuffer.height(); y++)
    for (int x=0; x<idBuffer.width(); x++)
    {
      curColor = QColor(idBuffer.pixel(x, y));
      currentFillIndex = -1;

      if (!areEqualColors(curColor, backgroundColor))
      {
        currentFillIndex = getColorTableIndex(curColor.rgb());

        if (currentFillIndex == -1)
        {
          // "curColor" must be an outline color
          // Search in current pixel's neighborhood for the associated fill color
          QColor accordingFillColor;
          if (findAccordingFillColor(idBuffer, QPoint(x,y), accordingFillColor))
          {
            // Append the two colors to the ID-Buffer's color table

            int objId, subObjId;
            getObjectId(accordingFillColor, objId, subObjId);

            _idBuffer.setNumColors(_idBuffer.numColors()+2);
        
            currentFillIndex = _idBuffer.numColors()-1;
            _idBuffer.setColor(currentFillIndex-1, accordingFillColor.rgb());
            _idBuffer.setColor(currentFillIndex, curColor.rgb());
        
            // Insert colors associated with "objId" into "_idColorMap"
            setIdColors(objId, subObjId, currentFillIndex-1, accordingFillColor.rgb(), 
            currentFillIndex, curColor.rgb());
          }
        }
        
        if (currentFillIndex != -1)
          _idBuffer.setPixel(x, y, currentFillIndex);
      }
    }  
}

//  When an outline color has been found, the outline pixel's neighboring pixels are
//  analyzed regarding their color => is their color the associated fill color?
//  Return if fill color has been found.
bool SmageIdBufferGenerator::findAccordingFillColor(const QImage &idBuffer, const QPoint &pos, QColor &fillColor)
{
  if (idBuffer.valid(pos.x()-1, pos.y()))
    if (isAmongFillColors(idBuffer.pixel(pos.x()-1, pos.y())))
    {
      fillColor = QColor(idBuffer.pixel(pos.x()-1, pos.y()));
      return TRUE;
    }
  if (idBuffer.valid(pos.x()+1, pos.y()))
    if (isAmongFillColors(idBuffer.pixel(pos.x()+1, pos.y())))
    {
      fillColor = QColor(idBuffer.pixel(pos.x()+1, pos.y()));
      return TRUE;
    }
  if (idBuffer.valid(pos.x(), pos.y()-1))
    if (isAmongFillColors(idBuffer.pixel(pos.x(), pos.y()-1)))
    {
      fillColor = QColor(idBuffer.pixel(pos.x(), pos.y()-1));
      return TRUE;
    }
  if (idBuffer.valid(pos.x(), pos.y()+1))
    if (isAmongFillColors(idBuffer.pixel(pos.x(), pos.y()+1)))
    {
      fillColor = QColor(idBuffer.pixel(pos.x(), pos.y()+1));
      return TRUE;
    }

  return FALSE;
}

//   Insert a vector of color information associated with "id" into "_idColorMap"
void SmageIdBufferGenerator::insertIdColorsValue(const unsigned int &id, const vector<SmageIdColorStruct> &colors)
{
  map<unsigned int, vector<SmageIdColorStruct> >::iterator idColorMapIter = _idColorMap.find(id);
  if (idColorMapIter != _idColorMap.end())
  {
    vector<SmageIdColorStruct>::const_iterator colorsIter;
    for (colorsIter=colors.begin(); colorsIter!=colors.end(); colorsIter++)
      idColorMapIter->second.push_back(*colorsIter);
  }
  else
  {
    pair<unsigned int, vector<SmageIdColorStruct> > p(id, colors);
    _idColorMap.insert(p);
  }
}

//   Insert a single color information associated with "id" into "_idColorMap"
void SmageIdBufferGenerator::insertIdColorsValue(const unsigned int &id, const SmageIdColorStruct &color)
{  
  map<unsigned int, vector<SmageIdColorStruct> >::iterator idColorMapIter = _idColorMap.find(id);
  
  if (idColorMapIter != _idColorMap.end())
    idColorMapIter->second.push_back(color);
  else
  {
    vector<SmageIdColorStruct> idColorVec;
    idColorVec.push_back(color);
    _idColorMap[id] = idColorVec;
  }
}
