#ifndef  _SMAGE_ID_BUFFER_GENERATOR_H
#define  _SMAGE_ID_BUFFER_GENERATOR_H

#include <qimage.h>

#include <stack>
#include <vector>
#include <map>

/** When a new ID-Buffer is generated, two colors are assigned by default
  * (a background and a fill color)
  */
#define NUM_DEFAULT_TABLE_ENTRIES 2

/** Predefined colors and their indices
  */
#define COL_BACKGROUND         255,255,255
#define COL_BOUNDARY           0,255,210
#define COL_BOUNDARY_DEFAULT   0,0,0
#define COL_FILL_DEFAULT       255,0,0

#define COL_BACKGROUND_INDEX        0
#define COL_BOUNDARY_INDEX          1
#define COL_FILL_DEFAULT_INDEX      NUM_DEFAULT_TABLE_ENTRIES
#define COL_BOUNDARY_DEFAULT_INDEX  (NUM_DEFAULT_TABLE_ENTRIES + 1)

#define COL_TMP_FILL_INDEX 1

#define DEFAULT_MAP_POS 0

static const int pot2[8] = {1,2,4,8,16,32,64,128};

/** Different types of externally generated ID-Buffers.
  *
  * @ingroup LibSmage
  *
  */
enum IdBufferType { 
  SMAGE_OUTLINES, 
  SMAGE_FILLED_REGIONS, 
  SMAGE_OUTLINED_REGIONS
};

/** Structure that holds necessary data concerning an ID-Buffer.
  *
  * @ingroup LibSmage
  *
  */
struct SmageIdBufferStruct {
  SmageIdBufferStruct();     // Constructor
  QImage idBuffer;           // The ID-Buffer (image)
  QColor backgroundColor;    // ID-Buffer's background color
  IdBufferType bufferType;   // Type of ID-Buffer
};

/** Structure that holds color information associated with a specific object ID
  *
  * @ingroup LibSmage
  *
  */
struct SmageIdColorStruct {
  SmageIdColorStruct();     // Constructor
  char subID;               // Sub-ID (a region may be composed of a number of sub-regions)
  QRgb fillColor;           // Color a region is filled with
  QRgb outlineColor;        // Color a region is outlined with
  int fillColorIndex;       // The according color indices
  int outlineColorIndex;
};

/** Class SmageIdBufferGenerator for basic ID-Buffer image operations (eg, fill a region).
  *
  * @ingroup LibSmage
  *
  * An ID-Buffer is an image that indicates, which regions in 
  *  the actual image shall be processed. It is used as a reference
  *  image<br>
  *
  *  Currently, the ID-Buffer has 2 default entries in the color 
  *  table (background and fill color);
  *  all pairs of following color table entries refer to regions
  *  in the buffer:<br>
  *  - first value  -> fill color of that region<br>
  *  - second value -> outline color of that region<br>
  *  <br>
  *  While the ID-Buffer set to this class may be a RGB or an indexed
  *  image, the class intern ID-Buffer is always an indexed image.
  *  The regions in the externally generated ID-Buffer may be filled,
  *  outlined, or already filled and outlined.
  *
  */
class SmageIdBufferGenerator
{
  public:
    /** Constructor.
      * (argument is original image's size)
      */ 
    SmageIdBufferGenerator(const int &width, const int &height);

    /** Destructor.
      */
    ~SmageIdBufferGenerator();

    /** Get color information for an object ID/sub-ID.
      */
    bool getFillColorIndices(const int &id, const int &subId, int &fill, int &boundary);
    bool getFillColor(const int &id, const int &subId, QColor &fill, QColor &boundary);

    /** Test if specified colors are associated with "id"
      */
    bool isAmongFillIndices(const int &id, const int &colorIndex);
    bool isAmongOutlineIndices(const int &id, const int &colorIndex);

    /** Is specified color among all fill colors.
      */
    bool isAmongFillColors(const QRgb &color);

    static bool isAmongColors(const int &color, const std::vector<SmageIdColorStruct> &colorValues);

    /** Get object ID/sub-ID associated with the specified color.
      */
    bool getObjectId(const QColor &color, int &id, int &subId);

    /** Set and get ID-Buffer (image)
      */
    void getIdBuffer(QImage &idBuffer) { idBuffer = _idBuffer; }
    void setIdBuffer(const QImage &idBuffer, const QColor &backgroundColor, const IdBufferType &bufferType);

    /** Insert the colors associated with an object-ID into a map "_idColorMap"
      */
    void insertIdColorsValue(const unsigned int &id, const std::vector<SmageIdColorStruct> &colors);
    void insertIdColorsValue(const unsigned int &id, const SmageIdColorStruct &color);

    static bool areEqualColors(const QRgb &, const QRgb &);
    static bool areEqualColors(const QColor &, const QColor &);
    static bool areEqualColors(const QRgb &, const QColor &);
      
    /** Is the number even
     */
    static bool isEven(const int &number);

  protected:
      
    /** Initialize the ID-Buffer and fill it with "COL_BACKGROUND"
     */
    void initializeIdBuffer();

    /** Set an ID-Buffer that has already outlined regions
      */
    void setOutlinedIdBuffer(const QImage &idBuffer, const QColor &backgroundColor);
    bool findAccordingFillColor(const QImage &idBuffer, const QPoint &pos, QColor &fillColor);

    /** Is specified "value" included in the vector
      */
    bool findValue(const std::vector<int> &values, const int &value);
    bool findValue(const std::vector<QColor> &values, const QColor &value);
  
    /** Is the specified point an upper or a lower peak in the outline
      */
    bool isVerticalPeak(const QPoint &point, const int &colIndex);

    /** Get the color table index for the specified color or return -1 if not included
      */
    int getColorTableIndex(const QRgb &color);

    /** Fill the region within an outline
      */
    unsigned long fillRegion(QImage &idBuffer, const QPoint &seedPointPos, const int &backgroundColorIndex, 
             const int &boundaryIndex, const int &fillColorIndex);
    void updateStacks(QImage &idBuffer, const QPoint &focusPoint, const int &backgroundColorIndex, 
    const int &fillColorIndex, const int &boundaryIndex,
    std::stack<QPoint> &stackUp, std::stack<QPoint> &stackDown, unsigned long &numOfAvailablePixel);

    /** Revise outlines; are there outline pixels that are not adjacent to filled color pixels
      */
    void reviseOutlines(const std::vector<int> &seedPointColorIndices,        
        const int &backgroundColorIndex, const int &boundaryColorIndex, 
        std::map<unsigned int, unsigned long> &outlinePixelMap);
    void reviseOutlines(const int &backgroundColorIndex, std::map<unsigned int, unsigned long> &outlinePixelMap);

    /** Set colors associated with specified "id" in "_idColorMap"
      */
    void setIdColors(const int &id, const int &subId, const int &fillIndex, const QRgb &fillColor,
           const int &outlineIndex, const QRgb &outlineColor);

    /** Temporary and actual ID-Buffer images
      */
    QImage _tmpBuffer;      
    QImage _idBuffer;      

    /** Map that holds for each object-ID the associated colors
      */
    std::map<unsigned int, std::vector<SmageIdColorStruct> > _idColorMap;
};

#endif
