#include "SmageImgCoder.h"
#include "SmageSettings.h"
#include "SmageImgProcessor.h"

#include <iostream>
#include <math.h>

#include <qbitmap.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qpen.h>

using namespace std;

// Constructor
SmageDataObjs::SmageDataObjs()
{
  data.clear();
  codeByteStream.clear();
  colorTable.clear();
  codePattern = SMAGE_AUTO;
}

// Constructor; image data is set;
SmageImgCoder::SmageImgCoder(const QImage &imgData)
{ 
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgCoder::SmageImgCoder()\n");
#endif

  _imgData = imgData;

  if (imgData.depth() >= 24) _isRGBImg = TRUE;
  else _isRGBImg = FALSE;

  _imgWidth = imgData.width();
  _imgHeight = imgData.height();

  _idBuffer = NULL;  
  _dataProcessor = NULL;

  _dataMap.clear();
}

// Destructor
SmageImgCoder::~SmageImgCoder()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgCoder::~SmageImgCoder()\n");
#endif

  if (_idBuffer != NULL)
    delete _idBuffer;

  if (_dataProcessor != NULL)
    delete _dataProcessor;
}

//  Initializing data that has not been initialized in the constructor
void SmageImgCoder::initialize(const bool &noIdBufferAvailable)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("\"SmageIdBuffer\" and \"SmageDataProcessor\" are initialized\n");
#endif

  _idBuffer = new SmageIdBuffer(_imgWidth, _imgHeight, noIdBufferAvailable);  
  _dataProcessor = new SmageDataProcessor;
}

// Returning a new "SmageDataObjs" structur with a generated "codeByteStream",
//  the byte stream to be encoded
SmageDataObjs SmageImgCoder::getPreparedData(const SmageDataObjs &data)
{
#ifdef SMAGE_PRINT_DEBUG
  if (_dataProcessor == NULL)
    _SmageLogFile.lprintf("*** Warning (SmageImgCoder::getPreparedData): \"_dataProcessor\" not initialized ***\n");
#endif

  SmageDataObjs preparedData;

  if (_dataProcessor != NULL)
  {
    SmagePropsStruct props;
    props.props = SMAGE_LINK;
    props.link.push_back('H');
    props.link.push_back('i');

    //_dataProcessor->setProps(props);
    _dataProcessor->getPreparedDataStream(data.data, preparedData.codeByteStream);
  }

  preparedData.data = data.data;
  preparedData.colorTable = data.colorTable;
  preparedData.codePattern = data.codePattern;
  
  return preparedData;
}

//  Converting a vector of QRgbs to a character vector
void SmageImgCoder::convertRgbToCharVec(const vector<QRgb> &rgb, vector<char> &charVec)
{
  charVec.clear();
  
  for (unsigned int i=0; i<rgb.size(); i++)
  {
    charVec.push_back((char)(rgb[i] & 0xff));          // Blue
    charVec.push_back((char)((rgb[i] >> 8) & 0xff));   // Green
    charVec.push_back((char)((rgb[i] >> 16) & 0xff));  // Red
  }
}

// Converting a vector of characters to a QRgb vector
void SmageImgCoder::convertCharToRgbVec(const vector<char> &charVec, vector<QRgb> &rgb)
{
  rgb.clear();
  
  for (unsigned int i=0; i<charVec.size(); i+=3)
  {
    if ((i+2) < charVec.size())
    {
      QRgb currentRgbValue = 0;

      currentRgbValue |= ((charVec[i+2] & 0xff) << 16); // Red
      currentRgbValue |= ((charVec[i+1] & 0xff) << 8);  // Green
      currentRgbValue |= (charVec[i] & 0xff);           // Blue

      rgb.push_back(currentRgbValue);
    }
  }
}

// Converting a field of characters to a QRgb vector
void SmageImgCoder::convertCharToRgbVec(const char *charField, const unsigned long &size, vector<QRgb> &rgb)
{
  if (charField == NULL)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgCoder::convertCharToRgbVec): \"charField\" is empty ***\n");
#endif
      return;
  }

  rgb.clear();
  const char *fieldPointer = charField;

  for (unsigned int i=0; i<size; i+=3)
  {
    if ((i+2) < size)
    {
      QRgb currentRgbValue = 0;

      currentRgbValue |= ((*fieldPointer) & 0xff);         fieldPointer++;   // Blue
      currentRgbValue |= (((*fieldPointer) & 0xff) << 8);  fieldPointer++;   // Green
      currentRgbValue |= (((*fieldPointer) & 0xff) << 16); fieldPointer++;   // Red

      rgb.push_back(currentRgbValue);
    }
  }
}

// Calculating a suitable code pattern according to the number of available pixels and the amount of data
//  to be encoded
bool SmageImgCoder::getSuitableCodePattern(const unsigned long &numOfPixels, const unsigned long &amountOfEncodableData,
             CodePattern &newCodePattern)
{
  if (numOfPixels <= 0)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgCoder::convertCharToRgbVec): \"charField\" is empty ***\n");
#endif
      return FALSE;
  }

  int bitsPerPixel = int(ceil((double(amountOfEncodableData)*8.0) / double(numOfPixels)));

#ifdef SMAGE_PRINT_INFOS
  _SmageLogFile.lprintf("SmageImgCoder::getSuitableCodePattern: data amount (bytes): %lu --- number of available pixels: %lu --- resulting bits per pixel: %d\n", amountOfEncodableData, numOfPixels, bitsPerPixel);
#endif

  // Assign code pattern according to "bitsPerPixel"
  if (bitsPerPixel > SMAGE_MAX_BITS_PER_PIXEL)
    return FALSE;
  else if (bitsPerPixel < 3)
    newCodePattern = SMAGE_LSB;
  else if (bitsPerPixel == 3)
    newCodePattern = SMAGE_BITS_3;
  else if (bitsPerPixel == 4)
    newCodePattern = SMAGE_BITS_4;
  else if (bitsPerPixel == 5)
    newCodePattern = SMAGE_LS_2B;
  else if (bitsPerPixel == 6)
    newCodePattern = SMAGE_BITS_6;
  else if (bitsPerPixel == 7)
    newCodePattern = SMAGE_BITS_7;
  else if (bitsPerPixel == 8)
    newCodePattern = SMAGE_LS_3B;
  else if (bitsPerPixel == 9)
    newCodePattern = SMAGE_BITS_9;
  else if (bitsPerPixel == 10)
    newCodePattern = SMAGE_BITS_10;
  else if (bitsPerPixel == 11)
    newCodePattern = SMAGE_LS_4B;
  else if (bitsPerPixel == 12)
    newCodePattern = SMAGE_BITS_12;
  else if (bitsPerPixel == 13)
    newCodePattern = SMAGE_BITS_13;
  else if (bitsPerPixel == 14)
    newCodePattern = SMAGE_LS_5B;
  else if (bitsPerPixel > 14 && bitsPerPixel <= SMAGE_MAX_BITS_PER_PIXEL)
    newCodePattern = SMAGE_LS_8B;
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgCoder::getSuitableCodePattern): amount of data too big ***\n");
#endif
  }

  return TRUE;
}


void SmageImgCoder::getOutlinesInImg(QPixmap &aPixmap)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgCoder::getOutlinesInImg()\n");
#endif

  if (aPixmap.isNull())
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgCoder::getOutlinesInImg): pixmap is empty ***\n");
#endif
    return;
  }

  if (aPixmap.width() != _imgData.width() || aPixmap.height() != _imgData.height())
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.
  lprintf("*** Warning (SmageImgCoder::getOutlinesInImg): pixmap and coded image are of different sizes ***\n");
#endif
    return;
  }
  
  QImage tmpImg(aPixmap.width(), aPixmap.height(), 32);

  SmageImgProcessor imgProcessor;
  imgProcessor.fillImgRandomized(tmpImg);

  QPainter painter;
  QBitmap bitmap(aPixmap.width(), aPixmap.height(), TRUE);
  QPen aPen(Qt::color1, 2);

  for (int y=0; y<bitmap.height(); y++)
    for (int x=0; x<bitmap.width(); x++)
    {
      int blue = qBlue(_imgData.pixel(x,y));

      if (blue & 1)
      {
        painter.begin(&bitmap);  
        painter.setPen(aPen);
        painter.drawRect(x,y,2,2);
        painter.end();
      }
    }

  if (aPixmap.convertFromImage(tmpImg))
    aPixmap.setMask(bitmap);
}
