#ifndef  _SMAGE_IMG_CODER_H
#define  _SMAGE_IMG_CODER_H

#include "SmageIdBuffer.h"
#include "SmageDataProcessor.h"

class QPixmap;

/** Structure that holds the information to be encoded or that has been 
  * decoded (for encoding only "data", "colorTable" and "codePattern" may be set;
  * "codeByteStream" is generated from "data" later on)
  *
  * @ingroup LibSmage
  *
  */
struct SmageDataObjs {
  SmageDataObjs();                   // Constructor
  std::vector<SmageData> data;       // Array of data items ("SmageData" -> "SmageDataProcessor.h")
  std::vector<char> codeByteStream;  // Compressed data stream (data to be encoded or that has been decoded)
  std::vector<QRgb> colorTable;      // Color table if any item of "data" is of type "SMAGE_IMG"
  CodePattern codePattern;           // How many bits per pixel ("CodePattern" -> "SmageDataProcessor.h")
};

/** Number of pixels that are reserved to store the amount of data to be coded.
  */
#define SMAGE_NUM_AMOUNT_OF_DATA_PIXELS 4

/** Number to be encoded if no color table is available.
  */
#define NO_COLOR_TABLE_CODE 1000

/** Class SmageImgCoder is base class for all encoding and decoding classes.
  *
  * @ingroup LibSmage
  *
  * Class is initialized with an image into which information shall 
  * be encoded or which already holds encoded information.
  */
class SmageImgCoder
{
  public:
    /** Constructor.
      * argument is the image into which data shall be encoded or which holds data.
      */
    SmageImgCoder(const QImage &imgData);

  protected:
      
    /** Destructor.
      */
    ~SmageImgCoder();

    /** Initialize "_idBuffer" and "_dataProcessor"
      */
    void initialize(const bool &noIdBufferAvailable);

    /** Preparing data for encoding.
      */
    SmageDataObjs getPreparedData(const SmageDataObjs &data);

    /** Conversions between characters and RGB values.
      */
    void convertRgbToCharVec(const std::vector<QRgb> &rgb, std::vector<char> &charVec);
    void convertCharToRgbVec(const std::vector<char> &charVec, std::vector<QRgb> &rgb);
    void convertCharToRgbVec(const char *charField, const unsigned long &size, std::vector<QRgb> &rgb);

    /**
      * If no specific code pattern is specified, a suitable code pattern is determined (according to 
      * the number of available pixels "numOfPixels" and the amount of data to be encoded "amountOfEncodableData")
      */
    bool getSuitableCodePattern(const unsigned long &numOfPixels, const unsigned long &amountOfEncodableData,
          CodePattern &newCodePattern);

    /** Data regarding the image that will be processed.
      */
    QImage _imgData;
    int _imgWidth, _imgHeight;
    bool _isRGBImg;

    /** ID-Buffer data.
      */
    SmageIdBuffer *_idBuffer;

    /** Data processor for generating the compressed data stream ("codeByteStream") in "SmageDataObjs".
      */
    SmageDataProcessor *_dataProcessor;

    /** 4 pixels in which the amount of data (number) is coded.
      */
    QPoint _amountOfDataPixels[4];      

    /** Map of object IDs and the according data.
      */
    std::map<unsigned int, SmageDataObjs> _dataMap;

    /** Get an image with outlines marked through a bitmap.
      */
    void getOutlinesInImg(QPixmap &aPixmap);
};

#endif
