#include "SmageImgLSBsDecoder.h"
#include "SmageSettings.h"

#include <iostream>

using namespace std;

// Constructor;
SmageImgLSBsDecoder::SmageImgLSBsDecoder(const QImage &imgData) : SmageImgCoder(imgData)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::SmageImgLSBsDecoder()\n");

  if (!_isRGBImg)
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::SmageImgLSBsDecoder): image is not a RGB image ***\n");
#endif
}

//   Destructor
SmageImgLSBsDecoder::~SmageImgLSBsDecoder()
{
}

//  Decode data from the image and fill "idBuffer" and "codeDataMap"
void SmageImgLSBsDecoder::decodeData(QImage &idBuffer, std::map<unsigned int, SmageDataObjs> &codedDataMap)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::decodeData()\n");
#endif

  if (_isRGBImg)
  {
    // Initialize "_idBuffer" and "_dataProcessor"
    initialize(FALSE);
      
    // Search for seed points and decode the information for each seed point
    scanForSeedPoints();
    decodeInformation();
      
    // Assign the data
    _idBuffer->getIdBuffer(idBuffer);
    codedDataMap = _dataMap;

    idBuffer.save("decodeIdBuffer.png", "PNG", 100);
  }
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::decodeData): image is not a RGB image ***\n");
#endif
  }
}

// Decode data from the image and fill "idBuffer", "codeDataMap" and
//  the "outlinePixmap" that will hold the outlines
void SmageImgLSBsDecoder::decodeData(QImage &idBuffer, std::map<unsigned int, SmageDataObjs> &codedDataMap,
             QPixmap &outlinePixmap)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::decodeData()\n");
#endif

  if (_isRGBImg)
  {
    // Initialize "_idBuffer" and "_dataProcessor"
    initialize(FALSE);

    // Search for seed points and decode the information for each seed point
    scanForSeedPoints();
    decodeInformation();
      
    // Assign the data
    _idBuffer->getIdBuffer(idBuffer);
    codedDataMap = _dataMap;

    // Generate outline pixmap
    outlinePixmap = QPixmap(_imgData.width(), _imgData.height());
    getOutlinesInImg(outlinePixmap);
  }
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::decodeData): image is not a RGB image ***\n");
#endif
  }
}

// Decode data from the image using specific method of image downgrading
//  (data must have been encoded using this method)
void SmageImgLSBsDecoder::decodeImgDownGrading(const char *fileName)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::decodeImgDownGrading()\n");
#endif

  // Based on the description in: 
  // C. Kurak and J. McHugh "A Cautionary Note on Image Downgrading"
  // "... replacing the low order bits of one image with the high order bits of another."

  if (_isRGBImg)
  {
    // Initialize "_idBuffer" and "_dataProcessor"
    initialize(FALSE);
  
    QImage encodedImg(_imgWidth, _imgHeight, 32);
      
    QRgb curCoverColor;
    int curCoverCol_Blue, curCoverCol_Green, curCoverCol_Red;
    int curCodeCol_Blue, curCodeCol_Green, curCodeCol_Red; 
  
    for (int y=0; y<_imgHeight; y++)
      for (int x=0; x<_imgWidth; x++)
      {
        curCodeCol_Blue = 0; curCodeCol_Green = 0; curCodeCol_Red = 0;
        curCoverColor = _imgData.pixel(x, y);
      
        curCoverCol_Blue = qBlue(curCoverColor);
        curCoverCol_Red = qRed(curCoverColor);
        curCoverCol_Green = qGreen(curCoverColor);

        curCodeCol_Red |= ((curCoverCol_Red & 15) << 4);
        curCodeCol_Green |= ((curCoverCol_Green & 15) << 4);
        curCodeCol_Blue |= ((curCoverCol_Blue & 15) << 4);
      
        encodedImg.setPixel(x, y, qRgb(curCodeCol_Red, curCodeCol_Green, curCodeCol_Blue));
      }

    encodedImg.save(fileName, "PNG", 100);
  }
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::decodeData): image is not a RGB image ***\n");
#endif
  }
}

// Test if image holds encoded data;
//  as soon as a seed point has been found, the image is expected to hold
//  additional information;
bool SmageImgLSBsDecoder::isInformationEncoded()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::isInformationEncoded()\n");
#endif

  QRgb *curRgbValueLine;
  int currentBlue;

  for (int y=0; y<_imgHeight; y++)
  {
    curRgbValueLine = (QRgb *) _imgData.scanLine(y);

    for (int x=0; x<_imgWidth; x++)
    {
      currentBlue = qBlue(curRgbValueLine[x]);

      if (currentBlue & 1)
      {
        if (_imgData.valid(x,y-1) && _imgData.valid(x-1,y))
        {
          if (qBlue(_imgData.pixel(x,y-1)) & 1 &&
              qBlue(curRgbValueLine[x-1]) & 1)
          {
            if (isSeedPoint(QPoint(x,y)))
	      return TRUE;
	  }
	}
      }
    }
  }
  
  return FALSE;
}

// Erase data from the image (data corresponds to the areas in the "idBuffer" specified by "colorValues" or it is
// all the encoded data if "allData" is TRUE)
void SmageImgLSBsDecoder::eraseDataFromImage(const bool &allData, 
					     const QImage &idBuffer, const vector<SmageIdColorStruct> &colorValues)
{
  QRgb *curRgbValueLine;
  int currentBlue, x, y;
  QRgb newColor;
  int curIdBufferIndex = 0;

  // Two image scans are necessary since otherwise the seed points could not clearly be identified

  if (!allData)
    {
      if (idBuffer.width() != _imgData.width() || idBuffer.height() != _imgData.height() ||
	  colorValues.empty())
	return;
    }

  // (1) Erase corresponding seed points
  for (y=0; y<_imgHeight; y++)
    {
      curRgbValueLine = (QRgb *) _imgData.scanLine(y);
      
      for (x=0; x<_imgWidth; x++)
	{
	  currentBlue = qBlue(curRgbValueLine[x]);
	  if (!allData)
	    curIdBufferIndex = idBuffer.pixelIndex(x, y);

	  if (currentBlue & 1)
	    {
	      if (SmageIdBufferGenerator::isAmongColors(curIdBufferIndex, colorValues) || allData)
		{
		  if (_imgData.valid(x,y-1) && _imgData.valid(x-1,y))
		    {
		      if (qBlue(_imgData.pixel(x,y-1)) & 1 &&
			  qBlue(curRgbValueLine[x-1]) & 1)
			{
			  if (isSeedPoint(QPoint(x,y)))
			    eraseSeedPointRegion(QPoint(x,y));
			}
		    }
		}
	    }
	}
    }

  // (2) Erase corresponding outlines
  for (y=0; y<_imgHeight; y++)
    {
      curRgbValueLine = (QRgb *) _imgData.scanLine(y);
      
      for (x=0; x<_imgWidth; x++)
	{
	  currentBlue = qBlue(curRgbValueLine[x]);
	  if (!allData)
	    curIdBufferIndex = idBuffer.pixelIndex(x, y);

	  if (currentBlue & 1)
	    {
	      if (SmageIdBufferGenerator::isAmongColors(curIdBufferIndex, colorValues) || allData)
		{
		  currentBlue &= 254;

		  newColor = qRgb(qRed(curRgbValueLine[x]), qGreen(curRgbValueLine[x]), currentBlue);
		  _imgData.setPixel(x, y, newColor);
		}
	    }
	}
    }
}

// Adapt the seed point and its 4 surrounding pixels to other pixels in this region
void SmageImgLSBsDecoder::eraseSeedPointRegion(const QPoint &seedPoint)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::eraseSeedPointRegion()\n");
#endif

  if (_imgData.valid(seedPoint.x()-1, seedPoint.y()-1) && 
      _imgData.valid(seedPoint.x()+1, seedPoint.y()+1))
  {
    // Colors of seed point's neighboring pixels (different to direct neighbor pixels used for coded image
    // identification)

    QRgb leftUpColor = _imgData.pixel(seedPoint.x()-1, seedPoint.y()-1);
    QRgb rightUpColor = _imgData.pixel(seedPoint.x()+1, seedPoint.y()-1);
    QRgb leftLowColor = _imgData.pixel(seedPoint.x()-1, seedPoint.y()+1);
    QRgb rightLowColor = _imgData.pixel(seedPoint.x()+1, seedPoint.y()+1);

    QRgb newColor;

    newColor = qRgb(int((qRed(leftUpColor) + qRed(rightUpColor))/2),
                    int((qGreen(leftUpColor) + qGreen(rightUpColor))/2),
                    int((qBlue(leftUpColor) + qBlue(rightUpColor))/2));
    _imgData.setPixel(seedPoint.x(), seedPoint.y()-1, newColor);

    newColor = qRgb(int((qRed(leftUpColor) + qRed(leftLowColor))/2),
                    int((qGreen(leftUpColor) + qGreen(leftLowColor))/2),
                    int((qBlue(leftUpColor) + qBlue(leftLowColor))/2));
    _imgData.setPixel(seedPoint.x()-1, seedPoint.y(), newColor);

    newColor = qRgb(int((qRed(rightUpColor) + qRed(rightLowColor))/2),
                    int((qGreen(rightUpColor) + qGreen(rightLowColor))/2),
                    int((qBlue(rightUpColor) + qBlue(rightLowColor))/2));
    _imgData.setPixel(seedPoint.x()+1, seedPoint.y(), newColor);

    newColor = qRgb(int((qRed(leftLowColor) + qRed(rightLowColor))/2),
                    int((qGreen(leftLowColor) + qGreen(rightLowColor))/2),
                    int((qBlue(leftLowColor) + qBlue(rightLowColor))/2));
    _imgData.setPixel(seedPoint.x(), seedPoint.y()+1, newColor);

    // Now the seed point itself as the mean of its for indirect neighbors
    newColor = qRgb(int((qRed(leftUpColor) + qRed(rightUpColor) + qRed(rightLowColor) + qRed(leftLowColor))/4),
                    int((qGreen(leftUpColor) + qGreen(rightUpColor) + qGreen(rightLowColor) + qGreen(leftLowColor))/4),
                    int((qBlue(leftUpColor) + qBlue(rightUpColor) + qBlue(rightLowColor) + qBlue(leftLowColor))/4));
    _imgData.setPixel(seedPoint.x(), seedPoint.y(), newColor);
  }
}

// Search for seed points within the image
void SmageImgLSBsDecoder::scanForSeedPoints()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::scanForSeedPoints()\n");
#endif

  QRgb *curRgbValueLine;
  int currentBlue;
  bool foundSeedPoint;
  int numColors = NUM_DEFAULT_TABLE_ENTRIES;

  QImage idBuffer;
  _idBuffer->getIdBuffer(idBuffer);
  if (idBuffer.isNull())
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::scanForSeedPoints): invalid ID_Buffer ***\n");
#endif
    return;
  }

  // A map with an object ID and ID-Buffer region information is filled
  map<unsigned int, SmageIdRegion> regionMap;
  regionMap.clear();
  map<unsigned int, SmageIdRegion>::iterator regionMapIter;

  for (int y=0; y<_imgHeight; y++)
  {
    curRgbValueLine = (QRgb *) _imgData.scanLine(y);

    for (int x=0; x<_imgWidth; x++)
    {
      currentBlue = qBlue(curRgbValueLine[x]);

      foundSeedPoint = FALSE;

      if ((currentBlue & 1) && (idBuffer.pixelIndex(x,y) == COL_BACKGROUND_INDEX))
      {
        if (_imgData.valid(x,y-1) && _imgData.valid(x-1,y))
        {
          if (qBlue(_imgData.pixel(x,y-1)) & 1 &&
              qBlue(curRgbValueLine[x-1]) & 1)
          {
            // If a seed point has been found, its position is inserted into 
            // "regionMap" at the object ID which coded in the seed point
            if (isSeedPoint(QPoint(x,y)))
            {
              foundSeedPoint = TRUE;

              int objId, subObjId;
              decodeObjId(QPoint(x,y), objId);
	      subObjId = 0;

              SmageIdSubRegion idSubRegion;
              idSubRegion.seedPointPos = QPoint(x,y);

              regionMapIter = regionMap.find(objId);
              if (regionMapIter == regionMap.end())
              {
                SmageIdRegion idRegion;
                idRegion.subRegions.push_back(idSubRegion);

                regionMap[objId] = idRegion;
              }
              else
		{
		  subObjId = regionMapIter->second.subRegions.size();
		  regionMapIter->second.subRegions.push_back(idSubRegion);
		}

              idSubRegion.subID = char(subObjId);

              int fillIndex, outlineIndex;
              if (!_idBuffer->getFillColorIndices(objId, subObjId, fillIndex, outlineIndex))
              {
                // A fill color and an outline color are determined that
                // will be used to fill the regarding regions in the ID-Buffer 
                // later on
                QRgb fillColor = qRgb(255, subObjId, objId);
                QRgb outlineColor = qRgb(0, subObjId, objId);

                SmageIdColorStruct idColorStruct;
                idColorStruct.subID = char(subObjId);
                idColorStruct.fillColor = fillColor;
                idColorStruct.outlineColor = outlineColor;
            
                fillIndex = numColors;
                outlineIndex = numColors+1;
                idColorStruct.fillColorIndex = fillIndex;
                idColorStruct.outlineColorIndex = outlineIndex;

                // ID-Buffer class has a map with object IDs and their
                // associated colors in the ID-Buffer image;
                // here a new item is inserted into this map
                _idBuffer->insertIdColorsValue((unsigned int)(objId), idColorStruct);

                // The color table of the ID-Buffer is updated with the 
                // two new colors
                numColors += 2;
                idBuffer.setNumColors(numColors);
                idBuffer.setColor(fillIndex, fillColor);
                idBuffer.setColor(outlineIndex, outlineColor);
              }

              idBuffer.setPixel(x, y, fillIndex);
            }
          }
        }

        if (!foundSeedPoint)
          idBuffer.setPixel(x, y, COL_BOUNDARY_INDEX);
      }
    }
  }
  
  // "regionMap" is assigned to the ID-Buffer class
  _idBuffer->setRegionMap(regionMap);

  // Fill the outlined regions in the ID-Buffer that belong to each seed point;
  // those filled regions will be used during the decoding process
  _idBuffer->fillSeedPointRegions();
}

// Is the current point a seed point?
bool SmageImgLSBsDecoder::isSeedPoint(const QPoint &focusPoint)
{
  // A seed point's blue LSB is unset;
  // here the surrounding pixels are analyzed, the characters "IW" and "WI"
  // should be coded in those pixels

  int counter_IW = 0;
  int counter_WI = 0;
  int counter_OutlinePixels = 0;
  int counter_IW_OutlinePixels = 0;

  QPoint points_IW[2], points_WI[2], currentPixel;

  currentPixel = QPoint(focusPoint.x(), focusPoint.y()-1);
  if (_imgData.valid(currentPixel.x(), currentPixel.y()))
      analyzePixel(currentPixel, counter_IW, counter_WI, counter_OutlinePixels,
    points_IW, points_WI, counter_IW_OutlinePixels);

  currentPixel = QPoint(focusPoint.x()+1, focusPoint.y());
  if (_imgData.valid(currentPixel.x(), currentPixel.y()))
      analyzePixel(currentPixel, counter_IW, counter_WI, counter_OutlinePixels,
    points_IW, points_WI, counter_IW_OutlinePixels);

  currentPixel = QPoint(focusPoint.x(), focusPoint.y()+1);
  if (_imgData.valid(currentPixel.x(), currentPixel.y()))
      analyzePixel(currentPixel, counter_IW, counter_WI, counter_OutlinePixels,
    points_IW, points_WI, counter_IW_OutlinePixels);

  currentPixel = QPoint(focusPoint.x()-1, focusPoint.y());
  if (_imgData.valid(currentPixel.x(), currentPixel.y()))
      analyzePixel(currentPixel, counter_IW, counter_WI, counter_OutlinePixels,
    points_IW, points_WI, counter_IW_OutlinePixels);

  if (counter_IW == 2 && counter_WI == 2 && counter_OutlinePixels >= 2)
  {
    if ((points_IW[0].x() != points_IW[1].x() && points_IW[0].y() != points_IW[1].y()) &&
        (points_WI[0].x() != points_WI[1].x() && points_WI[0].y() != points_WI[1].y()))
      if (counter_IW_OutlinePixels == 2)
        return TRUE;
  }

  return FALSE;
}

// Analyzing the seed point's neighboring pixels:
//  - both "IW" and "WI" should be coded in two pixels;
//  - at least two of the considered pixels should be outline pixels 
void SmageImgLSBsDecoder::analyzePixel(const QPoint &currentPixel, int &counter_IW, int &counter_WI, 
               int &counter_OutlinePixels, QPoint *points_IW, QPoint *points_WI,
               int &counter_IW_OutlinePixels)
{
  QColor currentColor = _imgData.pixel(currentPixel.x(), currentPixel.y());

  const int red = currentColor.red();
  const int green = currentColor.green();
  const int blue = currentColor.blue();
  bool isOutlinePixel = FALSE;
  int codeRed = 0;
  int codeGreen = 0;
  int codeBlue = 0;
  
  if (blue & 1) 
  {
    isOutlinePixel = TRUE;
    counter_OutlinePixels++;
  }

  codeRed |= (red & 31);
  codeGreen |= (green & 31);
  codeBlue |= (blue & 30);
  
  if (codeRed==9 && codeGreen==30 && codeBlue==20)
  {
    if (counter_IW<2)
      points_IW[counter_IW] = currentPixel;
    counter_IW++;

    if (isOutlinePixel)
      counter_IW_OutlinePixels++;
  }
  else if (codeRed==23 && codeGreen==6 && codeBlue==18)
  {
    if (counter_WI<2)
      points_WI[counter_WI] = currentPixel;
    counter_WI++;
  }
}

// Start decoding data
void SmageImgLSBsDecoder::decodeInformation()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::decodeInformation()\n");
#endif

  int codePattern, objId;

  const map<unsigned int, SmageIdRegion> &regionMap = _idBuffer->getRegionMap();
  map<unsigned int, SmageIdRegion>::const_iterator regionMapIter;
  vector<SmageIdSubRegion>::const_iterator subRegionsIter;

  vector<char> dataVec;
  vector<char> colorTableVec;  
  
  // Run through the ID-Buffer class' region map and decode information for each
  // seed point
  for (regionMapIter=regionMap.begin(); regionMapIter!=regionMap.end(); regionMapIter++)
  {
    dataVec.clear();
    colorTableVec.clear();

    SmageDataObjs dataObjs;
    SmagePropsStruct propsStruct;

    for (subRegionsIter=regionMapIter->second.subRegions.begin(); 
         subRegionsIter!=regionMapIter->second.subRegions.end(); subRegionsIter++)
    {  
      // Decode the information coded in the seed point
      decodeSeedpoint(subRegionsIter->seedPointPos, objId, codePattern);

#ifdef SMAGE_PRINT_INFOS
      cout << "Decoded Seedpoint: object ID: " << objId << " code pattern: " << codePattern << endl;
#endif
      _dataProcessor->convertIntToEnum(codePattern, dataObjs.codePattern);

      // Find the function for decoding according to the code pattern
      SIE_DecodeFuncCB *decodeFunc = NULL;

      switch (codePattern)
      {
        case SMAGE_LSB: decodeFunc = decodeBits_LSB; break;
        case SMAGE_BITS_3: decodeFunc = decode_BITS_3; break;
        case SMAGE_BITS_4: decodeFunc = decode_BITS_4; break;
        case SMAGE_LS_2B: decodeFunc = decode_LS_2B; break;
        case SMAGE_BITS_6: decodeFunc = decode_BITS_6; break;
        case SMAGE_BITS_7: decodeFunc = decode_BITS_7; break;
        case SMAGE_LS_3B: decodeFunc = decode_LS_3B; break;
        case SMAGE_BITS_9: decodeFunc = decode_BITS_9; break;
        case SMAGE_BITS_10: decodeFunc = decode_BITS_10; break;
        case SMAGE_LS_4B: decodeFunc = decode_LS_4B; break;
        case SMAGE_BITS_12: decodeFunc = decode_BITS_12; break;
        case SMAGE_BITS_13: decodeFunc = decode_BITS_13; break;
        case SMAGE_LS_5B: decodeFunc = decode_LS_5B; break;
        case SMAGE_LS_8B: decodeFunc = decode_LS_8B; break;
        default: 
        {
#ifdef SMAGE_PRINT_DEBUG
          _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::decodeInformation): undefined code pattern ***\n");
#endif
        }
      }

      if (decodeFunc != NULL)
        decodeBits(objId, regionMapIter->second, dataVec, colorTableVec, decodeFunc);
      else
      {
#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::decodeInformation): decode function is NULL ***\n");
#endif
      }
    }

    // Unsplit the decoded byte stream and extract the coded data items
    unsplitDataStream(objId, dataVec, dataObjs, propsStruct);

    // If a color table has been decoded, assign it to SmageDataObjs' colorTable
    if (colorTableVec.size() > 0)
      assignColorTable(colorTableVec, dataObjs);
    
    _dataMap[objId] = dataObjs;
  }
}

// Assign color table to "dataObjs.colorTable";
//  using "_dataProcessor" the byte stream is first uncompressed before the color table can be assigned
void SmageImgLSBsDecoder::assignColorTable(const vector<char> &colorTableVec, SmageDataObjs &dataObjs)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::assignColorTable()\n");
#endif

  _dataProcessor->setData(colorTableVec);
  _dataProcessor->decompressData();

  const char *decompressedColorTable = NULL;
  unsigned long tableSize = 0;
  decompressedColorTable = _dataProcessor->getDecompressedDataStream(tableSize);

  if (tableSize > 0)
    convertCharToRgbVec(decompressedColorTable, tableSize, dataObjs.colorTable);
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::assignColorTable): decompressed color table's size is NULL\n");
#endif
  }
}

void SmageImgLSBsDecoder::synchronizeSubIDs(SmageDataObjs &dataObjs)
{
  list<int> tmpList;
  list<int>::iterator listIter;
  char counter = 0;

  vector<SmageData>::iterator smageDataVecIter;
  for (smageDataVecIter=dataObjs.data.begin();
       smageDataVecIter!=dataObjs.data.end(); smageDataVecIter++)
    tmpList.push_back(int(smageDataVecIter->subID));

  tmpList.sort();
  tmpList.unique();

  for (listIter=tmpList.begin(); listIter!=tmpList.end(); listIter++)
    {
      for (smageDataVecIter=dataObjs.data.begin();
	   smageDataVecIter!=dataObjs.data.end(); smageDataVecIter++)
	{
	  if (((int)(smageDataVecIter->subID)) == *listIter)
	    smageDataVecIter->subID = counter;
	}

      counter++;
    }
}

// Unsplit the decoded data stream to extract the single embedded data items
void SmageImgLSBsDecoder::unsplitDataStream(const int &objId, const vector<char> &dataStream, 
              SmageDataObjs &dataObjs, SmagePropsStruct &propsStruct)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::unsplitDataStream() for object with ID: %d\n", objId);
#endif

  _dataProcessor->setData(dataStream);
  _dataProcessor->unsplitDataStream(dataObjs.data, propsStruct, TRUE);

  synchronizeSubIDs(dataObjs);

  vector<SmageData>::iterator smageDataVecIter;
  for (smageDataVecIter=dataObjs.data.begin();
       smageDataVecIter!=dataObjs.data.end(); smageDataVecIter++)
  {
    // The ID colors for the ID-Buffer and the single data items are synchronized
    QColor tmpFillColor, tmpOutlineColor;
    if (_idBuffer->getFillColor(objId, int(smageDataVecIter->subID), tmpFillColor, tmpOutlineColor))
    {    
      smageDataVecIter->idColor[0] = tmpFillColor.red();
      smageDataVecIter->idColor[1] = tmpFillColor.green();
      smageDataVecIter->idColor[2] = tmpFillColor.blue();
    }
  }
}

// Decode only the object ID and sub ID from a seed point
void SmageImgLSBsDecoder::decodeObjId(const QPoint &seedPointPos, int &objID)
{
  int blue, green;
  objID = 0;

  blue = qBlue(_imgData.pixel(seedPointPos.x(), seedPointPos.y()));
  green = qGreen(_imgData.pixel(seedPointPos.x(), seedPointPos.y()));

  objID |= ((blue & 14) >> 1);
  objID |= ((green & 7) << 3);
}

// Decode information coded in a seed point
void SmageImgLSBsDecoder::decodeSeedpoint(const QPoint &seedPointPos, int &objID, int &codePattern)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::decodeSeedpoint() for seed point at position: %d %d\n", 
      seedPointPos.x(), seedPointPos.y());
#endif

  int red, green, blue;
  codePattern = 0;  // Red
  objID = 0;        // Blue & Green

  red = qRed(_imgData.pixel(seedPointPos.x(), seedPointPos.y()));
  green = qGreen(_imgData.pixel(seedPointPos.x(), seedPointPos.y()));
  blue = qBlue(_imgData.pixel(seedPointPos.x(), seedPointPos.y()));
  codePattern |= (red & 15);
  objID |= ((blue & 14) >> 1);
  objID |= ((green & 7) << 3);

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::decodeSeedpoint() ... object ID: %d, code pattern: %d\n", objID, codePattern);
#endif
}

//  Decode the amount of data from the 4 pixels "_amountOfDataPixels"
void SmageImgLSBsDecoder::decodeAmountOfData(unsigned long &numEncodedBytes)
{
  int curImgCol_Blue, curImgCol_Green, curImgCol_Red;
  QRgb curColor;
  int bitCounter = 0;

  for (int i=0; i<4; i++)
  {
    curColor = _imgData.pixel(_amountOfDataPixels[i].x(), _amountOfDataPixels[i].y());

    curImgCol_Blue = qBlue(curColor);
    curImgCol_Red = qRed(curColor);
    curImgCol_Green = qGreen(curColor);

    numEncodedBytes |= ((curImgCol_Red & 7) << bitCounter);
    bitCounter += 3;
    numEncodedBytes |= ((curImgCol_Green & 7) << bitCounter);
    bitCounter += 3;
    numEncodedBytes |= (((curImgCol_Blue & 6) >> 1) << bitCounter);
    bitCounter += 2;
  }

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::decodeAmountOfData(): %lu\n", numEncodedBytes);
#endif
}

//   Decode the bit information
void SmageImgLSBsDecoder::decodeBits(const int &objId, const SmageIdRegion &idRegion, vector<char> &data, 
             vector<char> &colorTableVec, SIE_DecodeFuncCB *decodeFunc)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsDecoder::decodeBits()\n");
#endif

  char encodedChar = 0;
  int curImgCol_Blue, curImgCol_Green, curImgCol_Red;
  int currentIdBufferColorIndex;
  int bitCounter = 0;
  int dataAmountPixelCounter = 0;
  unsigned long byteCounter = 0;
  bool colorTableDecodingStarted = FALSE;

  QImage idBuffer;
  _idBuffer->getIdBuffer(idBuffer);
  if (idBuffer.isNull())
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsDecoder::decodeBits): invalid ID_Buffer ***\n");
#endif
    return;
  }

  //idBuffer.save("decodeIdBuffer.png", "PNG", 100);

  // Run through the image and check if the current color is among the colors for that
  // object ID (in ID-Buffer class)
  for (int y=0; y<_imgHeight; y++)
    for (int x=0; x<_imgWidth; x++)
    {
      currentIdBufferColorIndex = idBuffer.pixelIndex(x,y);
  
      // Does this pixel belong to the current seed point
      if ((_idBuffer->isAmongFillIndices(objId, currentIdBufferColorIndex) || 
	   _idBuffer->isAmongOutlineIndices(objId, currentIdBufferColorIndex)) && 
          isDecodablePoint(idRegion, QPoint(x, y)))
      {
        curImgCol_Blue = QColor(_imgData.pixel(x,y)).blue();
        curImgCol_Red = QColor(_imgData.pixel(x,y)).red();
        curImgCol_Green = QColor(_imgData.pixel(x,y)).green();
      
        // First 4 pixels are reserved and hold the amount of encoded data
        if (dataAmountPixelCounter < 4)
        {
          _amountOfDataPixels[dataAmountPixelCounter] = QPoint(x,y);
          dataAmountPixelCounter++;

          if (dataAmountPixelCounter == 4)
          {
            decodeAmountOfData(byteCounter);

            if (byteCounter==NO_COLOR_TABLE_CODE && colorTableDecodingStarted)
              return;
          }
        }
      
        else
        {
          if (byteCounter == 0)
          {  
#ifdef SMAGE_PRINT_DEBUG
            if (!colorTableDecodingStarted)
              _SmageLogFile.lprintf("Data decoding has been finished ... next is color table\n");
            else
              _SmageLogFile.lprintf("Data decoding has been finished.\n");
#endif
  
            // Decoding of data has been finished;
            // if a color table is coded, it is decoded next
            if (!colorTableDecodingStarted)
            {
#ifdef SMAGE_PRINT_DEBUG
	      _SmageLogFile.lprintf("Start decoding color table\n");
#endif 
              colorTableDecodingStarted = TRUE;
              _amountOfDataPixels[0] = QPoint(x,y);
              dataAmountPixelCounter = 1;

              resetDecCounter(bitCounter, byteCounter, encodedChar, data);
            }
            else
              return;
          }

          else
          {
            if (!colorTableDecodingStarted)
              (*decodeFunc)(curImgCol_Red, curImgCol_Green, curImgCol_Blue, bitCounter, byteCounter, encodedChar, data);
            else
              (*decodeFunc)(curImgCol_Red, curImgCol_Green, curImgCol_Blue, bitCounter, byteCounter, encodedChar, colorTableVec);
          }
        }
      }
    }
}

// Test if the current pixel is decodable;
//  it should neither be a seed point nor one of the seed points surrounding pixels
bool SmageImgLSBsDecoder::isDecodablePoint(const SmageIdRegion &idRegion, const QPoint &currentPos)
{
  vector<SmageIdSubRegion>::const_iterator subRegionVecIter;
  int seedPoint_X = idRegion.subRegions[0].seedPointPos.x();
  int seedPoint_Y = idRegion.subRegions[0].seedPointPos.y();

  for (subRegionVecIter=idRegion.subRegions.begin();
       subRegionVecIter!=idRegion.subRegions.end(); subRegionVecIter++)
  {
    seedPoint_X = subRegionVecIter->seedPointPos.x();
    seedPoint_Y = subRegionVecIter->seedPointPos.y();

    if ((currentPos.x() == seedPoint_X && currentPos.y() == seedPoint_Y) ||
        (currentPos.x() == seedPoint_X && 
        (currentPos.y() == (seedPoint_Y-1) || currentPos.y() == (seedPoint_Y+1))) ||
        (currentPos.y() == seedPoint_Y && 
        (currentPos.x() == (seedPoint_X-1) || currentPos.x() == (seedPoint_X+1))))
      return FALSE;
  }
  
  return TRUE;
}

// "bitCounter" equals 8; another byte is complete and data is reset
void SmageImgLSBsDecoder::resetDecCounter(int &bitCounter, unsigned long &byteCounter, char &encodedChar,
            vector<char> &data)
{
  if (byteCounter > 0)
  {
    data.push_back(encodedChar);
    byteCounter--;
  }

  encodedChar = 0;
  bitCounter = 0;
}

// Decode 2 bits
void SmageImgLSBsDecoder::decodeBits_LSB(const int &curImgCol_Red, const int &curImgCol_Green, 
           const int &, int &bitCounter, 
           unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

//Decode 3 bits
void SmageImgLSBsDecoder::decode_BITS_3(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
          int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

}

// Decode 4 bits
void SmageImgLSBsDecoder::decode_BITS_4(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
          int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 5 bits
void SmageImgLSBsDecoder::decode_LS_2B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
               int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 6 bits
void SmageImgLSBsDecoder::decode_BITS_6(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
          int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 7 bits
void SmageImgLSBsDecoder::decode_BITS_7(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
          int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 8 bits
void SmageImgLSBsDecoder::decode_LS_3B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
               int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 9 bits
void SmageImgLSBsDecoder::decode_BITS_9(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
          int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 10 bits
void SmageImgLSBsDecoder::decode_BITS_10(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
           int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 11 bits
void SmageImgLSBsDecoder::decode_LS_4B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
               int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 12 bits
void SmageImgLSBsDecoder::decode_BITS_12(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
           int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 13 bits
void SmageImgLSBsDecoder::decode_BITS_13(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
           int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);  
  if (curImgCol_Blue & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 14 bits
void SmageImgLSBsDecoder::decode_LS_5B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
               int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);  
  if (curImgCol_Blue & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 23 bits
void SmageImgLSBsDecoder::decode_LS_8B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
               int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  if (curImgCol_Red & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 32) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 64) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Red & 128) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Green & 1) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 32) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 64) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Green & 128) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  if (curImgCol_Blue & 2) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 4) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);  
  if (curImgCol_Blue & 8) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 16) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 32) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 64) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
  if (curImgCol_Blue & 128) encodedChar |= pot2[bitCounter];
  bitCounter++;
  if (bitCounter == 8) resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}

// Decode 24 bits; each color channel equals one byte
void SmageImgLSBsDecoder::decode_MAX(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, char &encodedChar, vector<char> &data)
{
  encodedChar = curImgCol_Red;
  resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  encodedChar = curImgCol_Green;
  resetDecCounter(bitCounter, byteCounter, encodedChar, data);

  encodedChar = curImgCol_Blue;
  resetDecCounter(bitCounter, byteCounter, encodedChar, data);
}
