#ifndef  _SMAGE_IMG_LSBS_DECODER_H
#define  _SMAGE_IMG_LSBS_DECODER_H

#include "SmageImgCoder.h"

/** Function type for decoding bits.
  */
typedef void SIE_DecodeFuncCB(const int &, const int &, const int &, int &, unsigned long &, char &, 
            std::vector<char> &);

/** Class SmageImgLSBsDecoder for decoding data from an image.
  *
  * @ingroup LibSmage
  *
  * The constructor's argument is the image from which the data shall be decoded.<br>
  *
  * Example: <br>
  * -------- <br>
  * <br>
  * <code><pre> 
    // "img" is the image that shall be processed;
    // "idBuffer" is the ID-Buffer that will hold (after the
    //  decoding process) the areas in "img" which are coded;
    
    // After decoding, "idBuffer" is a filled image and "codeMap"
    // holds the data with regard to the colors in "idBuffer"
  
    QImage idBuffer;
    std::map<unsigned int, SmageDataObjs> codeMap;
  
    SmageImgDecoder imgDecoder(img);
    imgDecoder.decodeData(idBuffer, codeMap);
  * </pre></code>
  */
class SmageImgLSBsDecoder : public SmageImgCoder
{
  public:
    /** Constructor.
      * argument is the image in which data is coded.
      */
    SmageImgLSBsDecoder(const QImage &imgData);

    /** Destructor.
      */
    ~SmageImgLSBsDecoder();

    /** Decode data from the image and fill "idBuffer" and "codeDataMap".
      */
    void decodeData(QImage &idBuffer, std::map<unsigned int, SmageDataObjs> &codedDataMap);

    /** Decode data from the image and fill "idBuffer", "codeDataMap" and fill pixmap with outlines.
      */
    void decodeData(QImage &idBuffer, std::map<unsigned int, SmageDataObjs> &codedDataMap,
          QPixmap &outlinePixmap);

    /** Decode data from the image using specific method of image downgrading (data must have been encoded using this method).
      */
    void decodeImgDownGrading(const char *fileName);

    /** Test if image holds encoded data.
      */
    bool isInformationEncoded();

    /** Erase data from the image that is associated with the specified colors in the ID-Buffer;
	or all the encoded data if "allData" is TRUE
     */
    void eraseDataFromImage(const bool &allData, const QImage &idBuffer=QImage(), 
			    const std::vector<SmageIdColorStruct> &colorValues=std::vector<SmageIdColorStruct>());

  private:
      
    /** Search for seed points within the image.
      */
    void scanForSeedPoints();

    /** Is the current point a seed point?
      */
    bool isSeedPoint(const QPoint &focusPoint);

    /** Decode information coded in a seed point.
      */
    void decodeSeedpoint(const QPoint &seedPointPos, int &objID, int &codePattern);

    /** Analyzing a pixel regarding seed point characteristics and neighboring pixels.
      */
    void analyzePixel(const QPoint &currentPixel, int &counter_I, int &counter_W, 
      int &counter_OutlinePixels, QPoint *points_I, QPoint *points_W,
      int &counter_IW_OutlinePixels);
      
    /** Start decoding data.
      */
    void decodeInformation();

    /** Adapt the seed point and its 4 surrounding pixels to other pixels in this region.
      */
    void eraseSeedPointRegion(const QPoint &seedPoint);

    /** Decode the bit information.
      */
    void decodeBits(const int &objId, const SmageIdRegion &idRegion, std::vector<char> &data, std::vector<char> &colorTable,
          SIE_DecodeFuncCB *decodeFunc);
      
    /** Test if current pixel is decodable.
      */
    bool isDecodablePoint(const SmageIdRegion &idRegion, const QPoint &currentPos);

    /** Assign color table to "dataObjs.colorTable".
      */
    void assignColorTable(const std::vector<char> &colorTableVec, SmageDataObjs &dataObjs);

    /** Decode only the object ID and sub ID from a seed point.
      */
    void decodeObjId(const QPoint &seedPointPos, int &objID);

    /** Decode the amount of data from the 4 pixels "_amountOfDataPixels".
      */
    void decodeAmountOfData(unsigned long &numEncodedBytes);

    /** Unsplit the decoded data stream to extract the single embedded data items.
      */
    void unsplitDataStream(const int &objId, const std::vector<char> &dataStream, SmageDataObjs &dataObjs,
           SmagePropsStruct &propsStruct);
    
    void synchronizeSubIDs(SmageDataObjs &dataObjs);

    /** Bit decoding for different numbers of modified bits.
      */
    static void resetDecCounter(int &bitCounter, unsigned long &byteCounter, char &curCodedByte, std::vector<char> &data);
    static void decodeBits_LSB(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
         int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_BITS_3(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
        int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_BITS_4(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
        int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_LS_2B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_BITS_6(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
        int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_BITS_7(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
        int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_LS_3B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_BITS_9(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
        int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_BITS_10(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
         int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_LS_4B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_BITS_12(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
        int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_BITS_13(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
         int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_LS_5B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_LS_8B(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
    static void decode_MAX(const int &curImgCol_Red, const int &curImgCol_Green, const int &curImgCol_Blue,
           int &bitCounter, unsigned long &byteCounter, char &curCodeByte, std::vector<char> &data);
};

#endif
