#include "SmageImgLSBsEncoder.h"
#include "SmageSettings.h"

#include <iostream>
#include <math.h>

using namespace std;


// Constructor;
SmageImgLSBsEncoder::SmageImgLSBsEncoder(QImage &imgData) : SmageImgCoder(imgData)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::SmageImgLSBsEncoder()\n");

  if (!_isRGBImg)
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::SmageImgLSBsEncoder): image is not a RGB image ***\n");
#endif

  _progressCBFunc = NULL;
  _progressUserData = NULL;
}

// Destructor
SmageImgLSBsEncoder::~SmageImgLSBsEncoder()
{
}

// Encode data into the whole image; there is no ID-Buffer necessary
bool SmageImgLSBsEncoder::encodeData(const SmageDataObjs &data)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::encodeData (into the whole image)\n");

  if (!_isRGBImg)
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::encodeData): image is not a RGB image, encoding impossible ***\n");
#endif

  if (_isRGBImg)
  {
    // Initialize "_idBuffer" and "_dataProcessor"
    initialize(TRUE);

    // Generate a compressed data stream and insert it into "_dataMap" at "DEFAULT_MAP_POS" (since there 
    // is only one object ID)
    _dataMap[DEFAULT_MAP_POS] = getPreparedData(data);
    SmageDataObjs *preparedData = &_dataMap[DEFAULT_MAP_POS];

    // Unset blue's LSB
    resetImgData();
      
    // "regionMap" holds image regions and their seed point positions
    const map<unsigned int, SmageIdRegion> &regionMap = _idBuffer->getRegionMap();
    map<unsigned int, SmageIdRegion>::const_iterator iter = regionMap.find(DEFAULT_MAP_POS);
    if (iter != regionMap.end())
      {
	if (!initDataCoding(DEFAULT_MAP_POS, *preparedData, iter->second))
	  return FALSE;
      }
    else
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::encodeData): \"regionMap\" may be empty ***\n");
#endif
      return FALSE;
    }
  }

  return TRUE;
}

// Encode data into regions specified by "idBufferData"
bool SmageImgLSBsEncoder::encodeData(const SmageIdBufferStruct &idBufferData, const vector<SmageDataObjs> &data,
				     const bool &resetImg, const int &startId)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::encodeData (into image regions)\n");

  if (data.empty())
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::encodeData): no data specified ***\n");
#endif

  if (_imgWidth != idBufferData.idBuffer.width() || _imgHeight != idBufferData.idBuffer.height())
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::encodeData): ID-Buffer must be of same size as cover image ***\n");
#endif
    return FALSE;
  }
  else if (_isRGBImg)
  { 
    // Initialize "_idBuffer" and "_dataProcessor"
    initialize(FALSE);

    int id = startId;
    vector<SmageDataObjs>::const_iterator dataIter;
    vector<SmageData>::const_iterator subDataIter;
    for (dataIter=data.begin(); dataIter!=data.end(); dataIter++)
    {
      _dataMap[id] = getPreparedData(*dataIter);
      char currentSubID = -1;

      // Extract for each data item its ID color and assign it to the ID-Buffer
      // (the correlation between ID-Buffer regions and their associated data items)
      vector<SmageIdColorStruct> colors;
      for (subDataIter=dataIter->data.begin(); subDataIter!=dataIter->data.end(); subDataIter++)
      {
        SmageIdColorStruct colorStruct;

        if (idBufferData.bufferType == SMAGE_OUTLINES)
        {
          colorStruct.outlineColor = qRgb(subDataIter->idColor[0], subDataIter->idColor[1], subDataIter->idColor[2]);
#ifdef SMAGE_PRINT_INFOS
          cout << "Object with ID: " << id << " and Sub-ID: " << (int(subDataIter->subID) & 255) << " has ID color: " 
               << qRed(colorStruct.outlineColor) << " " << qGreen(colorStruct.outlineColor) << " " 
               << qBlue(colorStruct.outlineColor) << endl;
#endif
        }
        else if (idBufferData.bufferType == SMAGE_FILLED_REGIONS || idBufferData.bufferType == SMAGE_OUTLINED_REGIONS)
        {
          colorStruct.fillColor = qRgb(subDataIter->idColor[0], subDataIter->idColor[1], subDataIter->idColor[2]);
#ifdef SMAGE_PRINT_INFOS
          cout << "Object with ID: " << id << " and Sub-ID: " << (int(subDataIter->subID) & 255) << " has ID color: " 
               << qRed(colorStruct.fillColor) << " " << qGreen(colorStruct.fillColor) << " " 
               << qBlue(colorStruct.fillColor) << endl;
#endif
        }

        colorStruct.subID = subDataIter->subID;

	if (currentSubID != subDataIter->subID)
	  {
	    colors.push_back(colorStruct);
	    currentSubID = subDataIter->subID;
	  }
      }

      _idBuffer->insertIdColorsValue(id, colors);
      id++;
    }

    if (_progressCBFunc != NULL && _progressUserData != NULL)
      (*_progressCBFunc)(_progressUserData);

    // Give ID-Buffer data to ID-Buffer class
    _idBuffer->setIdBuffer(idBufferData.idBuffer, idBufferData.backgroundColor, idBufferData.bufferType);

    if (_progressCBFunc != NULL && _progressUserData != NULL)
      (*_progressCBFunc)(_progressUserData);

    // Prepare ID-Buffer (fill outlines or generate outlines for already filled regions)
    if (idBufferData.bufferType == SMAGE_OUTLINES)  
      _idBuffer->fillOutlineRegions();
    else if (idBufferData.bufferType == SMAGE_FILLED_REGIONS)
      _idBuffer->generateOutlines();
    else if (idBufferData.bufferType == SMAGE_OUTLINED_REGIONS)
      _idBuffer->analyzeIdBuffer();
    else
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::encodeData): Unknown buffer type ***\n");
#endif
      return FALSE;
    }

    if (_progressCBFunc != NULL && _progressUserData != NULL)
      (*_progressCBFunc)(_progressUserData);

    // Unset blue's LSB
    if (resetImg)
      resetImgData();

    const map<unsigned int, SmageIdRegion> &regionMap = _idBuffer->getRegionMap();
    map<unsigned int, SmageIdRegion>::const_iterator regionMapIter;
    map<unsigned int, SmageDataObjs>::const_iterator dataMapIter;
      
    for (regionMapIter=regionMap.begin(); regionMapIter!=regionMap.end(); regionMapIter++)
    {
      dataMapIter = _dataMap.find(regionMapIter->first);
      if (dataMapIter != _dataMap.end())
        {
	  if (!initDataCoding(regionMapIter->first, dataMapIter->second, regionMapIter->second))
	    return FALSE;
	}
      else
      {
#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.
	  lprintf("*** Warning (SmageImgLSBsEncoder::encodeData): region map ID differs from data map ID ***\n");
#endif
	return FALSE;
      }

      if (_progressCBFunc != NULL && _progressUserData != NULL)
	(*_progressCBFunc)(_progressUserData);
    }
  }
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::encodeData): image is not a RGB image, encoding impossible ***\n");
#endif
    return FALSE;
  }
  
  return TRUE;
}

// Specific method of data coding (encoding an image into another image);
//  the 4 least significant bits of the cover image "_imgData" are replace by the 4 most significant bits
//  of the image to be encoded "codeData";
void SmageImgLSBsEncoder::doImageDowngrading(const QImage &codeData)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::doImageDowngrading()\n");
#endif

  // Based on the description in: 
  // C. Kurak and J. McHugh "A Cautionary Note on Image Downgrading"
  // "... replacing the low order bits of one image with the high order bits of another."

  if (_isRGBImg)
  {
    if (codeData.width() > _imgWidth || codeData.height() > _imgHeight)
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::doImageDowngrading): Image to be embedded is bigger than cover image ***\n");
#endif
    }
      
    QRgb curCoverColor, curCodeColor;
    int curCoverCol_Blue, curCoverCol_Green, curCoverCol_Red;
    int curCodeCol_Blue, curCodeCol_Green, curCodeCol_Red;

    for (int y=0; y<_imgHeight; y++)
      for (int x=0; x<_imgWidth; x++)
      {
        if (codeData.valid(x,y))
          curCodeColor = codeData.pixel(x, y);
        else
          curCodeColor = qRgb(255,255,255);
      
          curCoverColor = _imgData.pixel(x, y);
      
          // Modify color channels of cover image
          curCoverCol_Blue = qBlue(curCoverColor); curCoverCol_Blue &= 240;
          curCoverCol_Red = qRed(curCoverColor); curCoverCol_Red &= 240;
          curCoverCol_Green = qGreen(curCoverColor); curCoverCol_Green &= 240;
          curCodeCol_Blue = qBlue(curCodeColor);
          curCodeCol_Red = qRed(curCodeColor);
          curCodeCol_Green = qGreen(curCodeColor);

          curCoverCol_Red |= ((curCodeCol_Red & 240) >> 4);
          curCoverCol_Green |= ((curCodeCol_Green & 240) >> 4);
          curCoverCol_Blue |= ((curCodeCol_Blue & 240) >> 4);

          // Set modified colors
          _imgData.setPixel(x, y, qRgb(curCoverCol_Red, curCoverCol_Green, curCoverCol_Blue));
      }
  }
}

// Unset LSB of blue color chanel
void SmageImgLSBsEncoder::resetImgData()
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::resetImgData()\n");
#endif

  int curImgCol_Blue, curImgCol_Green, curImgCol_Red;

  for (int i=0; i<_imgHeight; i++)
    for (int j=0; j<_imgWidth; j++)
    {
      curImgCol_Blue = QColor(_imgData.pixel(j,i)).blue();
      curImgCol_Red = QColor(_imgData.pixel(j,i)).red();
      curImgCol_Green = QColor(_imgData.pixel(j,i)).green();
      curImgCol_Blue &= 254;

      _imgData.setPixel(j, i, qRgb(curImgCol_Red, curImgCol_Green, curImgCol_Blue));
    }
}

//   Initialize data encoding
bool SmageImgLSBsEncoder::initDataCoding(const int &objID, const SmageDataObjs &codeDataInfo,
					 const SmageIdRegion &idRegion)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::initDataCoding()\n");
#endif

  // Get the size of the compressed color table ...
  unsigned long compressedColorTableSize = 0;
  const char *compressedColorTable = NULL;
  vector<char> charVec;
  if (codeDataInfo.colorTable.size() > 0)
  {
    convertRgbToCharVec(codeDataInfo.colorTable, charVec);
    _dataProcessor->setData(charVec);
    _dataProcessor->compressData();
    compressedColorTable = _dataProcessor->getCompressedDataStream(compressedColorTableSize);
  }

  // Is specified code pattern suited or if "SMAGE_AUTO" find suitable code pattern
  CodePattern newCodePattern;
  if (!checkAvailableDataSpace(idRegion.numOfAllInnerPixels + idRegion.numOfAllOutlinePixels - (2*SMAGE_NUM_AMOUNT_OF_DATA_PIXELS), 
			       codeDataInfo.codePattern, codeDataInfo.codeByteStream.size() + compressedColorTableSize, newCodePattern))
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::initDataCoding): too much data for given image space ***\n");
#endif  
    return FALSE;
  }

#ifdef SMAGE_PRINT_INFOS
  cout << "Specified code pattern: " << codeDataInfo.codePattern << "  New code pattern: " << newCodePattern << endl;
#endif

  // Encode data in according seed points; seed point positions are set in "SmageIdSubRegion"
  vector<SmageIdSubRegion>::const_iterator subRegionIter;
  for (subRegionIter=idRegion.subRegions.begin();
       subRegionIter!=idRegion.subRegions.end(); subRegionIter++)
    codeSeedPoint(objID, subRegionIter->seedPointPos, newCodePattern);
  
  SIE_EncodeFuncCB *codeFunc = NULL;
  
  // call encode function according to "newCodePattern"
  switch (newCodePattern)
  {
    case SMAGE_LSB: codeFunc = code_LSB; break;
    case SMAGE_BITS_3: codeFunc = code_BITS_3; break;
    case SMAGE_BITS_4: codeFunc = code_BITS_4; break;
    case SMAGE_LS_2B: codeFunc = code_LS_2B; break;
    case SMAGE_BITS_6: codeFunc = code_BITS_6; break;
    case SMAGE_BITS_7: codeFunc = code_BITS_7; break;
    case SMAGE_LS_3B: codeFunc = code_LS_3B; break;
    case SMAGE_BITS_9: codeFunc = code_BITS_9; break;
    case SMAGE_BITS_10: codeFunc = code_BITS_10; break;
    case SMAGE_LS_4B: codeFunc = code_LS_4B; break;
    case SMAGE_BITS_12: codeFunc = code_BITS_12; break;
    case SMAGE_BITS_13: codeFunc = code_BITS_13; break;
    case SMAGE_LS_5B: codeFunc = code_LS_5B; break;
    case SMAGE_LS_8B: codeFunc = code_LS_8B; break;
    default: 
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::initDataCoding): undefined code pattern ***\n");
#endif 
    }
  }

  if (codeFunc != NULL)
    codeData(objID, idRegion, codeDataInfo, codeFunc);
  else
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::initDataCoding): code function is NULL ***\n");
#endif 
    return FALSE;
  }

  return TRUE;
}

// Code a seed point that holds important information for decoding;
// in addition, 4 surrounding pixels are coded in a way that the image can be regarded as a coded image
void SmageImgLSBsEncoder::codeSeedPoint(const int &objID, const QPoint &seedPointPos, 
					const CodePattern &codePattern)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::codeSeedPoint()\n");
#endif

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("Seedpoint: ID: %d  Position: %d %d\n", objID, seedPointPos.x(), seedPointPos.y());
#endif

  // SeedPoint position is given;
  // (1) Encode necessary data in SeedPoint
  // (2) Encode 'IW' and 'WI' in neighboring pixels (4-neighborhood), alternately

  QColor colorP1, colorP2;
  int blue, green, red, blueP1, greenP1, redP1, blueP2, greenP2, redP2;
  int codeRed, codeGreen, codeBlue;

  // (1) Necessary data in SeedPoint
  // Red:   "codePattern"
  // Blue:  Object ID (6 bits, 63 different IDs) ... first 3 bits of ID
  // Green: last 3 bits of ID
  
  red = qRed(_imgData.pixel(seedPointPos.x(), seedPointPos.y()));
  green = qGreen(_imgData.pixel(seedPointPos.x(), seedPointPos.y()));
  blue = qBlue(_imgData.pixel(seedPointPos.x(), seedPointPos.y()));

  red &= 240; red |= (codePattern & 15);
  green &= 248;
  blue &= 241; 
  
  blue |= ((objID & 7) << 1);
  green |= ((objID & 56) >> 3);
  
  _imgData.setPixel(seedPointPos.x(), seedPointPos.y(), qRgb(red, green, blue));

  // (2) Encoding 'IW' and 'WI':
  // Bit 8 in 'I' and 'W' is not needed, hence only 14 bits have to be coded;
  // Red:   Bits 1,2,3,4,5 are used
  // Green: Bits 1,2,3,4,5 are used
  // Blue:  Bit 1 is reserved for boundary, Bits 2,3,4,5 are used

  // 'I': 01001001 => 73
  // 'W': 01010111 => 87

  for (int k=0; k<2; k++)
  {
    if (k==0)
    {
      colorP1 = _imgData.pixel(seedPointPos.x(), seedPointPos.y()-1);
      colorP2 = _imgData.pixel(seedPointPos.x()-1, seedPointPos.y());
      codeRed = 9; codeGreen = 30; codeBlue = 20;
    }
    else
    {
      colorP1 = _imgData.pixel(seedPointPos.x()+1, seedPointPos.y());
      colorP2 = _imgData.pixel(seedPointPos.x(), seedPointPos.y()+1);
      codeRed = 23; codeGreen = 6; codeBlue = 18;
    }

    blueP1 = colorP1.blue(); blueP1 &= 225;
    greenP1 = colorP1.green(); greenP1 &= 224;
    redP1 = colorP1.red(); redP1 &= 224;

    blueP2 = colorP2.blue(); blueP2 &= 225;
    greenP2 = colorP2.green(); greenP2 &= 224;
    redP2 = colorP2.red(); redP2 &= 224;

    blueP1 |= codeBlue; blueP2 |= codeBlue;
    greenP1 |= codeGreen; greenP2 |= codeGreen;
    redP1 |= codeRed; redP2 |= codeRed;

    if (k==0)
    {
      _imgData.setPixel(seedPointPos.x(), seedPointPos.y()-1, qRgb(redP1, greenP1, blueP1));
      _imgData.setPixel(seedPointPos.x()-1, seedPointPos.y(), qRgb(redP2, greenP2, blueP2));
    }

    else
    {
      _imgData.setPixel(seedPointPos.x()+1, seedPointPos.y(), qRgb(redP1, greenP1, blueP1));
      _imgData.setPixel(seedPointPos.x(), seedPointPos.y()+1, qRgb(redP2, greenP2, blueP2));
    }
  }
}

// Encode the amount of data in the 4 regarding pixels ("_amountOfDataPixels")
void SmageImgLSBsEncoder::encodeAmountOfData(const unsigned long &numEncodedBytes)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::encodeAmountOfData(): %lu\n", numEncodedBytes);
#endif

#ifdef SMAGE_PRINT_INFOS
  cout << "Encode amount of data (bytes): " << numEncodedBytes << endl;
#endif

  int curImgCol_Blue, curImgCol_Green, curImgCol_Red;
  QRgb curColor;
  unsigned long numEncodedBytesCopy = numEncodedBytes;
  
  for (int i=0; i<4; i++)
  {
    curColor = _imgData.pixel(_amountOfDataPixels[i].x(), _amountOfDataPixels[i].y());

    curImgCol_Blue = qBlue(curColor);
    curImgCol_Red = qRed(curColor);
    curImgCol_Green = qGreen(curColor);
    curImgCol_Blue &= 249; curImgCol_Red &= 248; curImgCol_Green &= 248;
    curImgCol_Red |= (numEncodedBytesCopy & 7);
    numEncodedBytesCopy = (numEncodedBytesCopy >> 3);
    curImgCol_Green |= (numEncodedBytesCopy & 7);
    numEncodedBytesCopy = (numEncodedBytesCopy >> 3);
    curImgCol_Blue |= ((numEncodedBytesCopy & 3) << 1);
    numEncodedBytesCopy = (numEncodedBytesCopy >> 2);

    _imgData.setPixel(_amountOfDataPixels[i].x(), _amountOfDataPixels[i].y(), 
    qRgb(curImgCol_Red, curImgCol_Green, curImgCol_Blue));
  }
}

// Test if the current pixel is codable;
//  seed point pixel is not codable as well as the 4 surrounding pixels 
//  (identification of an coded image)
bool SmageImgLSBsEncoder::isCodablePixel(const int &x, const int &y,
             const vector<SmageIdSubRegion> &subRegions)
{
  vector<SmageIdSubRegion>::const_iterator iter; 
  for (iter=subRegions.begin(); iter!=subRegions.end(); iter++)
  {
    int seed_X = iter->seedPointPos.x();
    int seed_Y = iter->seedPointPos.y();

    if ((seed_X == x && seed_Y == y) ||
        (seed_X == x && ((seed_Y == (y-1)) || (seed_Y == (y+1)))) ||
        (seed_Y == y && ((seed_X == (x-1)) || (seed_X == (x+1)))))
      return FALSE;
  }

  return TRUE;
}

// Test if the current position is a seed point position 
bool SmageImgLSBsEncoder::isSeedPointPos(const int &x, const int &y,
           const vector<SmageIdSubRegion> &subRegions)
{
  vector<SmageIdSubRegion>::const_iterator iter; 
  for (iter=subRegions.begin(); iter!=subRegions.end(); iter++)
  {
    if ((iter->seedPointPos.x() == x) && (iter->seedPointPos.y() == y))
      return TRUE;
  }

  return FALSE;
}

//   Start encoding the data
void SmageImgLSBsEncoder::codeData(const int &objID, const SmageIdRegion &idRegion, 
           const SmageDataObjs &codeDataInfo, SIE_EncodeFuncCB *codeFunc)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::codeData(): for ID: %d\n", objID);
#endif

  if (_imgData.isNull() || !_isRGBImg)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::codeData): image is NULL or non-RGB ***\n");
#endif
    return;
  }
  
  int bitCounter = 0;
  unsigned long byteCounter = 0;
  int amountPixelCounter = 0;
  int curImgCol_Blue, curImgCol_Green, curImgCol_Red;
  vector<char>::const_iterator curCodeByte = codeDataInfo.codeByteStream.begin();
  bool endOfByteStream = FALSE;
  bool encodingDone = FALSE;
  unsigned long numOfAvailablePixel = idRegion.numOfAllInnerPixels + idRegion.numOfAllOutlinePixels;

  // Get the ID-Buffer from ID-Buffer class
  QImage idBuffer;
  _idBuffer->getIdBuffer(idBuffer);
  if (idBuffer.isNull())
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::codeData): invalid ID_Buffer ***\n");
#endif
    return;
  }

#ifdef SMAGE_PRINT_INFOS
  cout << "Number of available pixels (before encoding): " << numOfAvailablePixel << endl;
  cout << "Amount of Data to be encoded: " << codeDataInfo.codeByteStream.size() << endl;
#endif

  //idBuffer.save("idBuffer.png", "PNG", 100);

  int currentIdBufferColorIndex;

  // If a color table is available, it is prepared (compressed) by the "_dataProcessor"
  vector<char> charVec;
  unsigned long compressedColorTableSize = 0;
  const char *compressedColorTable = NULL;
  bool colorTableAvailable = FALSE;
  bool colorTableAmountOfDataCoded = FALSE;
  if (codeDataInfo.colorTable.size() > 0)
  {
    colorTableAvailable = TRUE;
    convertRgbToCharVec(codeDataInfo.colorTable, charVec);
    _dataProcessor->setData(charVec);
    _dataProcessor->compressData();
    compressedColorTable = _dataProcessor->getCompressedDataStream(compressedColorTableSize);

#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("SmageImgLSBsEncoder::codeData(): compressed color table size: %lu\n", compressedColorTableSize);
#endif
  }

  unsigned long currentByteStreamSize = codeDataInfo.codeByteStream.size();
  bool colorTableEncodingStarted = FALSE;

  // Go through the image and search for codable pixels (according to ID-Buffer)
  for (int i=0; i<_imgHeight; i++)
  {
    for (int j=0; j<_imgWidth; j++)
    {
      if (!endOfByteStream)
      {
        if (byteCounter >= currentByteStreamSize)
          endOfByteStream = TRUE;
      }

      curImgCol_Blue = QColor(_imgData.pixel(j,i)).blue();
      curImgCol_Red = QColor(_imgData.pixel(j,i)).red();
      curImgCol_Green = QColor(_imgData.pixel(j,i)).green();

      currentIdBufferColorIndex = idBuffer.pixelIndex(j,i);

      // If pixel is part of the outline, LSB of blue is set
      if (_idBuffer->isAmongOutlineIndices(objID, currentIdBufferColorIndex) ||
          isSeedPointPos(j, i, idRegion.subRegions))
        curImgCol_Blue |= 1;

      // Encode information within the outline
      if (!encodingDone)
      {
        // Is current color index among ID-Buffer's color indices for that ID
        if (_idBuffer->isAmongFillIndices(objID, currentIdBufferColorIndex) ||
	    _idBuffer->isAmongOutlineIndices(objID, currentIdBufferColorIndex))
        {          
          if (isCodablePixel(j, i, idRegion.subRegions))
          {  
            // First 4 codable pixels are reserved to encode the amount of data
            if (amountPixelCounter < 4)
            {
              _amountOfDataPixels[amountPixelCounter] = QPoint(j,i);
              amountPixelCounter++;
              if (numOfAvailablePixel != 0)
                numOfAvailablePixel--;
            }
            else
            {
              if (!endOfByteStream && (numOfAvailablePixel != 0))
              {
                (*codeFunc)(curImgCol_Red, curImgCol_Green, curImgCol_Blue,
                    bitCounter, byteCounter, curCodeByte, currentByteStreamSize);
                if (numOfAvailablePixel != 0)
                  numOfAvailablePixel--;
              }
      
              // If end of byte stream encode amount of data
              if (numOfAvailablePixel == 0 || endOfByteStream)
              {
#ifdef SMAGE_PRINT_DEBUG
                if (numOfAvailablePixel == 0)
                  _SmageLogFile.lprintf("SmageImgLSBsEncoder::codeData(): no pixels left\n");
                if (endOfByteStream)
                  _SmageLogFile.lprintf("SmageImgLSBsEncoder::codeData(): end of byte stream\n");
#endif
                if (!colorTableEncodingStarted || !colorTableAvailable)
		  encodeAmountOfData(currentByteStreamSize);
                else if (amountPixelCounter == 4)
                {
                  encodeAmountOfData(compressedColorTableSize);
                  colorTableAmountOfDataCoded = TRUE;
                }

                // End of byte stream ("codeDataInfo.codeByteStream") has been reached;
                // start encoding the color table
                if (!colorTableEncodingStarted)
                {
                  colorTableEncodingStarted = TRUE;
                  amountPixelCounter = 0;

                  if (numOfAvailablePixel  != 0)
                  {	      
		    _amountOfDataPixels[0] = QPoint(j,i);
                    amountPixelCounter = 1;
                    numOfAvailablePixel--;
                  }

                  byteCounter = bitCounter = 0;    

                  if (colorTableAvailable)
                  {
#ifdef SMAGE_PRINT_INFOS
                    cout << "Color table size: " << compressedColorTableSize << endl;
#endif

                    currentByteStreamSize = compressedColorTableSize;
                    endOfByteStream = FALSE;
                    curCodeByte = (vector<char>::const_iterator)(compressedColorTable);
                  }
                  else
                  {
                    endOfByteStream = TRUE;
                    // Color table size of NO_COLOR_TABLE_CODE means there is no color table encoded
                    currentByteStreamSize = NO_COLOR_TABLE_CODE;
                  }
                }
                else if (endOfByteStream)
                  encodingDone = TRUE;
              }
            }
          }
        }
      }   

      _imgData.setPixel(j, i, qRgb(curImgCol_Red, curImgCol_Green, curImgCol_Blue));
    }    
  }

#ifdef SMAGE_PRINT_INFOS
  cout << "Number of available pixel (after encoding) " << numOfAvailablePixel << endl;
#endif
}

// Verify if the specified code pattern is suited
bool SmageImgLSBsEncoder::checkAvailableDataSpace(const unsigned long &numOfPixels, const CodePattern &codePattern,
              const unsigned long &amountOfEncodableData, CodePattern &newCodePattern)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgLSBsEncoder::checkAvailableDataSpace()\n");
#endif

  newCodePattern = codePattern;

  int bitsPerPixel = int(ceil((double(amountOfEncodableData)*8.0) / double(numOfPixels)));

#ifdef SMAGE_PRINT_INFOS
  cout << "Number of available pixels: " << numOfPixels << " Amount of data: " << amountOfEncodableData 
       << " Bits per pixel: " << bitsPerPixel << endl;
#endif

  if (bitsPerPixel > SMAGE_MAX_BITS_PER_PIXEL)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::checkAvailableDataSpace): amount of data too large for image space ***\n");
#endif

    return FALSE;
  }

  switch (codePattern)
  {
    case SMAGE_LSB: if (bitsPerPixel < 3) return TRUE; break;
    case SMAGE_BITS_3: if (bitsPerPixel < 4) return TRUE; break;
    case SMAGE_BITS_4: if (bitsPerPixel < 5) return TRUE; break;
    case SMAGE_LS_2B: if (bitsPerPixel < 6) return TRUE; break;
    case SMAGE_BITS_6: if (bitsPerPixel < 7) return TRUE; break;
    case SMAGE_BITS_7: if (bitsPerPixel < 8) return TRUE; break;
    case SMAGE_LS_3B: if (bitsPerPixel < 9) return TRUE; break;
    case SMAGE_BITS_9: if (bitsPerPixel < 10) return TRUE; break;
    case SMAGE_BITS_10: if (bitsPerPixel < 11) return TRUE; break;
    case SMAGE_LS_4B: if (bitsPerPixel < 12) return TRUE; break;
    case SMAGE_BITS_12: if (bitsPerPixel < 13) return TRUE; break;
    case SMAGE_BITS_13: if (bitsPerPixel < 14) return TRUE; break;
    case SMAGE_LS_5B: if (bitsPerPixel < 15) return TRUE; break;
    case SMAGE_LS_8B: if (bitsPerPixel <= SMAGE_MAX_BITS_PER_PIXEL) return TRUE; break;
    case SMAGE_AUTO: return (getSuitableCodePattern(numOfPixels, amountOfEncodableData, newCodePattern));
    default: 
    {
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("*** Warning (SmageImgLSBsEncoder::checkAvailableDataSpace): undefined code pattern ***\n");
#endif
  break;
    }
  }

#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("***  Warning (SmageImgLSBsEncoder::checkAvailableDataSpace): current code pattern is not suited ***\n");
#endif

  // Specified code pattern is not suited, try to find a better one
  return (getSuitableCodePattern(numOfPixels, amountOfEncodableData, newCodePattern));
}

// How many bits can be encoded at maximum
void SmageImgLSBsEncoder::getMaxEncodableDataAmount(const unsigned long &numOfPixels, unsigned long &maxAmountOfEncodableData)
{
  maxAmountOfEncodableData = (unsigned long)(numOfPixels * SMAGE_MAX_BITS_PER_PIXEL);
}

// If "bitCounter" equals 8, another byte is complete and counters have to be reset
void SmageImgLSBsEncoder::resetEncCounter(int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
            const unsigned long &byteStreamSize)
{
  bitCounter = 0;
  
  if (byteCounter < byteStreamSize)
  {
    byteCounter++;
    curCodeByte++;
  }
}

// Encode 2 bits
void SmageImgLSBsEncoder::code_LSB(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
           int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
           const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 255; curImgCol_Red &= 254; curImgCol_Green &= 254;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 3 bits
void SmageImgLSBsEncoder::code_BITS_3(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
				      int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
				      const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 253; curImgCol_Red &= 254; curImgCol_Green &= 254;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
 
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 4 bits
void SmageImgLSBsEncoder::code_BITS_4(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
              int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
              const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 253; curImgCol_Red &= 252; curImgCol_Green &= 254;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

//  Encode 5 bits
void SmageImgLSBsEncoder::code_LS_2B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
             const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 253; curImgCol_Red &= 252; curImgCol_Green &= 252;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 6 bits
void SmageImgLSBsEncoder::code_BITS_6(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
              int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
              const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 249; curImgCol_Red &= 252; curImgCol_Green &= 252;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;  
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 7 bits
void SmageImgLSBsEncoder::code_BITS_7(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
              int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
              const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 249; curImgCol_Red &= 248; curImgCol_Green &= 252;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;  
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}


// Encode 8 bits
void SmageImgLSBsEncoder::code_LS_3B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
             const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 249; curImgCol_Red &= 248; curImgCol_Green &= 248;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
 
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 9 bits
void SmageImgLSBsEncoder::code_BITS_9(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
              int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
              const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 241; curImgCol_Red &= 248; curImgCol_Green &= 248;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
 
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 8;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 10 bits
void SmageImgLSBsEncoder::code_BITS_10(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
               int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
               const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 241; curImgCol_Red &= 240; curImgCol_Green &= 248;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 8;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 8;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 11 bits
void SmageImgLSBsEncoder::code_LS_4B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
             const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 241; curImgCol_Red &= 240; curImgCol_Green &= 240;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 8;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 8;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 8;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 12 bits
void SmageImgLSBsEncoder::code_BITS_12(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
               int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
               const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 225; curImgCol_Red &= 240; curImgCol_Green &= 240;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 8;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 8;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize); 
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 8;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 16;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}


// Encode 13 bits
void SmageImgLSBsEncoder::code_BITS_13(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
               int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
               const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 225; curImgCol_Red &= 224; curImgCol_Green &= 240;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 8;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 16;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 8;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 8;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 16;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 14 bits
void SmageImgLSBsEncoder::code_LS_5B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
             int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
             const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 225; curImgCol_Red &= 224; curImgCol_Green &= 224;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 8;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 16;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 8;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 16;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 8;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 16;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Encode 23 bits
void SmageImgLSBsEncoder::code_LS_8B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
           int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
           const unsigned long &byteStreamSize)
{
  curImgCol_Blue &= 1; curImgCol_Red = 0; curImgCol_Green = 0;

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 8;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 16;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 32;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 64;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Red |= 128;
  bitCounter++;   
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 1;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 8;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 16;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 32;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 64;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Green |= 128;
  bitCounter++; 
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 2;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 4;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 8;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 16;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 32;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 64;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
  if ((*curCodeByte) & pot2[bitCounter]) curImgCol_Blue |= 128;
  bitCounter++;
  if (bitCounter == 8) resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}

// Simply replace a color channel by a code byte;
//  24 bits can be coded this way, but original color is completely lost
void SmageImgLSBsEncoder::code_MAX(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue,
           int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte,
           const unsigned long &byteStreamSize)
{
  curImgCol_Red = *curCodeByte;
  resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  curImgCol_Green = *curCodeByte;
  resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);

  curImgCol_Blue = *curCodeByte;
  resetEncCounter(bitCounter, byteCounter, curCodeByte, byteStreamSize);
}
