
#ifndef  _SMAGE_IMG_LSBS_ENCODER_H
#define  _SMAGE_IMG_LSBS_ENCODER_H

#include "SmageImgCoder.h"

/** Function type for encoding bits.
  */
typedef void SIE_EncodeFuncCB(int &, int &, int &, int &, unsigned long &, std::vector<char>::const_iterator &, 
            const unsigned long &);

typedef void SIE_ProgressFuncCB(void *);

/** Class SmageImgLSBsEncoder for encoding data into an image using a method that modifies up to the 5 least significant bits of each RGB color channel.
  *
  * @ingroup LibSmage
  *
  * The constructor's argument is the image into which the 
  * data shall be embedded.
  *
  * Example: 
  * --------
  * <code><pre> 
    // "img" is the image that shall be processed;
    // "idBuffer" is an ID-Buffer (image) with regions (and their 
    //  specific colors, eg. qRgb(255,0,0)) that give the areas in 
    //  "img" into which the according data shall embedded;
    SmageIdBufferStruct idBufferData;
    idBufferData.idBuffer = idBuffer;
    idBufferData.backgroundColor = QColor(COL_BACKGROUND);
    idBufferData.bufferType = SMAGE_OUTLINED_REGIONS;

    vector<SmageDataObjs> dataObjsVec; 
    SmageDataObjs dataObjs;
    SmageData dataItem;
    dataItem.dataType = SMAGE_TEXT;
    dataItem.idColor[0]=255; dataItem.idColor[1]=dataItem.idColor[2]=0;
    vector<char> byteStream; byteStream.push_back('H'); byteStream.push_back('e');
    dataItem.byteStream = byteStream;
    dataObjs.data.push_back(dataItem);
    dataObjs.codePattern = SMAGE_AUTO;
    dataObjsVec.push_back(dataObjs);
  
    SmageImgLSBsEncoder encoder(img);
    encoder.encodeData(idBufferData, dataObjsVec);
  * </pre></code>
  */
class SmageImgLSBsEncoder : public SmageImgCoder
{
  public:
    /** Constructor.
      * Argument is the image into which data shall be encoded.
      */
    SmageImgLSBsEncoder(QImage &imgData);

    /** Destructor.
      */
    ~SmageImgLSBsEncoder();

    /** Encode data into the whole image; there is no ID-Buffer necessary.
      */
    bool encodeData(const SmageDataObjs &data);

    /** Encode data into regions specified by "idBufferData".
      */
    bool encodeData(const SmageIdBufferStruct &idBufferData, const std::vector<SmageDataObjs> &data, 
		    const bool &resetImg=TRUE, const int &startId=0);

    /** Specific method of data coding (encoding an image into another image).
      * the 4 least significant bits of the cover image "_imgData" are replace by the 4 most significant 
      * bits of the image to be encoded "codeData";
      */
    void doImageDowngrading(const QImage &codeData);      

    void setProgressCB(void *userData, SIE_ProgressFuncCB *func)
      { _progressCBFunc = func; _progressUserData = userData; }
    
  private:

    /** LSB of blue color chanel is unset.
      */
    void resetImgData();

    /** Initialize data encoding.
      */
    bool initDataCoding(const int &objID, const SmageDataObjs &codeDataInfo, const SmageIdRegion &idRegion);

    /** Code a seed point that holds important information for decoding.
      * in addition, 4 surrounding pixels are coded in a way that the image can be regarded as a coded image.
      */
    void codeSeedPoint(const int &objID, const QPoint &seedPointPos, const CodePattern &codePattern);

    /** Encode the amount of data in the 4 regarding pixels ("_amountOfDataPixels").
      */
    void encodeAmountOfData(const unsigned long &numEncodedBytes);

    /** Test if the current pixel is codable.
      */
    bool isCodablePixel(const int &x, const int &y, const std::vector<SmageIdSubRegion> &subRegions);

    /** Test if the current position is a seed point position.
      */
    bool isSeedPointPos(const int &x, const int &y, const std::vector<SmageIdSubRegion> &subRegions);

    /** Start encoding the data.
      */
    void codeData(const int &objID, const SmageIdRegion &idRegion, const SmageDataObjs &codeDataInfo, 
        SIE_EncodeFuncCB *codeFunc);

    /** Verify if the specified code pattern is suited.
      */
      bool checkAvailableDataSpace(const unsigned long &numOfPixels, const CodePattern &codePattern,
           const unsigned long &amountOfEncodableData, CodePattern &newCodePattern);

    /** How many bits can be encoded at maximum.
      */
    void getMaxEncodableDataAmount(const unsigned long &numOfPixels, unsigned long &maxAmountOfEncodableData);

    /** Bit coding; different numbers of bits to be modified.
      */
    static void resetEncCounter(int &bitCounter, unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
          const unsigned long &byteStreamSize);
    static void code_LSB(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
         unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
         const unsigned long &byteStreamSize);
    static void code_BITS_3(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
            unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
            const unsigned long &byteStreamSize);
    static void code_BITS_4(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
            unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
            const unsigned long &byteStreamSize);
    static void code_LS_2B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter,
           unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
           const unsigned long &byteStreamSize);
    static void code_BITS_6(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter,
            unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
            const unsigned long &byteStreamSize);
    static void code_BITS_7(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter,
            unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
            const unsigned long &byteStreamSize);
    static void code_LS_3B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
           unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
           const unsigned long &byteStreamSize);
    static void code_BITS_9(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
            unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
            const unsigned long &byteStreamSize);
    static void code_BITS_10(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
             unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
             const unsigned long &byteStreamSize);
    static void code_LS_4B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
           unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
           const unsigned long &byteStreamSize);
    static void code_BITS_12(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
             unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
             const unsigned long &byteStreamSize);
    static void code_BITS_13(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
             unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
             const unsigned long &byteStreamSize);
    static void code_LS_5B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
           unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
           const unsigned long &byteStreamSize);
    static void code_LS_8B(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
           unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
           const unsigned long &byteStreamSize);
    static void code_MAX(int &curImgCol_Red, int &curImgCol_Green, int &curImgCol_Blue, int &bitCounter, 
         unsigned long &byteCounter, std::vector<char>::const_iterator &curCodeByte, 
         const unsigned long &byteStreamSize);

    SIE_ProgressFuncCB *_progressCBFunc; 
    void *_progressUserData;
};

#endif
