#include "SmageImgProcessor.h"
#include "SmageSettings.h"

#include <iostream>
#include <ctime>

using namespace std;

// Structure's constructor
SmageImageObjs::SmageImageObjs()
{
  origImg = QImage();
  tableIndices.clear();
  colorTable.clear();
}

// Constructor
SmageImgProcessor::SmageImgProcessor()
{
  _origImg = QImage();
}

//  Destructor
SmageImgProcessor::~SmageImgProcessor()
{
}

// Get a vector of "SmageImageObjs" according to "idBufferData";
//  one "origImg" is specified;
void SmageImgProcessor::getIndexedImgData(const QImage &origImg, vector<SmageImageObjs> &imageObjs, 
            const SmageIdBufferStruct &idBufferData)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgProcessor::getIndexedImgData()\n");
#endif

  if (imageObjs.empty())
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgProcessor::getIndexedImgData): no image objects specified ***\n");
#endif
    return;
  }

  _origImg = origImg;
  getData(imageObjs, idBufferData);
}

// Get a vector of "SmageImageObjs" according to "idBufferData";
//  for each region in ID-Buffer a single original image is specified
void SmageImgProcessor::getIndexedImgData(vector<SmageImageObjs> &imageObjs, 
            const SmageIdBufferStruct &idBufferData)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgProcessor::getIndexedImgData()\n");
#endif

  if (imageObjs.empty())
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgProcessor::getIndexedImgData): no image objects specified ***\n");
#endif
    return;
  }

  _origImg = QImage();
  getData(imageObjs, idBufferData);
}

// Do calculation.
void SmageImgProcessor::getData(vector<SmageImageObjs> &imageObjs, const SmageIdBufferStruct &idBufferData)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgProcessor::getData()\n");
#endif

  // For each element of "imageObjs", the color table and its indices shall be filled;
  // the according regions in "_origImg" or "SmageImageObjs::origImg" are found by using "idBufferData";
  // the "imageObjs" elements refer to the ID-Buffer regions through "SmageImageObjs.idColor";
  // if no ID-Buffer is set, the whole image is processed

  // (1) Assign ID and colors from "imageObjs" to ID-Buffer class
  // (2) Process the ID-Buffer, since it may contain filled or outlined regions
  //     (outlined regions need to be filled and filled regions need to be outlined (since the 
  //      outline won't be considered))
  // (3) For each region in ID-Buffer that is refered to in "imageObjs", do an image
  //     quantization in "origImg" and get the indices and color tables

  SmageIdBuffer smageIdBuffer;

  if (!_origImg.isNull())
    smageIdBuffer = SmageIdBuffer(_origImg.width(), _origImg.height(), (idBufferData.idBuffer.isNull()));
  else
  {
    if (!imageObjs[0].origImg.isNull())
      smageIdBuffer = SmageIdBuffer(imageObjs[0].origImg.width(), imageObjs[0].origImg.height(), 
              (idBufferData.idBuffer.isNull()));
    else
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageImgProcessor::getData): no original image specified ***\n");
#endif
      return;
    }
  }
  
  QImage idBuffer;
  QImage imgCopy;

  // (1) Extract ID-Colors and assign them to ID-Buffer class
  unsigned int id = 0;
  vector<SmageImageObjs>::const_iterator imgObjVecIter;
  for (imgObjVecIter=imageObjs.begin(); imgObjVecIter!=imageObjs.end(); imgObjVecIter++)
  {
    SmageIdColorStruct colorStruct;
    if (idBufferData.bufferType == SMAGE_OUTLINES)
    {
      colorStruct.outlineColor = qRgb(imgObjVecIter->idColor.red(), imgObjVecIter->idColor.green(), 
            imgObjVecIter->idColor.blue());

#ifdef SMAGE_PRINT_INFOS
      cout << "Object with ID: " << id << " has ID color: " 
           << qRed(colorStruct.outlineColor) << " " << qGreen(colorStruct.outlineColor) << " " 
           << qBlue(colorStruct.outlineColor) << endl;
#endif
    }
    else if (idBufferData.bufferType == SMAGE_FILLED_REGIONS || idBufferData.bufferType == SMAGE_OUTLINED_REGIONS)
    {
      colorStruct.fillColor = qRgb(imgObjVecIter->idColor.red(), imgObjVecIter->idColor.green(), 
               imgObjVecIter->idColor.blue());

#ifdef SMAGE_PRINT_INFOS
      cout << "Object with ID: " << id << " has ID color: " 
           << qRed(colorStruct.fillColor) << " " << qGreen(colorStruct.fillColor) << " " 
           << qBlue(colorStruct.fillColor) << endl;
#endif
    }

    colorStruct.subID = 0;
    smageIdBuffer.insertIdColorsValue(id, colorStruct);
    id++;
  }

  // (2) Process the ID-Buffer
  if (!idBufferData.idBuffer.isNull())
  {
    smageIdBuffer.setIdBuffer(idBufferData.idBuffer, idBufferData.backgroundColor, idBufferData.bufferType);
      
    if (idBufferData.bufferType == SMAGE_OUTLINES)  
      smageIdBuffer.fillOutlineRegions();
    else if (idBufferData.bufferType == SMAGE_FILLED_REGIONS)
      smageIdBuffer.generateOutlines();
    else if (idBufferData.bufferType == SMAGE_OUTLINED_REGIONS)
      smageIdBuffer.analyzeIdBuffer();
  }

  QImage origImg;

  if (!_origImg.isNull())
    origImg = _origImg;

  // (3) Generate color tables and assign color indices
  smageIdBuffer.getIdBuffer(idBuffer);
  vector<SmageImageObjs>::iterator imageObjsIter;
  for (imageObjsIter=imageObjs.begin(); imageObjsIter!=imageObjs.end(); imageObjsIter++)
  {
    if (_origImg.isNull())
    {
      if (!imageObjsIter->origImg.isNull())
        origImg = imageObjsIter->origImg;
      else
      {
#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("*** Warning (SmageImgProcessor::getData): no original image specified ***\n");
#endif
      }
    }

    int id, subId, fillColorIndex, outlineColorIndex, currentColorIndex;
    if (smageIdBuffer.getObjectId(imageObjsIter->idColor, id, subId))
    {
      if (smageIdBuffer.getFillColorIndices(id, subId, fillColorIndex, outlineColorIndex))
      {
        if (origImg.depth() > 8)
        {
          // Copy the original image and fill colors outside current id buffer areas with 
          // background color to get a more detailed color table
          imgCopy = origImg.copy();
          QColor backgroundCol(COL_BACKGROUND);

          for (int y=0; y<imgCopy.height(); y++)
            for (int x=0; x<imgCopy.width(); x++)
            {
              currentColorIndex = idBuffer.pixelIndex(x, y);
              if (currentColorIndex != fillColorIndex)
                imgCopy.setPixel(x, y, backgroundCol.rgb());
            }

          // Image quantization
          //imgCopy = imgCopy.convertDepth(8, Qt::ThresholdDither);  
	  imgCopy = imgCopy.convertDepth(8);
        }
        else
        {      
          imgCopy = origImg.copy();
        }

        for (int y=0; y<imgCopy.height(); y++)
          for (int x=0; x<imgCopy.width(); x++)
          {
            // Get the color indices
            currentColorIndex = idBuffer.pixelIndex(x, y);

            if (currentColorIndex == fillColorIndex)
              imageObjsIter->tableIndices.push_back(char(imgCopy.pixelIndex(x, y) & 255));
          }

        copyColorTable(imgCopy.colorTable(), imageObjsIter->colorTable, imgCopy.numColors());
      }
      else
      {
#ifdef SMAGE_PRINT_DEBUG
        _SmageLogFile.lprintf("*** Warning (SmageImgProcessor::getIndexedImgData): could not get color indices ***\n");
#endif
      }
    }
    else
    {
#ifdef SMAGE_PRINT_DEBUG
      _SmageLogFile.lprintf("*** Warning (SmageImgProcessor::getIndexedImgData): could not get an object ID ***\n");
#endif
    }
  }
}

// Copy color table's field values to a vector
void SmageImgProcessor::copyColorTable(const QRgb *srcTable, vector<QRgb> &destTable, const int &size)
{
  const QRgb *srcTablePointer = srcTable;

  for (int i=0; i<size; i++)
  {
    destTable.push_back(*srcTablePointer);
    srcTablePointer++;
  }
}

// Fill an image with randomized colors
void SmageImgProcessor::fillImgRandomized(QImage &img)
{
#ifdef SMAGE_PRINT_DEBUG
  _SmageLogFile.lprintf("SmageImgProcessor::fillImgRandomized()\n");
#endif

  if (img.depth() < 24)
  {
#ifdef SMAGE_PRINT_DEBUG
    _SmageLogFile.lprintf("*** Warning (SmageImgProcessor::fillImgRandomized): image to be randomizely filled is not an RGB ***\n");
#endif
    return;
  }

  srand(static_cast<unsigned>(time(0)));
  int randomInt_R, randomInt_G, randomInt_B;

  for (int y=0; y<img.height(); y++)
    for (int x=0; x<img.width(); x++)
    {
      randomInt_R = int(256.0 * rand()/(RAND_MAX+1.0));
      randomInt_G = int(256.0 * rand()/(RAND_MAX+1.0));
      randomInt_B = int(256.0 * rand()/(RAND_MAX+1.0));
  
      img.setPixel(x, y, qRgb(randomInt_R, randomInt_G, randomInt_B));
    }
}
