#ifndef  _SMAGE_IMG_PROCESSOR_H
#define  _SMAGE_IMG_PROCESSOR_H

#include "SmageIdBuffer.h"

/** Structure of image objects.
  * According to the colors in an ID-Buffer, for each ID-Buffer region  a color table and its
  * indices are determined from the original image data (after an image quantization for each
  * such region has been done);
  * only "idColor" has to be set externally
  *
  * @ingroup LibSmage
  *
  */
struct SmageImageObjs {
  SmageImageObjs();                // Constructor
  QImage origImg;                  // Image that shall be encoded
  QColor idColor;                  // ID-Color, reference to the ID-Buffer
  std::vector<char> tableIndices;  // Color indices
  std::vector<QRgb> colorTable;    // Color table
};

/** Class SmageImgProcessor does the processing of images other than the ID-Buffer.
  *
  * @ingroup LibSmage
  *
  * The following image operations are included:<br>
  * - Image quantization (reducing the colors to get an indexed image)<br>
  * Example (Image quantization):<br>
  * <code><pre> 
    //Assuming, there is an ID-Buffer with "COL_BACKGROUND" as
    //   background color and a color region (255,0,0) that we are
    //   interested in within the original image "imgData"

    SmageIdBufferStruct idBufferData;
    idBufferData.idBuffer = idBuffer;
    idBufferData.backgroundColor = QColor(COL_BACKGROUND);
    idBufferData.bufferType = SMAGE_FILLED_REGIONS;

    std::vector<SmageImageObjs> imgObjects;
    SmageImageObjs imgObj;
    imgObj.idColor = QColor(255,0,0);
    imgObjects.push_back(imgObj);
  
    SmageImgProcessor imgProcessor;
    imgProcessor.getIndexedImgData(imgData, imgObjects, idBufferData);
  * </pre></code>
  */
class SmageImgProcessor
{
  public:
    /** Constructor.
      */
    SmageImgProcessor();
    
    /** Destructor.
      */
    ~SmageImgProcessor();

    /** Gets a vector of "SmageImageObjs" according to "idBufferData".
      * Only one original image.
      */
    void getIndexedImgData(const QImage &origImg, std::vector<SmageImageObjs> &imageObjs, const SmageIdBufferStruct &idBufferData);
    
    /** Gets a vector of "SmageImageObjs" according to "idBufferData".
      * Each region in ID-Buffer has its own image.
      */
    void getIndexedImgData(std::vector<SmageImageObjs> &imageObjs, const SmageIdBufferStruct &idBufferData);

    /** Fills an image with randomized colors.
      */
    void fillImgRandomized(QImage &img);

  private:

    /** Does calculation.
      */
    void getData(std::vector<SmageImageObjs> &imageObjs, const SmageIdBufferStruct &idBufferData);

    /** Copies color table's field values to a vector.
      */
    void copyColorTable(const QRgb *srcTable, std::vector<QRgb> &destTable, const int &size);

    /** Original image.
      * Taken for all ID-Buffer regions.
      */
    QImage _origImg;
};

#endif
