// project header
#include "NPSmagePluginInstance.h"

#include <qcursor.h>

NPSmagePluginInstance::NPSmagePluginInstance() : QNPInstance()
{
  _imgStreamCreated = FALSE;
}

NPSmagePluginInstance::~NPSmagePluginInstance()
{
}

QNPWidget*
NPSmagePluginInstance::newWindow()
{
  _npspw = new NPSmagePluginWidget();
  
  // get plugin parameters
  const char *imgUrl = arg( "smage" );
  const char *srcUrl = arg( "src" );
  _width = atoi( arg( "width" ) );
  _height = atoi( arg( "height" ) );
  _refUrl = arg( "href" );
  _target = arg( "target" );
  if ( _target == 0 )
  {
    _target = "_BLANK"; 
  }

  if ( imgUrl == 0 ) _imgUrl = 0;
  else
  {
    // treat urls like " " or "  " same as url ""
    // 'img'-url
    bool allBlanks = TRUE;
    int len;
    int i;

    len = strlen( imgUrl );
    for ( i = 0; i < len; i++ )
    {
      if ( isspace( imgUrl[i] ) == 0 )
      {
        allBlanks = FALSE;
        break;
      }
    }
    if ( allBlanks ) _imgUrl = "";
    else _imgUrl = imgUrl;
  }

  // get data from an url
  // --------------------
  //
  // There are two ways to get the image from the url:
  // getURLNotify( url ) and getURL( url ).
  // In the unmodified qnp library source both functions do not return any status code,
  // although the nepscape functions do. So the first thing is to modify the qnp source,
  // to return this status as well.
  //
  // The idea is, that the functions return 0, if no error occured, and another value, if
  // an error occured (e.g. the url is invalid). Netscape/Mozilla seem to handle this correctly,
  // but Opera always returns 0.
  //
  // If getURLNotify( url ) is used, the notifyURL function is called at the end of the
  // operations invoked by getURLNotify. In the case of an error, notifyURL is not called
  // by Netscape/Mozilla - but by Opera. And this time opera indicated the error :)
  // (In the case of no error all browser seem to call it).
  //
  // So this weird behaviour leads to the following solution:
  // If getURLNotify returns an errror (only Netscape/Mozilla) notifyURL is called by the 
  // code below. Opera will call this function itself, indicating an error.
  // 
  // But the unmodified qnp library source has no getURLNotify for unix/linux, so again it is 
  // time to modify the qnp and to replace the npunix.c file with this one:
  // http://www.damtp.cam.ac.uk/user/kp229/ndvi/ndvi/ndvi/npunix.c
  //
  // below is the source code for the modified getURLNotify function :
  //
  // int QNPInstance::getURLNotify( const char* url, const char* window, void *data )
  // {
  //   return NPN_GetURLNotify( pi->npp, url, window, data );
  // }
  //
  //
  //
  // 'src' vs. 'smage' (or any other) parameter in the embed tag
  // -----------------------------------------------------------
  // 
  // If a 'src' parameter is specified, the browser automatically trys to create a stream and send
  // the data of the url specified by the tag to the plugin. The problem is, that no notification 
  // about errors is given. Netscape/Mozilla even deletes the plugin if no stream can be created
  // ( == if an invalid url is given). (Firefox and Opera don't delete the plugin.)
  // This is a huge flaw in the plugin API.
  //
  // Using another parameter allows us to get the data ourselves. By this we can check for errors,
  // display status and progress information.
  // Nevertheless, if someone specifies a 'src' tag, the problems described above still apply.
  //
  // In the case that the 'src' parameter is specified there is absolutely no guarantee regarding
  // the order in which functions are called. Sometimes the browser will call newWindow() first, at
  // other times, newStreamCreated() will be called first.
  //
  // The flag _imgStreamCreated solves this problem. At initialization the flag is set to FALSE.
  // newStreamCreated() tests this flag. If newStreamCreated() is called first ('src' parameter),
  // the stream is rejected.
  // newWindow() checks the plugin parameters. If there is a 'src' parameter the plugin prints a
  // status message that the 'src' parameter is not allowed. Otherwise _imgStreamCreated is set to 
  // TRUE, and a stream to the url specified in 'smage' is created.
  //
  // The plugin doesn't know the url it is creted from.
  // In the case of relative urls in the 'src' and 'smage' parameters, it is not always possible to
  // distinguish between them in newStreamCreated(), as stream->url() returns the absolute url.
  //
  // a) http://smage.de/test/test/img.png
  // b) http://smage.de/test/img.png
  // 
  // http://smage.de/test/plugin.html
  // <embed src="img.png" smage="test/img.png" ...>
  //
  // 'src' -> b)
  // 'smage' -> a)
  //
  // Testing the relative url of the parameter against the absolute url of the stream is not 
  // definite. So just don't allow the use of the 'src' parameter.
  
  
  if ( srcUrl == 0  )
  {
    if ( _imgUrl != 0  )
    {
      if ( strcmp( _imgUrl, "" ) != 0 )
      {
        // print some feedback
        QString s = " Smage plugin >> try loading: ";
        s.append( _imgUrl );
        _npspw->_status1->setText( s );

        // this allows newStreamCreated() to process a stream
        _imgStreamCreated = TRUE;
        // try to get the data
        int error = getURLNotify( _imgUrl );
        // if an error occured, e.g. invalid img url
        if ( error != 0 )
        {
          // don't allow the creation of other streams
          _imgStreamCreated = FALSE;
          notifyURL( 0, ReasonError, 0 );
        }
      }
      else
      {
        // no url specified in the smage parameter
        // examples for this are:
        // <embed smage width="10" height="10">
        // <embed smage="" width="10" height="10">
        // in these cases the plugin would load the url of the page where it is embeded.
        _npspw->_status1->setText( " No image url specified in parameter 'smage'! " );
      }
    }
    else
    {
      // no smage parameter
      _npspw->_status1->setText( " No parameter 'smage' specified! " );
    }
  }
  else
  {
    _npspw->_status1->setText( " The parameter 'src' is not allowed! " );
  }
  return _npspw;
}

bool
NPSmagePluginInstance::newStreamCreated( QNPStream *stream, StreamMode& smode )
{
  // allow only streams that are created from the 'smage' parameter
  if ( !_imgStreamCreated ) return FALSE;

  // print some feedback
  QString s = " Smage plugin >> try loading: ";
  s.append( stream->url() );
  _npspw->_status1->setText( s );
  _npspw->_status2->setText( " Smage plugin >> receiving:  " );

  // set the streammode
  smode = Normal;
  return TRUE;
}

void
NPSmagePluginInstance::notifyURL( const char *url, Reason r, void *notifyData )
{
  if ( r != ReasonDone )
  {
    _npspw->_spw->hide();
    _npspw->_progress->hide();
    _npspw->_layout->setColStretch ( 0, 100 );
    _npspw->_layout->setColStretch ( 1, 0 );
    _npspw->_status2->setText( " Smage plugin >> an error occured, while loading the image! Propably the image link is broken." );
    _npspw->_error = TRUE;
    if ( _refUrl ) _npspw->setCursor( QCursor(PointingHandCursor) );

  }
  _npspw->update();
}

void
NPSmagePluginInstance::streamDestroyed( QNPStream *stream )
{
  if ( stream->okay() )
  {
    _npspw->_status1->hide();
    _npspw->_status2->hide();
    _npspw->_progress->hide();
    _npspw->_layout->setRowStretch ( 1, 0 );
    _npspw->_layout->setRowStretch ( 2, 0 );

    _img = QImage( _data );
    _npspw->_spw->resize( _width, _height );
    _npspw->_spw->setFileSize( _data.size() );
    _npspw->_spw->setScale(  _npspw->_spw->size(), _img.size() );
    _npspw->_spw->setImage( _img, stream->url() );
    _npspw->_spw->repaint( true );
  }

  status( "" );
}

int 
NPSmagePluginInstance::writeReady( QNPStream* )
{
  return 64000;
}

int
NPSmagePluginInstance::write( QNPStream* stream, int offset, int len, void *buffer )
{
  char* data = (char*)buffer;
  int old_size = _data.size();
  _data.resize( old_size + len );
  for ( int i = 0; i < len; i++ ) _data[old_size+i]=data[i];

  // print some feedback
  _npspw->_prog += 10;
  _npspw->_progress->setProgress( _npspw->_prog );

  char sbuf[64];
  float s = _data.size() / 1024.0;
  sprintf( sbuf, "%.2fKB", s ); 

  QString dlSize( sbuf );
  QFontMetrics fm = _npspw->_progress->fontMetrics();
  int tx = ( _npspw->_progress->width() - fm.width( dlSize ) ) / 2;
  int ty = _npspw->_progress->height() - (int)( ( _npspw->_progress->height() - fm.ascent() ) / 2.0 ) - 1;
  _npspw->_progress->drawText( tx, ty, dlSize );

//  _npspw->repaint( TRUE );

  return len;
}

void
NPSmagePluginInstance::visitLinkLocation()
{
  if ( _refUrl ) getURL( _refUrl, _target );
}

bool
NPSmagePluginInstance::hasRefURL()
{
  if ( _refUrl ) return true;
  else return false;
}

const char*
NPSmagePluginInstance::getAbsRefURL()
{
  return _refUrl;
}
