// project header
#include "SmageDecodeHandler.h"
#include "SmageTextWidget.h"
#include "SmageViewerShapedText.h"
#include "SmageViewerSimpleBubble.h"
#include "SmageViewerSpeechBubble.h"
#include "SmageViewerThinkBubble.h"
#include "SmageXmlParser.h"

// qt header
#include <qbitmap.h>
#include <qbuttongroup.h>
#include <qcheckbox.h>
#include <qpainter.h>
#include <qpopupmenu.h>
#include <qradiobutton.h>

SmageStructDecodeInfo
SmageDecodeHandler::decode( QImage image, QImage &refIDBuffer, QImage &outlineImg, std::map<unsigned int, SmageStructDecodedData> &decodedDataMap, 
          SmageViewerWidget *parent )
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageDecodeHandler::decode()\n");
#endif

  SmageStructDecodeInfo info;

  std::map<unsigned int, SmageDataObjs> decodeMap;
  SmageImgLSBsDecoder imgDecoder( image );

  QPixmap outlines;
  imgDecoder.decodeData( refIDBuffer, decodeMap, outlines );

  // Generate a copy of the original image with the outlines embedded
  QPixmap outlinePixmap( image );
  bitBlt( &outlinePixmap, 0,0, &outlines, 0,0, outlines.width(), outlines.height(), Qt::CopyROP, FALSE );
  outlineImg = outlinePixmap.convertToImage();

  std::map<unsigned int, SmageDataObjs>::const_iterator decodeMapIter;
  std::vector<SmageData>::const_iterator smageDataIter;
  std::vector<char>::const_iterator charVecIter;
  for ( decodeMapIter=decodeMap.begin(); decodeMapIter!=decodeMap.end(); decodeMapIter++ )
  {
    info._numOfEncodedAreas++;
    SmageStructDecodedData dataItem;
    // "smageDataIter" points to each data item (for each region in the ID-Buffer)
    for ( smageDataIter=decodeMapIter->second.data.begin(); smageDataIter!=decodeMapIter->second.data.end(); smageDataIter++ )
    {
      // Encoded text
      if ( smageDataIter->dataType == SMAGE_TEXT )
      {
        QString text = "";
        info._numOfEncodedTexts++;
        for ( charVecIter=smageDataIter->byteStream.begin(); charVecIter!=smageDataIter->byteStream.end(); charVecIter++ )
          text += *charVecIter;
        SmageTextWidget *stw = new SmageTextWidget( parent );
        stw->setText( text );
        stw->hide();
        dataItem._tw = stw;
      }
      // Encoded image
      if ( smageDataIter->dataType == SMAGE_IMG )
      {
        // 1. Initialize 8 bit image and bitmap
        // 2. Assign color table to 8 bit image
        // 3. Fill 8 bit image and bitmap
        // 4. Convert the 8 bit image to a pixmap and set its bitmask
  
        // (1) Initializing
        info._numOfEncodedImages++;
        QImage tmpImg( image.width(), image.height(), 8 );
        QBitmap bitmap( image.width(), image.height(), TRUE );
        std::vector<char>::const_iterator byteStreamIter = smageDataIter->byteStream.begin();
        QPainter painter( &bitmap );
        painter.setPen( Qt::color1 );

        // (2) Setting the color table
        std::vector<QRgb>::const_iterator colorTableIter;
        int counter = 0;
        for ( colorTableIter=decodeMapIter->second.colorTable.begin(); colorTableIter!=decodeMapIter->second.colorTable.end(); colorTableIter++ )
        {
          tmpImg.setNumColors( tmpImg.numColors()+1 );
          tmpImg.setColor( counter, *colorTableIter );
          counter++;
        }

        // (3) Filling the image and bitmap
        for ( int y=0; y<refIDBuffer.height(); y++ )
          for ( int x=0; x<refIDBuffer.width(); x++ )
          {
            QRgb currentRgb = refIDBuffer.colorTable()[refIDBuffer.pixelIndex(x, y)];
            if ( qBlue(currentRgb ) == int(decodeMapIter->first) && qRed(currentRgb) == 255 )
            {
              tmpImg.setPixel(x,y, int(*byteStreamIter & 255));
              byteStreamIter++;      
              painter.drawPoint(x,y);
            }
            else tmpImg.setPixel( x,y, int(smageDataIter->byteStream[0] & 255) );
          }

        // (4) Converting the image
        if ( dataItem._pixmap.convertFromImage(tmpImg) ) dataItem._pixmap.setMask( bitmap );

      }
      // Encoded formatted text
      if ( smageDataIter->dataType == SMAGE_FORM_TEXT )
      {  
        SmageObjShape objShape;
        std::vector<SmageTextShapes> shapesVec;
        objShape.getTextShapes(smageDataIter->byteStream, shapesVec);

        for (unsigned int i=0; i<shapesVec.size(); i++)
        {
          TextShapeType shapeType = shapesVec[i].shapeType;
          info._numOfEncodedFormattedTexts++;
          if (shapeType == SMAGE_SPEECH_BUBBLE || shapeType == SMAGE_THINK_BUBBLE)
          {
            SmageComicEllipseShape comicShape;
            comicShape.getComicEllipseShapeObjects(shapesVec[i].shapeString);
    
            std::vector<SmageComicEllipseObj *> ellipseObjVec = comicShape.getObjVec();
            std::vector<SmageComicEllipseObj *>::const_iterator ellipseObjVecIter;
    
            for (ellipseObjVecIter=ellipseObjVec.begin(); ellipseObjVecIter!=ellipseObjVec.end(); ellipseObjVecIter++)
            {
              if ((*ellipseObjVecIter)->ellipseObjData.textObjData.shapeType == SMAGE_SPEECH_BUBBLE)
              {
                 SmageViewerSpeechBubble *bubble = new SmageViewerSpeechBubble( parent, "SmageEditorSpeechBubble", **ellipseObjVecIter );
                 bubble->hide();
                 dataItem._shapedTextVec.push_back( bubble );
              }
              else if ((*ellipseObjVecIter)->ellipseObjData.textObjData.shapeType == SMAGE_THINK_BUBBLE)
              {
                SmageViewerThinkBubble *bubble = new SmageViewerThinkBubble( parent, "SmageEditorThinkBubble", **ellipseObjVecIter );
                bubble->hide();
                dataItem._shapedTextVec.push_back( bubble );
              }
            }
          }
          else if (shapeType == SMAGE_RECT || shapeType == SMAGE_ELLIPSE)
          {
            SmageEllipseShape ellipseShape;
            ellipseShape.getEllipseShapeObjects(shapesVec[i].shapeString);

            std::vector<SmageEllipseObj *> ellipseObjVec = ellipseShape.getObjVec();
            std::vector<SmageEllipseObj *>::const_iterator ellipseObjVecIter;

            for (ellipseObjVecIter=ellipseObjVec.begin(); ellipseObjVecIter!=ellipseObjVec.end(); ellipseObjVecIter++)
            {
              SmageViewerSimpleBubble *bubble = new SmageViewerSimpleBubble( parent, "SmageViewerSimpleBubble", **ellipseObjVecIter );
              bubble->hide();
              dataItem._shapedTextVec.push_back( bubble );
            }
          }
        }
      }
      
      // Encoded dictionary entry
      if ( smageDataIter->dataType == SMAGE_DICT )
      {
        QString text = "";
        for ( charVecIter=smageDataIter->byteStream.begin(); charVecIter!=smageDataIter->byteStream.end(); charVecIter++ )
          text += *charVecIter;

        DictEntriesPair dictEntries;
        QString codedTerm;
        if (getDictEntries(text, codedTerm, dictEntries))
        {
          info._numOfEncodedDictEntries++;
          SmageTextWidget *stw = new SmageTextWidget( parent );
          stw->setDictEntries(dictEntries, codedTerm);
          stw->setDictText(3, 0, FALSE, FALSE);
          stw->hide();
          dataItem._tw = stw;
        }
      }
    }

    // Store the current data item
    decodedDataMap[decodeMapIter->first] = dataItem;
  }
  return info;
}


bool SmageDecodeHandler::getDictEntries(const QString &text, QString &codedTerm, DictEntriesPair &dictEntries)
{
  // Get information from a string like "<_ID 2><tree>Baum"
  // "idNum" = 2; "idString" = tree; "codedTerm" = Baum

  int idNum = -1;
  QString idString;
  int pos;

  if (text.find("<_ID") != -1)
  {
    pos = text.find('>');
    if (pos != -1)
    {
      QString idNumString = text.mid(5, pos-5);
      idNum = idNumString.toInt();
    }
  }

  codedTerm = text.section('>', -1);
  idString = text.section('<', -1);
  pos = idString.find('>');
  if (pos != -1)
    idString = idString.left(pos);

  SmageXmlParser xmlParser;
  if (!xmlParser.searchForEntries("smageDict.xml", dictEntries, idString, idNum))
    return (xmlParser.searchForEntries("smageDictPriv.xml", dictEntries, idString, idNum));

  return TRUE;
}

QPopupMenu*
SmageDecodeHandler::createDictMenu( std::vector<TheS_Language> *languages, QWidget *parent, QButtonGroup *&fromGroup, QButtonGroup *&toGroup)
{
  QPopupMenu *dictMenu = new QPopupMenu(parent);
  Q_CHECK_PTR(dictMenu);

  QPopupMenu *translFrom = new QPopupMenu(dictMenu);
  fromGroup = new QButtonGroup(translFrom);
  fromGroup->hide();
  fromGroup->setRadioButtonExclusive(TRUE);
  QObject::connect(fromGroup, SIGNAL(clicked(int)), parent, SLOT(translFrom(int)));

  QPopupMenu *translTo = new QPopupMenu(dictMenu);
  toGroup = new QButtonGroup(translTo);
  toGroup->hide();
  toGroup->setRadioButtonExclusive(TRUE);
  QObject::connect(toGroup, SIGNAL(clicked(int)), parent, SLOT(translTo(int)));

  std::vector<TheS_Language>::iterator langIter;
  int i;
  for ( langIter=languages->begin(), i=0; langIter != languages->end(); ++langIter, ++i)
  {
    QRadioButton *from = new QRadioButton((*langIter)._name, translFrom);
    translFrom->insertItem(from);
    fromGroup->insert(from, i);

    QRadioButton *to = new QRadioButton((*langIter)._name, translTo);
    translTo->insertItem(to);
    toGroup->insert(to, i);
  }
  fromGroup->setButton(3);
  dictMenu->insertItem( QObject::tr("Translate from ..."), translFrom);

  toGroup->setButton(0);
  dictMenu->insertItem( QObject::tr("Translate to ..."), translTo);

  dictMenu->insertSeparator();

  dictMenu->insertItem( QObject::tr("Show coded word"), parent, SLOT( showCodedDictItem() ), 0, 3 );
  dictMenu->insertItem( QObject::tr("Show ambiguities"), parent, SLOT( showSecEntries() ), 0, 4 );

  return dictMenu;
}
