// project header
#include "SmageImgWidget.h"

// qt header
#include <qmessagebox.h>

// c/c++ header
#include <assert.h>
#include <iostream>

using namespace std;

SmageImgWidget::SmageImgWidget(QWidget *parent, const char* name, WFlags f)
  : QWidget(parent, name, f)
{ 
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgWidget::SmageImgWidget()\n");
#endif

  setFocusPolicy(QWidget::ClickFocus);

  _imgShouldBeDecoded = FALSE;
  _painter = new QPainter;

  _imgLoaded = _isRgbImg = _isCodedImg = _allowRegionSelection = _imgHasBeenModified = FALSE;
  _sceneChanged = TRUE;
  _img32BitsNeeded = TRUE;
  _imgStartPos = QPoint(0,0);
  _resizeDiff = QSize(0,0);

  // Initialize callbacks
  _showMsgCBFunc = NULL;
  _showMsgUserData = NULL;
  _closeCBFunc = NULL;
  _closeUserData = NULL;
  _saveCBFunc = NULL;
  _saveUserData = NULL;
  _forceOpenNewCBFunc = NULL;
  _forceOpenNewUserData = NULL;

  _smageGenerator = new SmageGenerator();

  _img = QImage();
  _imgPointer = NULL;
  _imgFileName = QString("");
}


SmageImgWidget::~SmageImgWidget()
{ 
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgWidget::~SmageImgWidget()\n");
#endif

  delete _painter;
  delete _smageGenerator;
}

// What should be done when refreshing the widget?
//  ... do be implemented in the derived classes
void SmageImgWidget::paintEvent(QPaintEvent *)
{
  if (_imgPointer != NULL)
    _painter->drawImage(_imgStartPos, *_imgPointer);
}

// Extracts only the file name of a string (no path);
//  e.g., /home/user/img.png => img.png
void SmageImgWidget::extractFileName(const QString &entireFileName, QString &fileName)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgWidget::extractFileName()\n");
#endif

  fileName = QString("");

  QChar sign;
#ifdef SMAGE_UNIX
  sign = QChar('/');
#else
  sign = QChar('\\');
#endif

  uint length = entireFileName.length();
  int index = entireFileName.findRev(sign);
  if (index != -1)
  {
    while (uint(index) < length-1)
      fileName += entireFileName[++index];
  }
  else
    fileName += entireFileName;
}

// Open an image (its location is specified by "fileName") and possibly decode
//  it if encoded with data
bool SmageImgWidget::loadImg(const char *fileName)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgWidget::loadImg()\n");
#endif

  QString warningMsg, captionMsg;
  warningMsg = QString(tr("The image could not be loaded."));
  captionMsg = QString(QObject::tr("Smage - Message"));

  _imgLoaded = _img.load(fileName);
  if (_imgLoaded)
  {
    // Set the widgets minimum size and its caption
    QSize minimum(100,100);
    if (minimum.width() > _img.size().width())
      minimum.setWidth(_img.size().width());
    if (minimum.height() > _img.size().height())
      minimum.setHeight(_img.size().height());

    setMinimumSize(minimum);
    //setFixedSize(_img.size());
    QString entireFileName(fileName);
    extractFileName(entireFileName, _imgFileName);

    _imgPointer = &_img;

    // If the image's depth is less than 25, convert its depth to 32
    if (_img.depth() <= 24 && _img32BitsNeeded)
    {
      _img = _img.convertDepth(32);
    }
      
    if (!_img.isNull())
    {
      _isRgbImg = TRUE;
      
      // Is the image coded with data? If so, decode it
      if (_imgShouldBeDecoded)
      {
        _isCodedImg = _smageGenerator->isCodedImg(_img);      
        if (_isCodedImg)
        {  
          QString msg;
          msg = QString(tr("The image has coded data."));
          if (_showMsgCBFunc != NULL)
            (*_showMsgCBFunc)(_showMsgUserData, msg, 10000);
        }
      }

      doWhenSuccesImgLoading();
    }
    else
      QMessageBox::warning(this, captionMsg, warningMsg, QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
  }
  else
    QMessageBox::warning(this, captionMsg, warningMsg, QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

  return _imgLoaded;
}

// When resizing the widget:
// the image start position has to be modified (to be centered)
void SmageImgWidget::resizeEvent(QResizeEvent *)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgWidget::resizeEvent()\n");
#endif

  QPoint tmpPos(_imgStartPos);

  _imgStartPos.setX(int(size().width()/2) - int(_img.size().width()/2));
  _imgStartPos.setY(int(size().height()/2) - int(_img.size().height()/2));

  _resizeDiff.setWidth(_imgStartPos.x() - tmpPos.x());
  _resizeDiff.setHeight(_imgStartPos.y() - tmpPos.y());

  _sceneChanged = TRUE;
}

// Get a position within the image (eg., if position is in widget but not in image)
void SmageImgWidget::getImgLikePos(const QPoint &curPos, QPoint &newPos)
{
  if (curPos.x() < _imgStartPos.x())
    newPos.setX(_imgStartPos.x());
  else if (curPos.x() > _imgStartPos.x() + _img.width()-1)
    newPos.setX(_imgStartPos.x() + _img.width()-1);
  else
    newPos.setX(curPos.x());

  if (curPos.y() < _imgStartPos.y())
    newPos.setY(_imgStartPos.y());
  else if (curPos.y() > _imgStartPos.y() + _img.height()-1)
    newPos.setY(_imgStartPos.y() + _img.height()-1);
  else
    newPos.setY(curPos.y());
}

//  Get position related to the image position
void SmageImgWidget::getLocalImgPos(const QPoint &curPos, QPoint &newPos)
{
  newPos = curPos - _imgStartPos;
}

//  Verify specified file name;
//  if file has ending, return it
QString SmageImgWidget::getFileEnding(const QString &string, bool &validEnding)
{
  // Has "string" something lile ".*"

  QString ending;

  uint length = string.length();
  int index = string.findRev('.');
  if (index != -1)
  {
    validEnding = TRUE;
    while (uint(index) < length-1)
      ending += string[++index];
  }

  else
    validEnding = FALSE;

  return ending;
}

// Save image as ...
void SmageImgWidget::saveImg(const QString &name)
{
  _img.save(name, "PNG", 100);
}

// Before closing the widget ...
void SmageImgWidget::closeEvent(QCloseEvent *event)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgWidget::closeEvent()\n");
#endif

  // Image has been modified
  if (_imgHasBeenModified)
  {
    QString s_msg;
    s_msg = QString(tr("The image has been modified.\nDo you want to save it?"));

    switch (QMessageBox::information(this, QObject::tr("Smage - Message"), s_msg, tr("Save"), QObject::tr("No"), QString::null, 0, 1))
    {
      case 0: 
      {
        if (_saveCBFunc != NULL)
          (*_saveCBFunc)(_saveUserData);
        break;
      }
    }
  }

  if (_closeCBFunc != NULL)
    (*_closeCBFunc)(_closeUserData);
  
  event->accept();
}


void SmageImgWidget::assignWidgetData(const SmageImgWidget *smageWidget)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgWidget::assignWidgetData()\n");
#endif

  setMinimumSize(smageWidget->size());

  _showMsgCBFunc = smageWidget->_showMsgCBFunc;
  _showMsgUserData = smageWidget->_showMsgUserData;
  _closeCBFunc = smageWidget->_closeCBFunc;
  _closeUserData = smageWidget->_closeUserData;
  _saveCBFunc = smageWidget->_saveCBFunc;
  _saveUserData = smageWidget->_saveUserData;
  _forceOpenNewCBFunc = smageWidget->_forceOpenNewCBFunc;
  _forceOpenNewUserData = smageWidget->_forceOpenNewUserData;

  if (smageWidget->_imgLoaded)
  {
    _img = smageWidget->_img.copy();
    _imgPointer = &_img;
    _imgLoaded = TRUE;
    _imgFileName = QString(smageWidget->_imgFileName.ascii());
    _isRgbImg = smageWidget->_isRgbImg;
    _isCodedImg = smageWidget->_isCodedImg;  
    _imgHasBeenModified = smageWidget->_imgHasBeenModified;

    doWhenSuccesImgLoading();
  }
}

QImage*
SmageImgWidget::getImgPointer()
{
  if(_imgLoaded) return _imgPointer;
  else return NULL;
}
