// project header
#include "SmageApp.h"
#include "SmageImgInputDlg.h"

// qt header
#include <qbuttongroup.h>
#include <qfiledialog.h> 
#include <qmessagebox.h>
#include <qmenubar.h>
#include <qstatusbar.h>

// c/c++ header
#include <assert.h>
#include <iostream>

using namespace std;

SmageApp::SmageApp(QWidget *topLevel, int argc, char *argv[]) : SmageAppBase(topLevel, argc, argv)
{   
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageApp::SmageApp()\n");
#endif

  if (_imgGrabber != NULL)
    _imgGrabber->setSelCopyImgCB(this, selCopyImgCB);

  _imgWidgetCounter = 0;
  
  _showCodedDictItem = FALSE;
  _showSecEntries = FALSE;
  _fromLang = 3;
  _toLang = 0;
}


SmageApp::~SmageApp()
{ 
}

// Open an image file
void SmageApp::doOpen()
{    
  char *d = getenv("SMAGE_DIR");
  QString smageDir ="";
  if ( d != NULL ) smageDir = QString( d + QObject::tr("My Smages") );
  QString string = QFileDialog::getOpenFileName(smageDir, tr("Images (*.bmp *.jpg *.jpeg *.png *.ppm *.xpm)"), this,
        "OpenFileDlg", tr("Open File"));

  if (!string.isEmpty())
  {
    bool validEnding;
    SmageImgWidget::getFileEnding(string, validEnding);
      
    if (validEnding) initializeNewImgWidget(string);
  }
}

// Assign a new type of image region selection
void SmageApp::assignNewSelectionType(const SmageSelectionType &selectionType)
{
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
  {
    SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;
    if (tmpWidget->getWidgetType() == "SMAGE_ENCODER_WIDGET")
    {
      SmageImgEncodeWidget *smageWidget = (SmageImgEncodeWidget *) widget;
      smageWidget->setSelectionType(selectionType);
    }
  }
}

// Initialize a new workspace widget (a workspace widget only contains the (un)coded image)
void SmageApp::initializeNewImgWidget(const char *name, const SmageImgWidget *tmpImgWidget)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageApp::initializeNewImgWidget()\n");
#endif

  _viewMenu->setItemEnabled( _showDataManagerId, FALSE );
  if (_wSpace->windowList().isEmpty())
    _imgWidgetCounter = 0;

  // Create a unique name for the workspace widget ("SmageImgWidget")
  char widgetName[15];
  sprintf(widgetName, "_IMG_WIDGET_%d", _imgWidgetCounter);

  SmageAppViewerWidget *viewerWidget = new SmageAppViewerWidget(_wSpace, widgetName, Qt::WDestructiveClose);// | Qt::WRepaintNoErase);
  SmageImgEncodeWidget *selectorWidget = NULL;
  SmageImgWidget *imgWidget = (SmageImgWidget *)(viewerWidget);

  if (tmpImgWidget == NULL)
  {
    if (!viewerWidget->loadImg(name))
      return;
            
    // Set a number of callback functions
    imgWidget->setShowMsgCB(this, showMsgCB);
    imgWidget->setCloseCB(this, closeCB);
    imgWidget->setSaveCB(this, saveCB);
    imgWidget->setForceOpenNewWidgetTypeCB(this, forceOpenNewCB);
  }
  else
    viewerWidget->assignWidgetData(tmpImgWidget);

  if (!imgWidget->isCodedImg())
  {
    selectorWidget = new SmageImgEncodeWidget(_wSpace, widgetName, Qt::WDestructiveClose);
    selectorWidget->assignWidgetData(viewerWidget);
    selectorWidget->setRegionSelectedExtCB(this, regionSelectedCB);
    selectorWidget->setOpenImgCB(this, openSelectorImgCB);
    selectorWidget->setBuildImgCB(this, buildAllowedCB);

    imgWidget = (SmageImgWidget *)(selectorWidget);
    
    delete viewerWidget;

    _viewMenu->setItemEnabled(_showDataManagerId, TRUE);
    _viewMenu->setItemEnabled(_switchToEncoderId, FALSE);
  }
  else
  {
    viewerWidget->setShowRegionsCB(this, showRegionsCB);
    viewerWidget->setOpenImgCB(this, openViewerImgCB);
    viewerWidget->setEraseCB(this, eraseCB);
    
    _viewMenu->setItemEnabled(_switchToEncoderId, TRUE);
  }

  _imgWidgetCounter++;
  imgWidget->showMaximized();
  _fileMenu->setItemEnabled(_saveItemId, TRUE);
  _viewMenu->setItemEnabled(_htmlGeneratorId, TRUE);
  _viewMenu->setItemChecked ( _showDataManagerId, FALSE );
}

// A region has been selected, data can now be inserted
void SmageApp::regionSelectedCB(void *userData, const bool &regionSelected, 
        const bool &allowBuildingImg)
{
  SmageApp *app = (SmageApp *) userData;
  
  if (app != NULL)
  {
    // Button for data insertion can now be activated or
    // is deactivated (if "regionSelected" is FALSE)
    app->_insertTInfo->setEnabled(regionSelected);
    app->_insertDictEntry->setEnabled(regionSelected);
    
    if (allowBuildingImg)
      app->_buildImg->setEnabled(TRUE);
    else
      app->_buildImg->setEnabled(FALSE);
  }
}

// Show a message in the status bar;
//  "ms" specifies the duration the message will be shown
void SmageApp::showMsgCB(void *userData, const QString &msg, const int &ms)
{
  SmageApp *app = (SmageApp *) userData;
  
  if (app != NULL)
  {
    if (ms < 0)
      app->statusBar()->message(msg);
    else
      app->statusBar()->message(msg, ms);
  }
}

// Show coded image regions;
//  button is enabled if "setEnabled" is TRUE;
//  button is activated if "showRegions" is TRUE
void SmageApp::showRegionsCB(void *userData, const bool &setEnabled, const bool &showRegions)
{
  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  app->_showOutlines->setEnabled(setEnabled);
  app->_showOutlines->setOn(showRegions);
}

// A workspace widget is about to get closed
void SmageApp::closeCB(void *userData)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageApp::closeCB()\n");
#endif

  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  app->_erase->setEnabled(FALSE);
  app->_showOutlines->setEnabled(FALSE);
  app->_selectRect->setEnabled(FALSE);
  app->_selectRect->setOn(FALSE);
  app->_insertTInfo->setEnabled(FALSE);
  app->_insertDictEntry->setEnabled(FALSE);
  app->_buildImg->setEnabled(FALSE);
  app->_fileMenu->setItemEnabled(app->_saveItemId, FALSE);
  app->_viewMenu->setItemEnabled (app->_htmlGeneratorId, FALSE);
  app->_viewMenu->setItemChecked (app->_showDataManagerId, FALSE);
  app->_viewMenu->setItemEnabled(app->_showDataManagerId, FALSE);
  app->menuBar()->setItemEnabled(app->_dictMenuId, FALSE);

#ifdef SMAGE_FULL
  app->_selectPoly->setEnabled(FALSE);
  app->_selectPoly->setOn(FALSE);
  app->_insertBInfo->setEnabled(FALSE);
#endif
}

// Should erasing be allowed
void SmageApp::eraseCB(void *userData, const bool &enableErasing)
{
  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  app->_erase->setEnabled(enableErasing);
}

// Should building image be allowed
void SmageApp::buildAllowedCB(void *userData, const bool &enableBuildingImg)
{
  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  app->_buildImg->setEnabled(enableBuildingImg);
}

// Set a new main window caption
void SmageApp::captionCB(void *userData, const QString &)
{
  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  QString newCaption(THE_S_NAME);
  app->setCaption(newCaption);
}

// New image (SmageAppViewerWidget class)
void SmageApp::openViewerImgCB(void *userData, const bool &allowErasing, const bool &hasDictEntriesEncoded, 
			       const TheS_DictProps &dictProps)
{
  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  app->doOpenImg(FALSE, SELECT_RECT, allowErasing, FALSE, FALSE, hasDictEntriesEncoded, dictProps);
}

// New image (SmageImgEncodeWidget class)
void SmageApp::openSelectorImgCB(void *userData, const SmageSelectionType &selectionType, 
				 const bool &allowDataInsertion, const bool &allowBuildingImg)
{
  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  TheS_DictProps dictProps;
  app->doOpenImg(TRUE, selectionType, FALSE, allowDataInsertion, allowBuildingImg, FALSE, dictProps);
}

// A new image has been opened or a different workspace widget is activated; 
//  is image coded with data ("isCodedImg"); which selection type was set ("selectionType")
void SmageApp::doOpenImg(const bool &enableDataEncoding, const SmageSelectionType &selectionType, const bool &allowErasing, 
       const bool &allowDataInsertion, const bool &allowBuildingImg, const bool &enableDictMenu, const TheS_DictProps &dictProps)
{
  _fileMenu->setItemEnabled(_saveItemId, TRUE);
  menuBar()->setItemEnabled(_dictMenuId, enableDictMenu);
  _erase->setEnabled(allowErasing);
  _insertTInfo->setEnabled(allowDataInsertion);
  _insertDictEntry->setEnabled(allowDataInsertion);
  
  if (enableDictMenu)
    {
      if (_translFromGroup != NULL)	_translFromGroup->setButton(dictProps.fromLang);
      //      _fromLang = dictProps.fromLang;
      if (_translToGroup != NULL)	_translToGroup->setButton(dictProps.toLang);
      //      _toLang = dictProps.toLang;
      _dictMenu->setItemChecked(3, dictProps.showCodedDictItem);
      _showCodedDictItem = dictProps.showCodedDictItem;
      _dictMenu->setItemChecked(4, dictProps.showSecEntries);
      _showSecEntries = dictProps.showSecEntries;
    }

  _buildImg->setEnabled(allowBuildingImg);
  _viewMenu->setItemEnabled(_showDataManagerId, FALSE);
  _viewMenu->setItemEnabled(_htmlGeneratorId, TRUE);

  if (!enableDataEncoding)
  {
    _showOutlines->setEnabled(TRUE);
    _selectRect->setEnabled(FALSE);
    _selectRect->setOn(FALSE);

#ifdef SMAGE_FULL
    _selectPoly->setEnabled(FALSE);
    _selectPoly->setOn(FALSE);
    _insertBInfo->setEnabled(FALSE);
#endif
  }
  else
  {
    _viewMenu->setItemEnabled(_showDataManagerId, TRUE);
    _selectRect->setEnabled(TRUE);
    if (selectionType == SELECT_RECT)
      _selectRect->setOn(TRUE);

    _showOutlines->setEnabled(FALSE);
    _showOutlines->setOn(FALSE);

#ifdef SMAGE_FULL
    _selectPoly->setEnabled(TRUE);
    if (selectionType == SELECT_POLY)
      _selectPoly->setOn(TRUE);

    _insertBInfo->setEnabled(TRUE);
#endif
  }
}

// Callback for saving
void SmageApp::saveCB(void *userData)
{
  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  app->doSaveP();
}

// Callback that forces to open a new widget
void SmageApp::forceOpenNewCB(void *userData, SmageImgWidget *imgWidget, const QString &newFileName)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageApp::forceOpenNewCB()\n");
#endif

  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  if (newFileName.isEmpty())
    app->initializeNewImgWidget(newFileName.ascii(), imgWidget);
  else
    app->initializeNewImgWidget(newFileName.ascii());
  
  delete imgWidget;
}

// Save current image
void SmageApp::doSaveP()
{
  if (!_wSpace->windowList().isEmpty())
  {
    QWidget *widget = _wSpace->activeWindow();
    if (widget != NULL)
    {
      char *d = getenv("SMAGE_DIR");
      QString smageDir ="";
      if ( d != NULL ) smageDir = QString( d + QObject::tr("My Smages") );
      QString fileName = QFileDialog::getSaveFileName(smageDir, tr("Images (*.png)"), this,
//      QString fileName = QFileDialog::getSaveFileName(".", tr("Images (*.png)"), this,
                "saveFileDlg", tr("Save File"));

      if (!fileName.isEmpty())
      {
        bool validEnding;
        QString fileType = SmageImgWidget::getFileEnding(fileName, validEnding);
        
        if (!validEnding || !fileType.contains("png"))
          fileName += ".png";
    {
      SmageImgWidget *smageWidget = (SmageImgWidget *) widget;
      smageWidget->saveImg(fileName);
    }
      }
    }
  }
}

// Current selection type: rectangular region
void SmageApp::doSelectRect()
{
  assignNewSelectionType(SELECT_RECT);
}

// Current selection type: any shape
#ifdef SMAGE_FULL
void SmageApp::doSelectPoly()
{
  assignNewSelectionType(SELECT_POLY);
}
#endif

// Specify text data to be embedded in selected region and
//  correlate specified data with image region
void SmageApp::doInsertTInfo()
{
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
  {
    SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;

    if (tmpWidget->getWidgetType() == "SMAGE_ENCODER_WIDGET")
    {
      SmageImgEncodeWidget *smageWidget = (SmageImgEncodeWidget *) widget;
      
      // Region is verified ... does it overlap with an already chosen region?
      if (smageWidget->verifyRegion())
      {
	TextShapeType textType;
	if (SmageTextTypeDlg::getTextType(textType, this))
	  {
	    if (textType == SMAGE_PLAIN_TEXT)
	      {
		QString textInput;
		if (SmageImgInputDlg::getTextInput(textInput, this))
		  {
		    // Set specified text if has been typed in
		    if (!textInput.isEmpty())
		      smageWidget->integratePlainText(textInput);
		    
		    // Allow to build the image
		    if (!textInput.isEmpty())
		      _buildImg->setEnabled(TRUE);
		  }
	      }
	    else
	      {
		smageWidget->integrateFormattedText(textType);
	      }
	  }
      }
      else
      {
        QMessageBox::warning(smageWidget, QObject::tr("Smage - Message"), tr("Selected regions should not overlap."),
           QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

      }
    }
  }
}

// Specify image data to be embedded in selected region
void SmageApp::doInsertBInfo()
{
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
  {
    SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;

    if (tmpWidget->getWidgetType() == "SMAGE_ENCODER_WIDGET")
    {      
      _imgGrabber->show();
      _imgGrabber->setActiveWindow();
      _imgGrabber->raise();
    }
  }
}

void SmageApp::doInsertDictEntry()
{
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
    {
      SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;
      
      if (tmpWidget->getWidgetType() == "SMAGE_ENCODER_WIDGET")
	{      
	  SmageImgEncodeWidget *smageWidget = (SmageImgEncodeWidget *) widget;
	  smageWidget->doInsertDictEntry();
	}
    }
}

// Encode data into current image

void SmageApp::doBuildImg()
{
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
  {
    SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;

    if (tmpWidget->getWidgetType() == "SMAGE_ENCODER_WIDGET")
    {
      SmageImgEncodeWidget *smageWidget = (SmageImgEncodeWidget *) widget;
      smageWidget->buildImg();
    }
  }
}

// Delete data from image.
void SmageApp::doErase()
{
   QWidget *widget = _wSpace->activeWindow();
   if (widget != NULL)
   {
     SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;

     if (tmpWidget->getWidgetType() == "SMAGE_APP_VIEWER_WIDGET")
     {
       SmageAppViewerWidget *smageWidget = (SmageAppViewerWidget *) widget;
       smageWidget->eraseDataFromImage();
     }
   }
}

// Show outlines of coded image regions
void SmageApp::doShowOutlines()
{
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
  {
    SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;

    if (tmpWidget->getWidgetType() == "SMAGE_APP_VIEWER_WIDGET")
    {
      SmageAppViewerWidget *smageWidget = (SmageAppViewerWidget *) widget;
      smageWidget->showOutlines(_showOutlines->isOn());
    }
  }
}

void SmageApp::selCopyImgCB(void *userData, const QImage* img, const QPointArray &points)
{
  SmageApp *app = (SmageApp *) userData;
  if (app == NULL)
    return;

  QWidget *widget = app->_wSpace->activeWindow();
  if (widget != NULL)
  {
    SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;
      
    if (tmpWidget->getWidgetType() == "SMAGE_ENCODER_WIDGET")
    {
      SmageImgEncodeWidget *smageWidget = (SmageImgEncodeWidget *) widget;
      smageWidget->integrateImg(img, points);
    }
  }
}


void SmageApp::doShowDataManager(const bool &show)
{
  if (show)
    {
      QWidget *widget = _wSpace->activeWindow();
      if (widget != NULL)
	{
	  SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;
	  
	  if (tmpWidget->getWidgetType() == "SMAGE_ENCODER_WIDGET")
	    {
	      SmageImgEncodeWidget *smageWidget = (SmageImgEncodeWidget *) widget;
	      smageWidget->getDataManageWidget()->show();
	    }
	}
    }
  else
    {
      for (uint i=0; i<_wSpace->windowList().count(); ++i)
	{
	  if (((SmageImgWidget *)(_wSpace->windowList().at(i)))->getWidgetType() == "SMAGE_ENCODER_WIDGET")
	    ((SmageImgEncodeWidget *) _wSpace->windowList().at(i))->getDataManageWidget()->hide();
	}
    }
}


void SmageApp::doSwitchToEncoder()
{
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
    {
      SmageImgWidget *tmpWidget = (SmageImgWidget *) widget;
      
      if (tmpWidget->getWidgetType() == "SMAGE_APP_VIEWER_WIDGET")
	{
	  SmageAppViewerWidget *viewerWidget = (SmageAppViewerWidget *) tmpWidget;

	  char widgetName[15];
	  sprintf(widgetName, "_IMG_WIDGET_%d", _imgWidgetCounter);
	  _imgWidgetCounter++;

	  SmageImgEncodeWidget *encWidget = new SmageImgEncodeWidget(_wSpace, widgetName, Qt::WDestructiveClose);
	  encWidget->assignWidgetData(viewerWidget);
	  encWidget->setRegionSelectedExtCB(this, regionSelectedCB);
	  encWidget->setOpenImgCB(this, openSelectorImgCB);
	  encWidget->setBuildImgCB(this, buildAllowedCB);
	  
	  encWidget->setCodedImg(viewerWidget->getOutlineImg(), viewerWidget->getIdBuffer(), 
				 viewerWidget->getNumCodedDataItems());
	  delete viewerWidget;
	  
	  encWidget->showMaximized();

	  _viewMenu->setItemEnabled(_switchToEncoderId, FALSE);
	  _showOutlines->setEnabled(FALSE);
	  _showOutlines->setOn(FALSE);
	  _viewMenu->setItemChecked(_showDataManagerId, FALSE);
	}
    }
}

void SmageApp::doShowSecEntries(bool show)
{
  _showSecEntries = show;
  updateDictText();
}


void SmageApp::doShowCodedDictItem(bool show)
{
  _showCodedDictItem = show;
  updateDictText();
}

void SmageApp::doTranslFrom(int from)
{
  _fromLang = from;
  updateDictText();
}
  

void SmageApp::doTranslTo(int to)
{
  _toLang = to;
  updateDictText();
}


void SmageApp::updateDictText()
{
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
    {
      if (((SmageImgWidget *)(widget))->getWidgetType() == "SMAGE_APP_VIEWER_WIDGET")
	{
	  SmageAppViewerWidget *smageWidget = (SmageAppViewerWidget *) widget;
	  smageWidget->updateDictText(_fromLang, _toLang, _showSecEntries, _showCodedDictItem);
	}
    }
}


void SmageApp::doGenerateHtml()
{ 
  QWidget *widget = _wSpace->activeWindow();
  if (widget != NULL)
    {
      SmageImgWidget *imgWidget = (SmageImgWidget *)(widget);
      QImage *img = imgWidget->getImgPointer();

      if (img != NULL)
	{
	  _htmlGenerator->setValues(ImageProperties(imgWidget->getFileName(), img->width(), img->height()));
	  _htmlGenerator->show();
	}
    }
}
