// project header
#include "SmageAppViewerWidget.h"
#include "SmageMsgDlg.h"
#include "../smageSharedFiles/SmageDecodeHandler.h"

// qt header
#include <qbitmap.h>
#include <qcursor.h>
#include <qlabel.h>
#include <qmessagebox.h>

// c/c++ header
#include <assert.h>
#include <iostream>

using namespace std;

SmageAppViewerWidget::SmageAppViewerWidget(QWidget *parent, const char* name, WFlags f)
  : SmageViewerWidget(parent, name, f)
{ 
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageAppViewerWidget::SmageAppViewerWidget()\n");
#endif
  // Initialize callbacks
  _showRegionsCBFunc = NULL;
  _showRegionsUserData = NULL;
  _openImgCBFunc = NULL;
  _openImgUserData = NULL;
  _eraseCBFunc = NULL;
  _eraseUserData = NULL;

  _waitForClick = FALSE;
  _allowErasing = TRUE;
  _hasDictEntriesEncoded = FALSE;
}

SmageAppViewerWidget::~SmageAppViewerWidget()
{ 
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageAppViewerWidget::~SmageAppViewerWidget()\n");
#endif
}

void SmageAppViewerWidget::paintEvent(QPaintEvent *event)
{
  if (_imgLoaded)
  {
    // Update buttons in main application
    if (hasFocus())
    {
      if (_openImgCBFunc != NULL)
        (*_openImgCBFunc)(_openImgUserData, _allowErasing, _hasDictEntriesEncoded, _dictProps);
      if (_showRegionsCBFunc != NULL)
        (*_showRegionsCBFunc)(_showRegionsUserData, !_allowRegionSelection, _showOutlines);
    }
    
    if (_waitForClick)
      _currentPixmapID = _previousPixmapID = -1;

    SmageViewerWidget::paintEvent(event);      
  }  
}

void
SmageAppViewerWidget::mouseMoveEvent( QMouseEvent *e )
{
  _currentMousePos = e->pos();
  getLocalImgPos(_currentMousePos, _mousePos);

  if (!_waitForClick)
    onMouseMove();
}

void 
SmageAppViewerWidget::mouseReleaseEvent(QMouseEvent *)
{
  if (_waitForClick)
    {
      _waitForClick = FALSE;
      showOutlines(FALSE);
      eraseDataPartFromImage(mapFromGlobal(QCursor::pos()));
    }
}

void
SmageAppViewerWidget::resizeEvent(QResizeEvent *event)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageAppViewerWidget::resizeEvent()\n");
#endif
  inherited::resizeEvent( event );

  // update bounds of each SmageTextWidget
  // and keep any anchored SmageTextWidget at its position
  if ( !_imgLoaded ) return;
  std::map<unsigned int, SmageStructDecodedData>::iterator itr;
  for ( itr = _decodedDataMap.begin(); itr != _decodedDataMap.end(); itr++ )
  {
    if ( itr->second._tw != NULL )
    {
      SmageTextWidget *stw = itr->second._tw;
      stw->updateBounds();
      if ( stw->_anchored )
      {
        QPoint oldPos = QPoint( stw->pos().x() + stw->_xOffset, stw->pos().y() + stw->_yOffset );
        QPoint newPos = QPoint( oldPos.x() + _resizeDiff.width(), oldPos.y() + _resizeDiff.height() );
        stw->reshape( newPos );
        stw->move( newPos.x() - stw->_xOffset, newPos.y() - stw->_yOffset );
        stw->show();
      }
    }
  }

  // keep bubble at its position
  if ( _bubbleVec != NULL  ) 
  {
    std::vector<SmageViewerShapedText *>::iterator bubbleIter;
    for ( bubbleIter = _bubbleVec->begin(); bubbleIter != _bubbleVec->end(); bubbleIter++ )
    {
      (*bubbleIter)->move((*bubbleIter)->_position + _imgStartPos);
      (*bubbleIter)->show();
    }
  }
}

// Class specific operations when an image has been successfully loaded
void SmageAppViewerWidget::doWhenSuccesImgLoading()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageAppViewerWidget::doWhenSuccesImgLoading()\n");
#endif

  _decodeIdBuffer = QImage();
  _outlineImg = QImage();
  _decodedDataMap.clear();
  _allowErasing = FALSE;
  setCaption(_imgFileName);

  if (_isCodedImg)
  {  
    setMouseTracking(TRUE);
    _allowErasing = TRUE;
    // doDecoding
    SmageStructDecodeInfo info = SmageDecodeHandler::decode( _img, _decodeIdBuffer, _outlineImg, _decodedDataMap, this );
    if (info._numOfEncodedDictEntries > 0) _hasDictEntriesEncoded = TRUE;

#ifdef THE_S_PRINT_DEBUG
    _decodeIdBuffer.save("decodeIdBuffer.png", "PNG", 100);
#endif

    if (_openImgCBFunc != NULL)
      (*_openImgCBFunc)(_openImgUserData, _allowErasing, _hasDictEntriesEncoded, _dictProps);
  }
}


//  Delete data from image;
//  - remove coded data, if image is a coded image or
//  - delete user selected regions and their associated data
void SmageAppViewerWidget::eraseDataFromImage()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageViewerWidget::eraseDataFromImage()\n");
#endif

  QString aMsg;

  if (_isCodedImg)
  {
    int result;
    if (SmageMsgDlg::getResult(result, this, tr("Delete Data"), tr("Delete all the data from the image?"),
			       QObject::tr("Yes"), 3, QObject::tr("No"), QObject::tr("Cancel")))
      {
	if (result == 3)
	  return;

	if (result == 1)
	  {
	    // Remove all encoded data
	    _smageGenerator->eraseDataFromImage(_img);
	    _decodedDataMap.clear();
	    
	    _isCodedImg = _showOutlines = _hasDictEntriesEncoded = FALSE;
	    resetImgWidget();

	    setMouseTracking(FALSE);
	    _imgHasBeenModified = TRUE;

	    if (_showRegionsCBFunc != NULL)
	      (*_showRegionsCBFunc)(_showRegionsUserData, !_allowRegionSelection, _showOutlines);

	    if (_eraseCBFunc != NULL)
	      (*_eraseCBFunc)(_eraseUserData, _allowErasing);

	    if (_forceOpenNewCBFunc != NULL && _forceOpenNewUserData != NULL)
	      (*_forceOpenNewCBFunc)(_forceOpenNewUserData, this, QString(""));
	  }

	else if (result == 2)
	  {		
	    if (SmageMsgDlg::getResult(result, this, tr("Delete Data"), 
				       tr("Select the region in the image\nto be deleted ..."), QObject::tr("OK")))
	      {
		_waitForClick = TRUE;
		showOutlines(TRUE);
	      }
	  }
      }
  }
}

void SmageAppViewerWidget::eraseDataPartFromImage(const QPoint &pos)
{
  QPoint imgLikePos;
  getLocalImgPos(pos, imgLikePos);

  QString failInfo(tr("There is no data encoded in this region"));

  if (imgLikePos.x() < 0 || imgLikePos.y() < 0 ||
      imgLikePos.x() > _imgPointer->width() || imgLikePos.y() > _imgPointer->height() ||
      !_decodeIdBuffer.valid(imgLikePos.x(), imgLikePos.y()))
    {
      QMessageBox::information(this, tr("Delete data"), failInfo);
      return;
    }

  int idBufferColorIndex;
  idBufferColorIndex = _decodeIdBuffer.pixelIndex(imgLikePos.x(), imgLikePos.y());

  if (idBufferColorIndex < 2)
    {
      QMessageBox::information(this, tr("Delete Data"), failInfo);
      return;
    }

  _smageGenerator->eraseDataPartFromImage(_img, _decodeIdBuffer, idBufferColorIndex);
  _isCodedImg = _smageGenerator->isCodedImg(_img);

  _imgHasBeenModified = TRUE;

  if (_forceOpenNewCBFunc != NULL && _forceOpenNewUserData != NULL)
    (*_forceOpenNewCBFunc)(_forceOpenNewUserData, this, QString::null);
}

// Before closing the widget ...
//  - ask if modified image should be stored
void SmageAppViewerWidget::closeEvent(QCloseEvent *event)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageAppViewerWidget::closeEvent()\n");
#endif
  
  inherited::closeEvent(event);
}


// Reset the image widgets data
void SmageAppViewerWidget::resetImgWidget()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageAppViewerWidget::resetImgWidget()\n");
#endif

  _imgPointer = &_img;
  _currentPixmapID = _previousPixmapID = -1;

  _decodeIdBuffer = QImage();
  _outlineImg = QImage();

  _allowErasing = FALSE;
  _sceneChanged = TRUE;
  repaint(FALSE);
}


void SmageAppViewerWidget::assignWidgetData(const SmageImgWidget *smageWidget)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageAppViewerWidget::assignWidgetData()\n");
#endif

  SmageImgWidget::assignWidgetData(smageWidget);

  if (smageWidget->getWidgetType() == "SMAGE_APP_VIEWER_WIDGET")
    _allowErasing = ((SmageAppViewerWidget *) smageWidget)->_allowErasing;
}
