// project header
#include "SmageBubble.h"

// qt header
#include <qcursor.h>
#include <qpainter.h>

// c/c++ header
#include <iostream>
#include <math.h>

using namespace std;

SmageBubble::SmageBubble(QWidget *parent, const char* name, const QRect &rect, 
			 const bool &hasAutoMovement, const bool &isEllipse)
  : QWidget(parent, name)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageBubble::SmageBubble()\n");
#endif
  
  _pressedPos = QPoint(0,0);
  _hasAutoMovement = hasAutoMovement;
  _isEllipse = isEllipse;

  _ellipseRect = QRect(rect.topLeft(), rect.bottomRight());

  _buttonBoundingBox = NULL;

  _mouseMoveCBFunc = NULL;
  _mouseMoveUserData = NULL;
  _mouseReleaseCBFunc = NULL;
  _mouseReleaseUserData = NULL;
  _mousePressCBFunc = NULL;
  _mousePressUserData = NULL;
  _updateMaskCBFunc = NULL;
  _updateMaskUserData = NULL;
  _updateBBoxCBFunc = NULL;
  _updateBBoxUserData = NULL;

  QPoint p = _ellipseRect.bottomRight();
  _buttonBoundingBox = new SmageButton(this, "ButtonBBox");
  _buttonBoundingBox->setAutoMovement(FALSE);
  _buttonBoundingBox->setMouseReleaseCB(this, buttonReleaseCB);
  _buttonBoundingBox->setMousePressCB(this, buttonPressCB);
  _buttonBoundingBox->setMouseMoveCB(this, updateBBoxCB);
  _buttonBoundingBox->setGeometry(p.x() - int(_buttonBoundingBox->width()/2.0), p.y() - int(_buttonBoundingBox->height()/2.0), 
				      _buttonBoundingBox->width(), _buttonBoundingBox->height());
  
  assignWidgetSize();
  calcWidgetMask();
  setMask(_bubbleMask);
}

SmageBubble::~SmageBubble()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageBubble::~SmageBubble()\n");
#endif
}

void SmageBubble::setBubbleData(const TheS_BubbleShapeData &data)
{
  _shapeData.bubbleBackCol = QColor(data.bubbleBackCol);
  _shapeData.bubbleOutlineCol = QColor(data.bubbleOutlineCol);
  _shapeData.bubbleOutlineWidth = data.bubbleOutlineWidth;
  _shapeData.bubbleBackgroundAlpha = data.bubbleBackgroundAlpha;
  repaint();
}



void SmageBubble::assignWidgetSize()
{  
  int top = geometry().top();
  int left = geometry().left();
 
  int buttonWidth = 0;
  int buttonHeight = 0;

  buttonWidth = int(ceil(double(_buttonBoundingBox->width()/2.0)));
  buttonHeight = int(ceil(double(_buttonBoundingBox->height()/2.0)));

  setGeometry(left, top, _ellipseRect.width() + buttonWidth, _ellipseRect.height() + buttonHeight);
}

void SmageBubble::calcWidgetMask()
{
  _bubbleMask = QBitmap(width(), height(), TRUE);

  QPainter painter(&_bubbleMask);
  QPen pen(Qt::color1,_shapeData.bubbleOutlineWidth);
  QBrush brush;
  if ( _shapeData.bubbleBackgroundAlpha != 0 )
    brush = QBrush(Qt::color1);
  else
    brush = QBrush( Qt::NoBrush );
  painter.setPen(pen);
  painter.setBrush(brush);
  
  if (_isEllipse)
    painter.drawEllipse(_ellipseRect);
  else
    painter.drawRect(_ellipseRect);
  
  painter.drawRect(_buttonBoundingBox->geometry());
  
  if (_isEllipse)
	{
	  painter.setBrush(Qt::NoBrush);
	  painter.drawRect(_ellipseRect);
	}
}


void SmageBubble::paintEvent(QPaintEvent *)
{
  QPainter p(this);

  QPen pen;
  if (_shapeData.bubbleOutlineWidth > 0)
    pen = QPen(_shapeData.bubbleOutlineCol, 2*_shapeData.bubbleOutlineWidth- _shapeData.bubbleOutlineWidth%2);
  else
    pen = QPen( Qt::NoPen );
  QBrush brush;
  if ( _shapeData.bubbleBackgroundAlpha != 0 )
    brush = QBrush(_shapeData.bubbleBackCol);
  else
    brush.setPixmap( QPixmap("pixmaps/transparentBackground.xpm") );
  p.setPen(pen);
  p.setBrush(brush);
  if (_isEllipse)
    p.drawEllipse(_ellipseRect);
  else
//    p.drawRect(_ellipseRect);
  {
    if (_shapeData.bubbleOutlineWidth == 0 ) p.drawRect(_ellipseRect);
    else
    {
      p.setPen(QPen(_shapeData.bubbleOutlineCol, 1));
      p.setBrush(Qt::NoBrush);
      uint i;
      for (i=0; i<_shapeData.bubbleOutlineWidth-1;i++ )
        p.drawRect(_ellipseRect.x()+i,_ellipseRect.y()+i, _ellipseRect.width()-2*i, _ellipseRect.height()-2*i );
      p.setBrush(brush);
      p.drawRect(_ellipseRect.x()+i,_ellipseRect.y()+i, _ellipseRect.width()-2*i, _ellipseRect.height()-2*i );
    }
  }

  if (_isEllipse)
	{
	  p.setBrush(Qt::NoBrush);
	  p.setPen(Qt::SolidLine);
	  p.drawRect(_ellipseRect);
	}
}

void SmageBubble::mousePressEvent(QMouseEvent *event)
{  
  _pressedPos = event->pos();

  if (_mousePressCBFunc != NULL && _mousePressUserData != NULL)
    (*_mousePressCBFunc)(_mousePressUserData, event);

  if (_updateMaskCBFunc != NULL && _updateMaskUserData != NULL)
    (*_updateMaskCBFunc)(_updateMaskUserData, FALSE);
}

void SmageBubble::mouseReleaseEvent(QMouseEvent *)
{  
  if ( _hasAutoMovement)
    setMask(_bubbleMask);

  if (_mouseReleaseCBFunc != NULL && _mouseReleaseUserData != NULL)
    (*_mouseReleaseCBFunc)(_mouseReleaseUserData);

  if (_updateMaskCBFunc != NULL && _updateMaskUserData != NULL)
    (*_updateMaskCBFunc)(_updateMaskUserData, FALSE);
}

void SmageBubble::mouseMoveEvent(QMouseEvent *event)
{
  if ( _hasAutoMovement)
    {
      clearMask();
      move(pos() - _pressedPos + mapFromGlobal(QCursor::pos()));

      event->accept();
    }

  if (_mouseMoveCBFunc != NULL && _mouseMoveUserData != NULL)
    (*_mouseMoveCBFunc)(_mouseMoveUserData);

  if (_updateMaskCBFunc != NULL && _updateMaskUserData != NULL)
    (*_updateMaskCBFunc)(_updateMaskUserData, TRUE);
}


void SmageBubble::buttonReleaseCB(void *userData)
{
  SmageBubble *db = (SmageBubble *) userData;
  if (db == NULL)
    return;

  db->calcWidgetMask();
  db->setMask(db->_bubbleMask);

  if (db->_updateMaskCBFunc != NULL && db->_updateMaskUserData != NULL)
  	(*db->_updateMaskCBFunc)(db->_updateMaskUserData, FALSE);
}

void SmageBubble::buttonPressCB(void *userData)
{
  SmageBubble *db = (SmageBubble *) userData;
  if (db == NULL)
    return;

  db->clearMask();
  if (db->_updateMaskCBFunc != NULL && db->_updateMaskUserData != NULL)
	(*db->_updateMaskCBFunc)(db->_updateMaskUserData, TRUE);
}


void SmageBubble::updateBBoxCB(void *userData, const QPoint &pos)
{
  SmageBubble *db = (SmageBubble *) userData;
  if (db == NULL)
    return;

  db->updateBBox(pos);

  if (db->_updateMaskCBFunc != NULL && db->_updateMaskUserData != NULL)
	(*db->_updateMaskCBFunc)(db->_updateMaskUserData, TRUE);

  if (db->_updateBBoxCBFunc != NULL && db->_updateBBoxUserData != NULL)
	(*db->_updateBBoxCBFunc)(db->_updateBBoxUserData, db->name());
}


void SmageBubble::updateBBox(const QPoint &pos)
{
  QPoint currentPos = _buttonBoundingBox->pos();
  QPoint newPos = pos - QPoint(int(_buttonBoundingBox->width()/2.0), int(_buttonBoundingBox->height()/2.0));
  int diffX = newPos.x() - currentPos.x();
  int diffY = newPos.y() - currentPos.y();

  int newEllipseWidth = _ellipseRect.width() + diffX;
  int newEllipseHeight = _ellipseRect.height() + diffY;

  if (newEllipseWidth <= 0 || newEllipseHeight <= 0)
    return;

  _ellipseRect.setWidth(newEllipseWidth);
  _ellipseRect.setHeight(newEllipseHeight);

  newPos = _ellipseRect.bottomRight() - QPoint(int(_buttonBoundingBox->width()/2.0), int(_buttonBoundingBox->height()/2.0));
  _buttonBoundingBox->move(newPos);

  assignWidgetSize();
  clearMask();

  update();
}
