// project header
#include "SmageDataManageWidget.h"
#include "./pixmaps/managerView.xpm"
#include "./pixmaps/managerDel.xpm"
#include "./pixmaps/managerViewFocus.xpm"

// qt header
#include <qcheckbox.h>
#include <qgroupbox.h>
#include <qinputdialog.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qtooltip.h>

// c/c++ header
#include <assert.h>
#include <iostream>

using namespace std;


SmageDataManageWidgetItem::SmageDataManageWidgetItem(QWidget *parent, const char *name, const int &widgetItemID,
						     const QString &itemName)
  : QWidget(parent, name )
{ 
  setPaletteBackgroundColor(QColor(232,227,227));

  _widgetItemID = widgetItemID;

  _dataManagerItemUserData = NULL;
  _dataManagerItemCBFunc = NULL;
  _dataManagerFocusItemCBFunc = NULL;
  _dataManagerFocusItemUserData = NULL;

  QGroupBox *box = new QGroupBox(this);
  box->setColumns(3);
  box->setInsideMargin(5);

  _widgetItemGridLayout = new QGridLayout(this, 1, 1);
  
  _widgetItemCheckBox = new QCheckBox(box);
  _widgetItemCheckBox->setPixmap(QPixmap(managerView_xpm));
  _widgetItemCheckBox->setChecked(TRUE);
  QToolTip::add(_widgetItemCheckBox, "Show/hide all data specified for this region.\nCaution: Hidden data will not be encoded into the image!");
  connect(_widgetItemCheckBox, SIGNAL(clicked()), SLOT(viewPressed()));

  _widgetItemLabel = new QLabel(itemName, box);
  _widgetItemLabel->setFixedSize(100,25);

  _widgetItemDelButton = new QPushButton(box);
  _widgetItemDelButton->setPixmap(QPixmap(managerDel_xpm));
  _widgetItemDelButton->setFixedSize(17,17);
  QToolTip::add(_widgetItemDelButton, tr("Delete all data specified for this region."));
  connect(_widgetItemDelButton, SIGNAL(clicked()), SLOT(delPressed()));

  _widgetItemGridLayout->addWidget(box, 0, 0, Qt::AlignHCenter);
  
  resize(150, 30);
}

SmageDataManageWidgetItem::~SmageDataManageWidgetItem()
{ 
}


void SmageDataManageWidgetItem::mousePressEvent(QMouseEvent *event)
{
  if (event->button() == Qt::RightButton)
    {
      bool ok;
      QString text = QInputDialog::getText(tr("New Identifier"), QString::null, QLineEdit::Normal, _widgetItemLabel->text(), &ok, this);
      if (ok && !text.isEmpty()) 
	{
	  if (text.length() > 12)
	    text = text.left(12);
	  _widgetItemLabel->setText(text);
	} 
    }
  else
    {
      setPaletteBackgroundColor(QColor(158,174,255));
      _widgetItemCheckBox->setPixmap(QPixmap(managerViewFocus_xpm));
      
      if (_dataManagerFocusItemUserData != NULL && _dataManagerFocusItemCBFunc != NULL)
	(*_dataManagerFocusItemCBFunc)(_dataManagerFocusItemUserData, _widgetItemID);
    }
}

void SmageDataManageWidgetItem::viewPressed()
{
  int isToggled;
  if (_widgetItemCheckBox->isOn()) isToggled = 1;
  else isToggled = 0;

  if (_dataManagerItemCBFunc != NULL && _dataManagerItemUserData != NULL)
    (*_dataManagerItemCBFunc)(_dataManagerItemUserData, isToggled, _widgetItemID);
}
  

void SmageDataManageWidgetItem::delPressed()
{
  if (_dataManagerItemCBFunc != NULL && _dataManagerItemUserData != NULL)
    (*_dataManagerItemCBFunc)(_dataManagerItemUserData, 2, _widgetItemID);
}


void SmageDataManageWidgetItem::resetItem()
{
  setPaletteBackgroundColor(QColor(232,227,227));
  _widgetItemCheckBox->setPixmap(QPixmap(managerView_xpm));
}


SmageDataManageWidget::SmageDataManageWidget(QWidget *parent, const char* name)
  : QWidget(parent, name, Qt::WStyle_Customize | Qt::WStyle_Title |/* Qt::WType_Dialog */ Qt::WType_TopLevel | Qt::WStyle_DialogBorder)
{ 
  _manageWidgetGridLayout = new QGridLayout(this, 0, 0);
  _itemMap.clear();

  _dataManagerWidgetCBFunc = NULL;
  _dataManagerWidgetUserData = NULL;
  _dataManagerFocusCBFunc = NULL;
  _dataManagerFocusUserData = NULL;

  _focusedItemID = -1;

  resize(150, 0);

  hide();
}

SmageDataManageWidget::~SmageDataManageWidget()
{ 
}

void SmageDataManageWidget::addItem(const QString &name, const int &id)
{
  if (_itemMap.find(id) != _itemMap.end())
    return;

  SmageDataManageWidgetItem *item = new SmageDataManageWidgetItem(this, "Item", id, name);
  _manageWidgetGridLayout->addWidget(item, _manageWidgetGridLayout->numRows(), 0);
  item->setDataManagerItemCB(this, itemButtonPressedCB);
  item->setDataManagerFocusItemCB(this, itemFocusedCB);
  item->show();

  pair<int, SmageDataManageWidgetItem *> p(id, item);
  _itemMap.insert(p);

  setFixedSize(width(), height() + item->height());
}

void SmageDataManageWidget::removeItem(const int &id)
{
  map<int, SmageDataManageWidgetItem *>::iterator mapIter = _itemMap.find(id);
  if (mapIter != _itemMap.end())
    {
      SmageDataManageWidgetItem *widgetItem = (SmageDataManageWidgetItem *) mapIter->second;
      _manageWidgetGridLayout->remove(mapIter->second);

      if (_itemMap.size() == 1)
	setFixedSize(width(), 0);
      else
	setFixedSize(width(), height() - mapIter->second->height());

      _itemMap.erase(mapIter);
      delete widgetItem;
    }
}


void SmageDataManageWidget::itemButtonPressedCB(void *userData, const int &whichButton, 
						const int &itemID)
{
  SmageDataManageWidget *widget = (SmageDataManageWidget *) userData;
  if (widget == NULL)
    return;

  if (widget->_dataManagerWidgetCBFunc != NULL && widget->_dataManagerWidgetUserData != NULL)
    (*widget->_dataManagerWidgetCBFunc)(widget->_dataManagerWidgetUserData, whichButton, itemID);

  if (whichButton == 2)
    widget->removeItem(itemID);
}


void SmageDataManageWidget::itemFocusedCB(void *userData, const int &whichItem)
{
  SmageDataManageWidget *widget = (SmageDataManageWidget *) userData;
  if (widget == NULL)
    return;

  widget->itemFocused(whichItem);
}


void SmageDataManageWidget::itemFocused(const int &whichItem)
{
  map<int, SmageDataManageWidgetItem *>::iterator itemMapIter;

  if (whichItem == -1)
    {
      for (itemMapIter=_itemMap.begin(); itemMapIter!=_itemMap.end(); itemMapIter++)
	itemMapIter->second->resetItem();
    }

  else if (_focusedItemID != -1 && _focusedItemID != whichItem)
    {
      itemMapIter = _itemMap.find(_focusedItemID);
      if (itemMapIter != _itemMap.end())
	itemMapIter->second->resetItem();
    }

  _focusedItemID = whichItem;

  if (_dataManagerFocusCBFunc != NULL && _dataManagerFocusUserData != NULL)
    (*_dataManagerFocusCBFunc)(_dataManagerFocusUserData, whichItem);
}


void SmageDataManageWidget::closeEvent(QCloseEvent *event)
{
  event->ignore();
}
