// project header
#include "SmageEditorSimpleBubble.h"
#include "SmageEditorSpeechBubble.h"
#include "SmageEditorThinkBubble.h"
#include "SmageEncDoneDlg.h"
#include "SmageImgEncodeWidget.h"
#include "SmageImgHandler.h"

// qt header
#include <qmessagebox.h>

// c/c++ header
#include <assert.h>
#include <iostream>

using namespace std;

#include "./pixmaps/textInsertedIcon.xpm"
#include "./pixmaps/dictInsertedIcon.xpm"

SmageImgEncodeWidget::SmageImgEncodeWidget(QWidget *parent, const char* name, WFlags f)
  : SmageImgSelectionWidget(parent, name, f)
{ 
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::SmageImgEncodeWidget()\n");
#endif

  _allowBuildingImg = FALSE;
  _outlineImg = QImage();
  _codedRegionsBitmap = QBitmap();

  _startId = 0;

  // Initialize callbacks
  _openImgCBFunc = NULL;
  _openImgUserData = NULL;
  _buildImgCBFunc = NULL;
  _buildImgUserData = NULL;
  _regionSelectedExtCBFunc = NULL;
  _regionSelectedExtUserData = NULL;

  _dataMap.clear();
  _dataManageWidget = new SmageDataManageWidget(this, "DataManageWidget");
  _dataManageWidget->setDataManagerWidgetCB(this, dataManageWidgetButtonPressedCB);
  _dataManageWidget->setDataManagerFocusCB(this, dataManageWidgetFocusCB);
  _dataManageWidget->hide();
}


SmageImgEncodeWidget::~SmageImgEncodeWidget()
{ 
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::~SmageImgEncodeWidget()\n");
#endif
}

void SmageImgEncodeWidget::mouseReleaseEvent(QMouseEvent *event)
{
  inherited::mouseReleaseEvent(event);

  _dataManageWidget->itemFocused(-1);

  if (_regionSelectedExtCBFunc != NULL)
    (*_regionSelectedExtCBFunc)(_regionSelectedExtUserData, _regionSelected, _allowBuildingImg);
}

void SmageImgEncodeWidget::paintEvent(QPaintEvent *event)
{
  inherited::paintEvent(event);

  if (_imgLoaded)
  {
    // Do repainting several things
    if (_sceneChanged)
    {
      _painter->begin(this);
      _painter->setRasterOp(NotROP);
      _pen->setStyle(Qt::DotLine);
      _painter->setPen(*_pen);
    
      map<int, TheS_Data>::const_iterator dataMapIter;
      for (dataMapIter=_dataMap.begin(); dataMapIter!=_dataMap.end(); dataMapIter++)
	{	
	  if ((!dataMapIter->second.data.formTextVec.empty() ||
	       !dataMapIter->second.data.textVec.empty() ||
	       dataMapIter->second.data.dictItem != NULL) &&
	      dataMapIter->second.data.pixVec.empty())
	    {
	      for (unsigned int i=0; i<dataMapIter->second.polygonVec.size(); i++)
		{
		  QPointArray tmpPointArray(dataMapIter->second.polygonVec[i]);
		  tmpPointArray.detach();
		  tmpPointArray.translate(_imgStartPos.x(), _imgStartPos.y());
		  _painter->drawPolygon(tmpPointArray);
		}
	    }
	}

      _painter->end();
    }

    // Update buttons in main application
    if (hasFocus())
    {
      if (_openImgCBFunc != NULL)
        (*_openImgCBFunc)(_openImgUserData, _selectionType, _regionSelected, _allowBuildingImg);
    }
  }
}

// Init buffers that store already selected regions
void SmageImgEncodeWidget::initReservedRegionBuffers()
{   
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::initReservedRegionBuffers()\n");
#endif
 
  _reservedRegionsBitmap = QBitmap(_img.size(), TRUE);
  _reservedRegionsBitmap.fill(Qt::color0);
}

// Class specific operations when an image has been successfully loaded
void SmageImgEncodeWidget::doWhenSuccesImgLoading()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::doWhenSuccesImgLoading()\n");
#endif

  // Initialize buffers that store the regions a user has selected
  initReservedRegionBuffers();

  setCaption(_imgFileName);

  if (_openImgCBFunc != NULL)
    (*_openImgCBFunc)(_openImgUserData, _selectionType, _regionSelected, _allowBuildingImg);
}

// Verify the last selected image region
bool SmageImgEncodeWidget::verifyRegion()
{  
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::verifyRegion()\n");
#endif

  if (_pointVec.empty() || _tmpWidgetRelatedPointArray.isEmpty())
    return TRUE;

  QPointArray tmpPointArray(_pointVec.size());
  for (unsigned int i=0; i<_pointVec.size(); i++)
    tmpPointArray.setPoint(i, _pointVec[i] - _imgStartPos);
  
  // "currentlySelectedRegion" is filled with currently selected polygon
  QBitmap currentlySelectedRegion(_img.size(), TRUE);
  currentlySelectedRegion.fill(Qt::color0);
  QPainter painter;
  painter.begin(&currentlySelectedRegion);
  painter.setPen(Qt::color1);
  painter.setBrush(Qt::color1);
  painter.drawPolygon(tmpPointArray);
  painter.end();

  // "_reservedRegionsBitmap" is filled with all polygonal areas that are already associated with
  // data
  _reservedRegionsBitmap.fill(Qt::color0);
  painter.begin(&_reservedRegionsBitmap);
  painter.setPen(Qt::color1);
  painter.setBrush(Qt::color1);

  map<int, TheS_Data>::const_iterator dataMapIter;
  for (dataMapIter=_dataMap.begin(); dataMapIter!=_dataMap.end(); dataMapIter++)
    {	  
      if ((!dataMapIter->second.data.formTextVec.empty() ||
	   !dataMapIter->second.data.textVec.empty()) &&
	  dataMapIter->second.data.pixVec.empty())
	{
	  for (unsigned int i=0; i<dataMapIter->second.polygonVec.size(); i++)
	    painter.drawPolygon(dataMapIter->second.polygonVec[i]);
	}
    }

  painter.end();
  
  // If regions do intersect, current selected region is not suited
  if (SmageImgHandler::regionsDoNotIntersect(_reservedRegionsBitmap, currentlySelectedRegion))
    {
      if (!_codedRegionsBitmap.isNull())
	return (SmageImgHandler::regionsDoNotIntersect(_codedRegionsBitmap, currentlySelectedRegion));
	  
      return TRUE;
    }
  
  return FALSE;
}

// Verify if none of the regions in the image associated with data overlap
bool SmageImgEncodeWidget::verifyAllRegions()
{  
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::verifyAllRegions()\n");
#endif

  if (_dataMap.empty())
    return TRUE;

  map<int, TheS_Data>::const_iterator dataMapIter;
  unsigned int i;

  // "currentlySelectedRegion" is filled with polygons that are associated with text data
  // but not with image data ("SmageImgSnip")
  QBitmap reservedRegionsBitmap(_img.size(), TRUE);
  reservedRegionsBitmap.fill(Qt::color0);
  QPainter painter;
  painter.begin(&reservedRegionsBitmap);
  painter.setPen(Qt::color1);
  painter.setBrush(Qt::color1);

  for (dataMapIter = _dataMap.begin(); dataMapIter != _dataMap.end(); dataMapIter++)
    {
      if ((!dataMapIter->second.data.formTextVec.empty() ||
	   !dataMapIter->second.data.textVec.empty()) &&
	  dataMapIter->second.data.pixVec.empty())
	{
	  for (i=0; i<dataMapIter->second.polygonVec.size(); i++)
	    painter.drawPolygon(dataMapIter->second.polygonVec[i]);
	}
    }

  painter.end();

  // Next, all regions associated with image data (that could be moved within the image)
  // are compared with the already filled areas

  QBitmap imgSnipRegionsBitmap(_img.size(), TRUE);
  imgSnipRegionsBitmap.fill(Qt::color0);
  painter.begin(&imgSnipRegionsBitmap);
  painter.setPen(Qt::color1);
  painter.setBrush(Qt::color1);

  for (dataMapIter = _dataMap.begin(); dataMapIter != _dataMap.end(); dataMapIter++)
    {
      if (!dataMapIter->second.data.pixVec.empty())
	{
	  for (i=0; i<dataMapIter->second.polygonVec.size(); i++)
	    painter.drawPolygon(dataMapIter->second.polygonVec[i]);
	}
    }
  
  painter.end();

  // If regions do not intersect, return TRUE
  if (SmageImgHandler::regionsDoNotIntersect(reservedRegionsBitmap, imgSnipRegionsBitmap))
    {
      if (!_codedRegionsBitmap.isNull())
	return (SmageImgHandler::regionsDoNotIntersect(_codedRegionsBitmap, imgSnipRegionsBitmap));
      
      return TRUE;
    }
  
  return FALSE;
}

// Assign data to current active image region
void SmageImgEncodeWidget::integratePlainText(const QString &data)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::integratePlainText()\n");
#endif

  // Set the data for the currently selected region(s);
  // "_dataMap" holds all point arrays for this image that are associated with coded information;

  QPoint textHolderPos(_imgStartPos.x(), _imgStartPos.y());

  char *d = getenv("SMAGE_DIR");
  QString smageDir ="";
  if ( d != NULL ) smageDir = QString( d );

  SmageTextHolderWidget *textHolder = new SmageTextHolderWidget(this, "TextHolder", data, textInsertedIcon_xpm);

  map<int, TheS_Data>::iterator dataMapIter = _dataMap.find(_regionID);
  if (dataMapIter != _dataMap.end())
    {
      if (dataMapIter->second.data.dictItem == NULL && dataMapIter->second.data.pixVec.empty())
	{
	  dataMapIter->second.data.textVec.push_back(textHolder);
	  
	  if (!dataMapIter->second.polygonVec.empty())
	    textHolderPos += dataMapIter->second.polygonVec[0].point(0);
	}
      else
	{
	  QMessageBox::information(this, QObject::tr("Smage - Message"), tr("An image dictionary entry or another image is already specified for this region."),
				   QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

	  delete textHolder;
	  return;
	}
    }
  else
    { 
      if (_pointVec.size() < 3)
	{
	  delete textHolder;
	  return;
	}

      textHolderPos = _pointVec[0];

      TheS_Data dataItem;
      dataItem.data.textVec.push_back(textHolder);

      QPointArray imgRelatedPoints(_pointVec.size());
      imgRelatedPoints.putPoints(0, _tmpImgRelatedPointArray.size()-1, _tmpImgRelatedPointArray);
      dataItem.polygonVec.push_back(imgRelatedPoints);

      _tmpImgRelatedPointArray = QPointArray();
      _tmpWidgetRelatedPointArray = QPointArray();

      pair<int, TheS_Data> p(_regionID, dataItem);
      _dataMap.insert(p);

      addNewDataItem(_regionID);
    }

  _shouldIncRegionID = TRUE;

  // Since data has been specified, coded image can be generated
  _allowBuildingImg = TRUE;

  if (_regionSelectedExtCBFunc != NULL)
    (*_regionSelectedExtCBFunc)(_regionSelectedExtUserData, _regionSelected, _allowBuildingImg);

  textHolder->show();
  textHolder->move(textHolderPos);
}


// Encode data in image;
void SmageImgEncodeWidget::buildImg()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::buildImg()\n");
#endif
  
  // Data encoding is done by "_smageGenerator"

  QImage result;
  prepareDataMapForCoding();

  if (!verifyAllRegions())
    {
      QMessageBox::warning(this, QObject::tr("Smage - Message"), tr("Selected regions should not overlap."),
			   QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

      return;
    }

  _smageGenerator->initProgressDlg(this, _dataMap.size() + 3);

  if (_smageGenerator->buildImg(_img, result, &_dataMap, !_isCodedImg, _startId))
  {
    QString fileName;
    if (encodingDone(result, fileName))
    {
      if (_forceOpenNewCBFunc != NULL && _forceOpenNewUserData != NULL)
        (*_forceOpenNewCBFunc)(_forceOpenNewUserData, this, fileName);
    }
  }
  else
  {
#ifdef THE_S_PRINT_DEBUG
    _TheSmageLogFile.lprintf("*** Warning (SmageImgEncodeWidget::buildImg): could not build the image ***\n");
#endif

    QMessageBox::warning(this, QObject::tr("Smage - Message"), tr("A problem occured during generating the coded picture."), QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
  }
}

// Data has been encoded in the image ... need a file name for storing
bool SmageImgEncodeWidget::encodingDone(const QImage &img, QString &fileName)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::encodingDone()\n");
#endif

  if (SmageEncDoneDlg::getFileName(fileName, this))
  {
    bool validEnding;
    QString ending = getFileEnding(fileName, validEnding);

    if (!validEnding || !ending.contains("png"))
      fileName += ".png";
    SmageImgHandler::saveImgToFile(fileName, img, "PNG");

    return TRUE;
  }

  return FALSE;
}

// Before closing the widget ...
//  - ask if coded image should be generated, if not yet done
void SmageImgEncodeWidget::closeEvent(QCloseEvent *event)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::closeEvent()\n");
#endif

  // Coded image has not yet been generated
  if (!_dataMap.empty())
  {
    QString s_msg;
    s_msg = QString(tr("Image parts are associated with data.\nWould you like to generate the coded image now?"));

    switch (QMessageBox::information(this, QObject::tr("Smage - Message"), s_msg, QObject::tr("Yes"), QObject::tr("No"), QString::null, 0, 1))
    {
      case 0: 
      {
        buildImg();
        break;
      }
    }
  }

  inherited::closeEvent(event);
}

// Reset the image widgets data
void SmageImgEncodeWidget::resetImgWidget()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::resetImgWidget()\n");
#endif

  _reservedRegionsBitmap.fill(Qt::color0);
  inherited::resetImgWidget();
}

// When resizing the widget:
//  - user selected regions need to be updated
void SmageImgEncodeWidget::resizeEvent(QResizeEvent *event)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::resizeEvent()\n");
#endif

  inherited::resizeEvent(event);

  unsigned int i;

  map<int, TheS_Data>::iterator dataMapIter;
  for (dataMapIter=_dataMap.begin(); dataMapIter!=_dataMap.end(); dataMapIter++)
    {
      for (i=0; i<dataMapIter->second.data.pixVec.size(); i++)
	{
	  SmageImgSnip *imgSnip = dataMapIter->second.data.pixVec[i].imgSnip;
	  imgSnip->move(imgSnip->pos().x()+_resizeDiff.width(),
			imgSnip->pos().y()+_resizeDiff.height());
	  imgSnip->setParentImgStartPos(_imgStartPos);
	}
      for (i=0; i<dataMapIter->second.data.textVec.size(); i++)
	{
	  SmageTextHolderWidget *th = (SmageTextHolderWidget *) dataMapIter->second.data.textVec[i];
	  th->move(th->pos().x()+_resizeDiff.width(), th->pos().y()+_resizeDiff.height());
	}
      for (i=0; i<dataMapIter->second.data.formTextVec.size(); i++)
	{
	  SmageShapedTextBase *st = 
	    (SmageShapedTextBase *) dataMapIter->second.data.formTextVec[i].shapedText;
	  st->move(st->pos().x()+_resizeDiff.width(), st->pos().y()+_resizeDiff.height());
	}
      if (dataMapIter->second.data.dictItem != NULL)
	{
	  SmageTextHolderWidget *dth = (SmageTextHolderWidget *) dataMapIter->second.data.dictItem;
	  dth->move(dth->pos().x()+_resizeDiff.width(), dth->pos().y()+_resizeDiff.height());
	}
    }
}


void SmageImgEncodeWidget::integrateImg(const QImage* img, const QPointArray &points)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::integrateImg()\n");
#endif

  if (img->isNull() || points.isEmpty())
    return;

  // Show a new widget containing the image's part to be integrated
  SmageImgSnip *imgSnip = new SmageImgSnip(this, "ImgSnip", _imgStartPos, _imgPointer->size(),
					   Qt::WStyle_Customize | Qt::WStyle_NoBorder | Qt::WStyle_StaysOnTop);

  QRect geomRect = rect();
  QPoint pos = _imgStartPos;
  bool sameSize = TRUE;

  if (_imgPointer->width() != img->width() || _imgPointer->height() != img->height())
    {
      pos = QPoint(int((geomRect.x()+geomRect.width())/2.0), int((geomRect.y()+geomRect.height())/2.0));
      sameSize = FALSE;
    }

  if (!imgSnip->setImg(img, points, pos, sameSize))
    {
      delete imgSnip;
      return;
    }

  imgSnip->show();

  TheS_ImgDataItem pixDataItem;
  imgSnip->getImgData(pixDataItem.image, pixDataItem.regionPoints);
  pixDataItem.imgSnip = imgSnip;

  TheS_Data dataItem;
  dataItem.data.pixVec.push_back(pixDataItem);
  
  // Map image region to widget's image coordinates
  if (!pixDataItem.regionPoints.isEmpty())
    {
      incrementRegionID();
      _shouldIncRegionID = TRUE;

      pair<int, TheS_Data> p(_regionID, dataItem);
      _dataMap.insert(p);

      addNewDataItem(_regionID);
    }

  _allowBuildingImg = TRUE;
}

void SmageImgEncodeWidget::integrateFormattedText(const TextShapeType &textType)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::integrateFormattedText()\n");
#endif

  bool matchedTextType = FALSE;

  // Find a position at which the bubble can be initially positioned;
  // if "pointVec" is not empty, take its first point;
  // if "pointVec" is empty, the bubble is associated with an integrated image snip ... take its position
  QPoint textPos(0,0);
  if (!_pointVec.empty())
    textPos = _pointVec[0];
  else
    {
      map<int, TheS_Data>::iterator dataMapIter = _dataMap.find(_regionID);
      if (dataMapIter != _dataMap.end())
	{
	  if (!dataMapIter->second.data.pixVec.empty())
	    textPos = dataMapIter->second.data.pixVec[0].imgSnip->pos();
	}
    }

  // Create a formated text widget (a bubble)
  if (textType == SMAGE_SPEECH_BUBBLE)
    {
      SmageEditorSpeechBubble *aBubble = new SmageEditorSpeechBubble(this, "SmageEditorSpeechBubble");
      insertDataItem(_regionID, aBubble);
      aBubble->show();
      aBubble->move(textPos);
      matchedTextType = TRUE;
    }

  else if (textType == SMAGE_THINK_BUBBLE)
    {
      SmageEditorThinkBubble *aBubble = new SmageEditorThinkBubble(this, "SmageEditorThinkBubble");
      insertDataItem(_regionID, aBubble);
      aBubble->show();
      aBubble->move(textPos);
      matchedTextType = TRUE;
    }

  else if (textType == SMAGE_ELLIPSE || textType == SMAGE_RECT)
    {
      SmageEditorSimpleBubble *aBubble = new SmageEditorSimpleBubble(this, "SmageEditorSimpleBubble", textType);
      insertDataItem(_regionID, aBubble);
      aBubble->show();
      aBubble->move(textPos);
      matchedTextType = TRUE;
    }


  if (matchedTextType)
    {
      _shouldIncRegionID = TRUE;
      _allowBuildingImg = TRUE;
    }
}

void SmageImgEncodeWidget::insertDataItem(const int &id, SmageEditorShapedText *shapedText)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::insertDataItem()\n");
#endif

  TheS_TextDataItem textDataItem;
  textDataItem.shapedText = shapedText;

  map<int, TheS_Data>::iterator dataMapIter = _dataMap.find(id);
  if (dataMapIter != _dataMap.end())
    dataMapIter->second.data.formTextVec.push_back(textDataItem);
  else
    {
      if (_pointVec.size() < 3) return;

      TheS_Data dataItem;
      dataItem.data.formTextVec.push_back(textDataItem);

      QPointArray imgRelatedPoints(_pointVec.size());
      imgRelatedPoints.putPoints(0, _tmpImgRelatedPointArray.size()-1, _tmpImgRelatedPointArray);
      dataItem.polygonVec.push_back(imgRelatedPoints);

      _tmpImgRelatedPointArray = QPointArray();
      _tmpWidgetRelatedPointArray = QPointArray();

      pair<int, TheS_Data> p(id, dataItem);
      _dataMap.insert(p);

      addNewDataItem(id);
    }
}


void SmageImgEncodeWidget::prepareDataMapForCoding()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::prepareDataMapForCoding()\n");
#endif

  map<int, TheS_Data>::iterator dataMapIter;
  vector<TheS_TextDataItem>::iterator textDataItemVecIter;
  vector<TheS_ImgDataItem>::iterator imgDataItemVecIter;
  
  for (dataMapIter=_dataMap.begin(); dataMapIter!=_dataMap.end(); dataMapIter++)
    {
      for (textDataItemVecIter=dataMapIter->second.data.formTextVec.begin();
	   textDataItemVecIter!=dataMapIter->second.data.formTextVec.end(); textDataItemVecIter++)
	{
	  if (textDataItemVecIter->shapedText->getClassName() == "SmageEditorSpeechBubble")
	    {	  	  
	      textDataItemVecIter->shapedTextString =
		((SmageEditorSpeechBubble *)(textDataItemVecIter->shapedText))->getBubbleData(_imgStartPos);
	    }
	  else if (textDataItemVecIter->shapedText->getClassName() == "SmageEditorThinkBubble")
	    {
	      textDataItemVecIter->shapedTextString = 
		((SmageEditorThinkBubble *)(textDataItemVecIter->shapedText))->getBubbleData(_imgStartPos);
	    }
	  else if (textDataItemVecIter->shapedText->getClassName() == "SmageEditorSimpleBubble")
	    {
	      textDataItemVecIter->shapedTextString = 
		((SmageEditorSimpleBubble *)(textDataItemVecIter->shapedText))->getBubbleData(_imgStartPos);
	    }	      
	}

      if (!dataMapIter->second.data.pixVec.empty())
	dataMapIter->second.polygonVec.clear();

      for (imgDataItemVecIter=dataMapIter->second.data.pixVec.begin();
	   imgDataItemVecIter!=dataMapIter->second.data.pixVec.end(); imgDataItemVecIter++)
	{
	  QPointArray regionPointsCopy = imgDataItemVecIter->regionPoints.copy();
	  QPoint refPoint = imgDataItemVecIter->imgSnip->mapToParent(regionPointsCopy.point(0)) - _imgStartPos;
	  int x_diff = regionPointsCopy.point(0).x() - refPoint.x();
	  int y_diff = regionPointsCopy.point(0).y() - refPoint.y();

	  regionPointsCopy.translate(-x_diff, -y_diff);
	  dataMapIter->second.polygonVec.push_back(regionPointsCopy);
	}
    }
}


void SmageImgEncodeWidget::deleteDataItem(SmageEditorShapedText *shapedText)
{
  map<int, TheS_Data>::iterator dataMapIter;
  for (dataMapIter = _dataMap.begin(); dataMapIter != _dataMap.end(); dataMapIter++)
    {
      std::vector<TheS_TextDataItem>::iterator itemIter;
      for (itemIter=dataMapIter->second.data.formTextVec.begin(); 
	   itemIter!=dataMapIter->second.data.formTextVec.end(); itemIter++)
	{
	  if (itemIter->shapedText == shapedText)
	    {
	      dataMapIter->second.data.formTextVec.erase(itemIter);
	      shapedText->close();

	      if (dataMapIter->second.data.formTextVec.empty() &&
		  dataMapIter->second.data.pixVec.empty() &&
		  dataMapIter->second.data.textVec.empty() &&
		  dataMapIter->second.data.dictItem == NULL)
		{
		  _dataManageWidget->removeItem(dataMapIter->first);
		  _dataMap.erase(dataMapIter);
		}

	      return;
	    }
	}	
    }
}


void SmageImgEncodeWidget::addNewDataItem(const int &id)
{
  QString dataItemName = QString(tr("Data %1")).arg(id);
  _dataManageWidget->addItem(dataItemName, id);
}


void SmageImgEncodeWidget::dataManageWidgetButtonPressedCB(void *userData, const int &whichButton, 
							   const int &itemID)
{
  SmageImgEncodeWidget *widget = (SmageImgEncodeWidget *) userData;
  if (widget == NULL)
    return;
  
  map<int, TheS_Data>::iterator dataMapIter = widget->_dataMap.find(itemID);
  if (dataMapIter == widget->_dataMap.end())
    return;
    
  vector<TheS_ImgDataItem>::iterator imgDataItemIter;
  vector<TheS_TextDataItem>::iterator textDataItemIter;
  vector<SmageTextHolderWidget *>::iterator simpleTextDataItemIter;

  if (whichButton == 2)
    {
      for (imgDataItemIter=dataMapIter->second.data.pixVec.begin();
	   imgDataItemIter!=dataMapIter->second.data.pixVec.end(); imgDataItemIter++)
	imgDataItemIter->imgSnip->close(TRUE);
      for (textDataItemIter=dataMapIter->second.data.formTextVec.begin();
	   textDataItemIter!=dataMapIter->second.data.formTextVec.end(); textDataItemIter++)
	textDataItemIter->shapedText->close(TRUE);
      for (simpleTextDataItemIter=dataMapIter->second.data.textVec.begin();
	   simpleTextDataItemIter!=dataMapIter->second.data.textVec.end(); simpleTextDataItemIter++)
	(*simpleTextDataItemIter)->close(TRUE);
      if (dataMapIter->second.data.dictItem != NULL)
	{
	  dataMapIter->second.data.dictItem->close(TRUE);
	  dataMapIter->second.data.dictItem = NULL;
	}

      widget->_dataMap.erase(dataMapIter);

      if (widget->_dataMap.empty())
	widget->_allowBuildingImg = FALSE;

      widget->_regionSelected = FALSE;
      if (widget->_regionSelectedExtCBFunc != NULL)
	(*widget->_regionSelectedExtCBFunc)(widget->_regionSelectedExtUserData, FALSE, widget->_allowBuildingImg);
    }
  else
    {	  
      if (whichButton == 0) 
	{
	  dataMapIter->second.disabled = TRUE;
	  
	  for (imgDataItemIter=dataMapIter->second.data.pixVec.begin();
	       imgDataItemIter!=dataMapIter->second.data.pixVec.end(); imgDataItemIter++)
	    imgDataItemIter->imgSnip->hide();
	  for (textDataItemIter=dataMapIter->second.data.formTextVec.begin();
	       textDataItemIter!=dataMapIter->second.data.formTextVec.end(); textDataItemIter++)
	    textDataItemIter->shapedText->hide();
	  for (simpleTextDataItemIter=dataMapIter->second.data.textVec.begin();
	       simpleTextDataItemIter!=dataMapIter->second.data.textVec.end(); simpleTextDataItemIter++)
	    (*simpleTextDataItemIter)->hide();
	  if (dataMapIter->second.data.dictItem != NULL) 
	    dataMapIter->second.data.dictItem->hide();
	}
      else 
	{
	  dataMapIter->second.disabled = FALSE;

	  for (imgDataItemIter=dataMapIter->second.data.pixVec.begin();
	       imgDataItemIter!=dataMapIter->second.data.pixVec.end(); imgDataItemIter++)
	    imgDataItemIter->imgSnip->show();
	  for (textDataItemIter=dataMapIter->second.data.formTextVec.begin();
	       textDataItemIter!=dataMapIter->second.data.formTextVec.end(); textDataItemIter++)
	    textDataItemIter->shapedText->show();
	  for (simpleTextDataItemIter=dataMapIter->second.data.textVec.begin();
	       simpleTextDataItemIter!=dataMapIter->second.data.textVec.end(); simpleTextDataItemIter++)
	    (*simpleTextDataItemIter)->show();
	  if (dataMapIter->second.data.dictItem != NULL) 
	    dataMapIter->second.data.dictItem->show();

	}
    }
}


void SmageImgEncodeWidget::dataManageWidgetFocusCB(void *userData, const int &whichId)
{
  SmageImgEncodeWidget *widget = (SmageImgEncodeWidget *) userData;
  if (widget == NULL)
    return;
  
  if (whichId == -1)
    return;

  widget->_pointVec.clear();
  widget->_tmpImgRelatedPointArray = QPointArray();
  widget->_tmpWidgetRelatedPointArray = QPointArray();

  widget->_regionID = whichId;

  widget->_regionSelected = TRUE;
  if (widget->_regionSelectedExtCBFunc != NULL)
    (*widget->_regionSelectedExtCBFunc)(widget->_regionSelectedExtUserData, TRUE, widget->_allowBuildingImg);

  widget->update();
}


void SmageImgEncodeWidget::incrementRegionID()
{
  if (_dataMap.empty())
    _regionID = 1;
  else
    {
      map<int, TheS_Data>::const_iterator dataMapIter = _dataMap.end();
      dataMapIter--;
      _regionID = dataMapIter->first + 1;
    }
}


void SmageImgEncodeWidget::setCodedImg(const QImage &outlineImg, const QImage &idBuffer, 
				       const int &startId)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("(SmageImgEncodeWidget::setCodedImg)\n");
#endif

  if (outlineImg.size() != idBuffer.size() ||
      outlineImg.size() != _img.size())
    {
#ifdef THE_S_PRINT_DEBUG
      _TheSmageLogFile.lprintf("*** Warning (SmageImgEncodeWidget::setCodedImg): image and ID buffer have different sizes ***\n");
#endif
      return;
    }

  _startId = startId;

  // Copy the image that shows the already coded regions
  _outlineImg = QImage(outlineImg);
  _outlineImg.detach();

  // Fill the "_codedRegionsBitmap"
  _codedRegionsBitmap = QBitmap(idBuffer.size(), TRUE);
  _codedRegionsBitmap.fill(Qt::color0);

  int currentColIndex;
  QPainter painter;
  painter.begin(&_codedRegionsBitmap);
  painter.setPen(Qt::color1);

  for (int y=0; y<idBuffer.height(); y++)
    for (int x=0; x<idBuffer.width(); x++)
      {
	currentColIndex = idBuffer.pixelIndex(x,y);
	
	if (currentColIndex != COL_BACKGROUND_INDEX)
	  painter.drawPoint(x, y);
      }

  painter.end();

  // Show the image with the highlighted regions already coded
  _imgPointer = &_outlineImg;

  if (_openImgCBFunc != NULL)
    (*_openImgCBFunc)(_openImgUserData, _selectionType, _regionSelected, _allowBuildingImg);
}


void SmageImgEncodeWidget::doInsertDictEntry()
{
  char *d = getenv("SMAGE_DIR");
  QString smageDir ="";
  if ( d != NULL ) smageDir = QString( d );

  SmageTextHolderWidget *textHolder = new SmageTextHolderWidget(this, "TextHolder", QString::null, dictInsertedIcon_xpm, TRUE);
  
  if (textHolder->getDictEntry())
    {
      if (!integrateDictEntry(textHolder))
	delete textHolder;
    }
  else
    delete textHolder;
}

bool SmageImgEncodeWidget::integrateDictEntry(SmageTextHolderWidget *textHolder)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgEncodeWidget::integrateDictEntry()\n");
#endif

  QPoint textHolderPos(_imgStartPos.x(), _imgStartPos.y());

  map<int, TheS_Data>::iterator dataMapIter = _dataMap.find(_regionID);
  if (dataMapIter != _dataMap.end())
    {
      if (dataMapIter->second.data.textVec.empty() && dataMapIter->second.data.pixVec.empty())
	{
	  dataMapIter->second.data.dictItem = textHolder;

	  if (!dataMapIter->second.polygonVec.empty())
	    textHolderPos += dataMapIter->second.polygonVec[0].point(0);
	}
      else
	{
	  QMessageBox::information(this, QObject::tr("Smage - Message"), tr("Unformatted text or another image is already specified for this region."),
				   QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
	  return FALSE;
	}
    }
  else
    {      
      if (_pointVec.size() < 3)
	return FALSE;

      textHolderPos = _pointVec[0];

      TheS_Data dataItem;
      dataItem.data.dictItem = textHolder;

      QPointArray imgRelatedPoints(_pointVec.size());
      imgRelatedPoints.putPoints(0, _tmpImgRelatedPointArray.size()-1, _tmpImgRelatedPointArray);
      dataItem.polygonVec.push_back(imgRelatedPoints);

      _tmpImgRelatedPointArray = QPointArray();
      _tmpWidgetRelatedPointArray = QPointArray();

      pair<int, TheS_Data> p(_regionID, dataItem);
      _dataMap.insert(p);

      addNewDataItem(_regionID);
    }

  _shouldIncRegionID = TRUE;

  // Since data has been specified, coded image can be generated
  _allowBuildingImg = TRUE;

  if (_regionSelectedExtCBFunc != NULL)
    (*_regionSelectedExtCBFunc)(_regionSelectedExtUserData, _regionSelected, _allowBuildingImg);

  textHolder->show();
  textHolder->move(textHolderPos);

  return TRUE;
}
