// project header
#include "SmageImgHandler.h"

// qt header
#include <qbuffer.h>
#include <qpainter.h>
#include <qpointarray.h>

// c/c++ header
#include <assert.h>
#include <iostream>
#include <vector>

using namespace std;

SmageImgHandler::SmageImgHandler()
{ 
}

SmageImgHandler::~SmageImgHandler()
{ 
}

// Do the filled regions in "bitmap" overlap with filled regions in "img"
bool SmageImgHandler::regionsDoNotIntersect(QImage &img, const QBitmap &bitmap)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgHandler::regionsDoNotIntersect()\n");
#endif

  // Pixel operations can only be done in "QImage"
  QImage tmpImg = bitmap.convertToImage();

  if (tmpImg.height() != img.height() || tmpImg.width() != img.width())
    return FALSE;

  int currentColIndexImg, currentColIndexBitmap;

  // Compare the pixels in "img" and "tmpImg"
  for (int y=0; y<tmpImg.height(); y++)
    for (int x=0; x<tmpImg.width(); x++)
    {
      currentColIndexImg = img.pixelIndex(x,y);
      currentColIndexBitmap = tmpImg.pixelIndex(x,y);
  
      if (currentColIndexImg == 1 && currentColIndexBitmap == 1)
        return FALSE;
    }

  return TRUE;
}

//  Do the filled regions in "bitmap1" overlap with filled regions in "bitmap2"
bool SmageImgHandler::regionsDoNotIntersect(const QBitmap &bitmap1, const QBitmap &bitmap2)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgHandler::regionsDoNotIntersect()\n");
#endif

  // Pixel operations can only be done in "QImage"
  QImage tmpImg1 = bitmap1.convertToImage();
  QImage tmpImg2 = bitmap2.convertToImage();

  if (tmpImg1.height() != tmpImg2.height() || tmpImg1.width() != tmpImg2.width())
    return FALSE;

  int currentColIndexImg1, currentColIndexImg2;

  // Compare the pixels in "img" and "tmpImg"
  for (int y=0; y<tmpImg1.height(); y++)
    for (int x=0; x<tmpImg1.width(); x++)
    {
      currentColIndexImg1 = tmpImg1.pixelIndex(x,y);
      currentColIndexImg2 = tmpImg2.pixelIndex(x,y);
  
      if (currentColIndexImg1 == 1 && currentColIndexImg2 == 1)
        return FALSE;
    }

  return TRUE;
}

// Copy "bitmap"'s filled regions to "img"
void SmageImgHandler::fillRegionsInImg(QImage &img, const QBitmap &bitmap)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgHandler::fillRegionsInImg()\n");
#endif

  QImage tmpImg = bitmap.convertToImage();

  if (tmpImg.height() != img.height() || tmpImg.width() != img.width())
    return;

  int currentColIndexBitmap;

  for (int y=0; y<tmpImg.height(); y++)
    for (int x=0; x<tmpImg.width(); x++)
    {
      currentColIndexBitmap = tmpImg.pixelIndex(x,y);
  
      if (currentColIndexBitmap == 1)
        img.setPixel(x,y, currentColIndexBitmap);
    }
}

// Try to compress the image as much as possible without loss
bool SmageImgHandler::saveImgToFile(const QString &fileName, const QImage &img, const QString &fileFormat)
{
  QImageIO imgIO_In;
  imgIO_In.setImage(img);
  imgIO_In.setFormat(fileFormat);
  
  int comprRate = 70;

  while (comprRate <= 100)
    {  
      QBuffer aBuffer;
      aBuffer.open(IO_WriteOnly);

      imgIO_In.setIODevice(&aBuffer);
      imgIO_In.setQuality(comprRate);
      imgIO_In.write();

      QByteArray a = aBuffer.buffer();
      QBuffer aBuffer1(a);
      aBuffer1.open(IO_ReadOnly);

      QImageIO imgIO_Out(&aBuffer1, fileFormat);
      imgIO_Out.read();

      aBuffer.close();
      aBuffer1.close();

      if (imgIO_Out.image() == img)
	{
	  img.save(fileName, fileFormat, comprRate);
	  return TRUE;
	}

      comprRate += 10;      
    }

  return FALSE;
}


bool SmageImgHandler::findOutline(QImage *img, QPointArray &points, const QRgb &color)
{
  if (img->isNull())
    return FALSE;

  QImage *imgPointer = img;
  QImage tmpImage;

  if (img->depth() < 32)
    {
      tmpImage = img->convertDepth(32);
      imgPointer = &tmpImage;
    }

  QPoint startPos(-1,-1);
  vector<QPoint> pointVec;

  QImage filledBitmap(img->width(), img->height(), 8, 3);
  filledBitmap.setColor(0, qRgb(255,255,255)); // Background
  filledBitmap.setColor(1, qRgb(0,0,0)); // Marked as outline
  filledBitmap.setColor(2, qRgb(0,255,255)); // Marked as visited non-background
  filledBitmap.fill(0);

  // Search for the outline's first point which serves as start point for the search 
  // for the other outline points 
  for (int y=0; y<imgPointer->height(); y++)
    {
      if (startPos.x() >= 0)
	break;
      for (int x=0; x<imgPointer->width(); x++)
	{
	  if (isOutlinePixel(imgPointer, filledBitmap, QPoint(x,y), color))
	    {
	      filledBitmap.setPixel(x,y, 1);
	      startPos = QPoint(x,y);
	      pointVec.push_back(startPos);
	      break;
	    }
	}
    }

  // Was a start outline position found?
  if (startPos.x() < 0 || startPos.y() < 0)
    return FALSE;
  
  QPoint nextPos;
  QPoint curPos = startPos;
  while (searchForNextOutlinePixel(img, filledBitmap, curPos, nextPos))
    {
      pointVec.push_back(nextPos);
      curPos = nextPos;
    }

  if (pointVec.size() < 3)
    return FALSE;

  if (areNeighbors(startPos, *(pointVec.end()-1)))
    {
      points = QPointArray(pointVec.size());

      for (unsigned int k=0; k<pointVec.size(); k++)
	points.setPoint(k, pointVec[k]);

      return TRUE;
    }

  return FALSE;
}

bool SmageImgHandler::searchForNextOutlinePixel(const QImage *img, QImage &bitmap, 
						const QPoint &curPos, QPoint &nextPos)
{
  QRgb curCol = img->pixel(curPos.x(), curPos.y());

  if (img->valid(curPos.x()+1, curPos.y()))
    {
      if (bitmap.pixelIndex(curPos.x()+1, curPos.y()) == 0)
	{
	  if (isVisitedNoneBackAround(bitmap, QPoint(curPos.x()+1, curPos.y())))
	    {
	      if (isOutlinePixel(img, bitmap, QPoint(curPos.x()+1, curPos.y()), curCol))
		{
		  nextPos = QPoint(curPos.x()+1, curPos.y());
		  bitmap.setPixel(curPos.x()+1, curPos.y(), 1);
		  return TRUE;
		}
	    }
	}
    }

  if (img->valid(curPos.x(), curPos.y()+1))
    {
      if (bitmap.pixelIndex(curPos.x(), curPos.y()+1) == 0)
	{
	  if (isVisitedNoneBackAround(bitmap, QPoint(curPos.x(), curPos.y()+1)))
	    {
	      if (isOutlinePixel(img, bitmap, QPoint(curPos.x(), curPos.y()+1), curCol))
		{
		  nextPos = QPoint(curPos.x(), curPos.y()+1);
		  bitmap.setPixel(curPos.x(), curPos.y()+1, 1);
		  return TRUE;
		}
	    }
	}
    }

  if (img->valid(curPos.x()-1, curPos.y()))
    {
      if (bitmap.pixelIndex(curPos.x()-1, curPos.y()) == 0)
	{
	  if (isVisitedNoneBackAround(bitmap, QPoint(curPos.x()-1, curPos.y())))
	    {
	      if (isOutlinePixel(img, bitmap, QPoint(curPos.x()-1, curPos.y()), curCol))
		{
		  nextPos = QPoint(curPos.x()-1, curPos.y());
		  bitmap.setPixel(curPos.x()-1, curPos.y(), 1);
		  return TRUE;
		}
	    }
	}
    }

  if (img->valid(curPos.x(), curPos.y()-1))
    {
      if (bitmap.pixelIndex(curPos.x(), curPos.y()-1) == 0)
	{
	  if (isVisitedNoneBackAround(bitmap, QPoint(curPos.x(), curPos.y()-1)))
	    {
	      if (isOutlinePixel(img, bitmap, QPoint(curPos.x(), curPos.y()-1), curCol))
		{
		  nextPos = QPoint(curPos.x(), curPos.y()-1);
		  bitmap.setPixel(curPos.x(), curPos.y()-1, 1);
		  return TRUE;
		}
	    }
	}
    }

  if (img->valid(curPos.x()+1, curPos.y()+1))
    {
      if (bitmap.pixelIndex(curPos.x()+1, curPos.y()+1) == 0)
	{
	  if (isVisitedNoneBackAround(bitmap, QPoint(curPos.x()+1, curPos.y()+1)))
	    {
	      if (isOutlinePixel(img, bitmap, QPoint(curPos.x()+1, curPos.y()+1), curCol))
		{
		  nextPos = QPoint(curPos.x()+1, curPos.y()+1);
		  bitmap.setPixel(curPos.x()+1, curPos.y()+1, 1);
		  return TRUE;
		}
	    }
	}
    }

  if (img->valid(curPos.x()-1, curPos.y()+1))
    {
      if (bitmap.pixelIndex(curPos.x()-1, curPos.y()+1) == 0)
	{
	  if (isVisitedNoneBackAround(bitmap, QPoint(curPos.x()-1, curPos.y()+1)))
	    {
	      if (isOutlinePixel(img, bitmap, QPoint(curPos.x()-1, curPos.y()+1), curCol))
		{
		  nextPos = QPoint(curPos.x()-1, curPos.y()+1);
		  bitmap.setPixel(curPos.x()-1, curPos.y()+1, 1);
		  return TRUE;
		}
	    }
	}
    }

  if (img->valid(curPos.x()-1, curPos.y()-1))
    {
      if (bitmap.pixelIndex(curPos.x()-1, curPos.y()-1) == 0)
	{
	  if (isVisitedNoneBackAround(bitmap, QPoint(curPos.x()-1, curPos.y()-1)))
	    {
	      if (isOutlinePixel(img, bitmap, QPoint(curPos.x()-1, curPos.y()-1), curCol))
		{
		  nextPos = QPoint(curPos.x()-1, curPos.y()-1);
		  bitmap.setPixel(curPos.x()-1, curPos.y()-1, 1);
		  return TRUE;
		}
	    }
	}
    }

  if (img->valid(curPos.x()+1, curPos.y()-1))
    {
      if (bitmap.pixelIndex(curPos.x()+1, curPos.y()-1) == 0)
	{
	  if (isVisitedNoneBackAround(bitmap, QPoint(curPos.x()+1, curPos.y()-1)))
	    {
	      if (isOutlinePixel(img, bitmap, QPoint(curPos.x()+1, curPos.y()-1), curCol))
		{
		  nextPos = QPoint(curPos.x()+1, curPos.y()-1);
		  bitmap.setPixel(curPos.x()+1, curPos.y()-1, 1);
		  return TRUE;
		}
	    }
	}
    }

  return FALSE;
}

bool SmageImgHandler::isOutlinePixel(const QImage *img, QImage &bitmap, 
				     const QPoint &pos, const QRgb &color)
{
  // Compare the colors of the focused pixel at "pos" and its four neighbors;
  // if one of the neighbor pixels has a different color, focused pixel is an outline pixel

  // All four neighbors of focused pixel are visited and marked if they are none-background pixels

  if (!img->valid(pos.x(), pos.y()))
    return FALSE;

  bool result = FALSE;

  QRgb curCol = img->pixel(pos.x(), pos.y());

  if (!compareRgb(curCol, color))
    return FALSE;

  if (img->valid(pos.x()+1, pos.y()))
    if (!compareRgb(img->pixel(pos.x()+1, pos.y()), color))
      {
	bitmap.setPixel(pos.x()+1, pos.y(), 2);
	result = TRUE;
      }
  if (img->valid(pos.x()-1, pos.y()))
    if (!compareRgb(img->pixel(pos.x()-1, pos.y()), color))
      {
	bitmap.setPixel(pos.x()-1, pos.y(), 2);
	result = TRUE;
      }
  if (img->valid(pos.x(), pos.y()+1))
    if (!compareRgb(img->pixel(pos.x(), pos.y()+1), color))
      {
	bitmap.setPixel(pos.x(), pos.y()+1, 2);
	result = TRUE;
      }
  if (img->valid(pos.x(), pos.y()-1))
    if (!compareRgb(img->pixel(pos.x(), pos.y()-1), color))
      {
	bitmap.setPixel(pos.x(), pos.y()-1, 2);
	result = TRUE;
      }

  return result;
}


bool SmageImgHandler::areNeighbors(const QPoint &p1, const QPoint &p2)
{
  return ((p1.x()==p2.x() || p1.x()+1==p2.x() || p1.x()-1==p2.x()) &&
	  (p1.y()==p2.y() || p1.y()+1==p2.y() || p1.y()-1==p2.y()));
}

bool SmageImgHandler::compareRgb(const QRgb &c1, const QRgb &c2)
{
  return (qRed(c1)==qRed(c2) && qGreen(c1)==qGreen(c2) && qBlue(c1)==qBlue(c2));
}


bool SmageImgHandler::isVisitedNoneBackAround(const QImage &bitmap, const QPoint &pos)
{
  if (bitmap.valid(pos.x()+1, pos.y()+1))
    if (bitmap.pixelIndex(pos.x()+1, pos.y()+1) == 2)
      return TRUE;
  if (bitmap.valid(pos.x()-1, pos.y()+1))
    if (bitmap.pixelIndex(pos.x()-1, pos.y()+1) == 2)
      return TRUE;
  if (bitmap.valid(pos.x()-1, pos.y()-1))
    if (bitmap.pixelIndex(pos.x()-1, pos.y()-1) == 2)
      return TRUE;
  if (bitmap.valid(pos.x()+1, pos.y()-1))
    if (bitmap.pixelIndex(pos.x()+1, pos.y()-1) == 2)
      return TRUE;

  if (bitmap.valid(pos.x()+1, pos.y()))
    if (bitmap.pixelIndex(pos.x()+1, pos.y()) == 2)
      return TRUE;
  if (bitmap.valid(pos.x(), pos.y()+1))
    if (bitmap.pixelIndex(pos.x(), pos.y()+1) == 2)
      return TRUE;
  if (bitmap.valid(pos.x()-1, pos.y()))
    if (bitmap.pixelIndex(pos.x()-1, pos.y()) == 2)
      return TRUE;
  if (bitmap.valid(pos.x(), pos.y()-1))
    if (bitmap.pixelIndex(pos.x(), pos.y()-1) == 2)
      return TRUE;

  return FALSE;
}
