// project header
#include "SmageImgSelectionWidget.h"

// qt header
#include <qmessagebox.h>
#include <qpen.h>

// c/c++ header
#include <assert.h>
#include <iostream>

using namespace std;

SmageImgSelectionWidget::SmageImgSelectionWidget(QWidget *parent, const char* name, WFlags f)
  : SmageImgWidget(parent, name, f)
{ 
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgSelectionWidget::SmageImgSelectionWidget()\n");
#endif

  setMouseTracking(FALSE);

  _pen = new QPen(black, 2, Qt::DotLine);
  _selectionType = SELECT_RECT;

  _rubberOn = FALSE;
  _mouseHasBeenMoved = FALSE;
  _regionSelected = FALSE;
  _sceneChanged = TRUE;
  _firstMousePress = FALSE;
  _regionID = 1;
  _shouldIncRegionID = FALSE;
  _allowRegionSelection = TRUE;

  _rubberStart = _rubberEnd = QPoint(0,0);

  _pointVec.clear();
  
  // Initialize callbacks
  _regionSelectedCBFunc = NULL;
  _regionSelectedUserData = NULL;

  _tmpImgRelatedPointArray = QPointArray();
  _tmpWidgetRelatedPointArray = QPointArray();
}


SmageImgSelectionWidget::~SmageImgSelectionWidget()
{ 
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgSelectionWidget::~SmageImgSelectionWidget()\n");
#endif

  delete _pen;
}


void SmageImgSelectionWidget::paintEvent(QPaintEvent *)
{
  if (_imgLoaded)
  {
    // Do repainting several things
    if (_sceneChanged)
    {
      _painter->begin(this);

      // Draw the actual image (pointer to original image or original image with outlines in it)
      if (_imgPointer != NULL)
        _painter->drawImage(_imgStartPos, *_imgPointer);

      if (_allowRegionSelection)
      {    
        // If image has no data encoded, user selection lines are drawn

        _painter->setRasterOp(NotROP);
        _pen->setStyle(Qt::DotLine);
        _painter->setPen(*_pen);

        // "_tmpWidgetRelatedPointArray" is the current outline; no data specified yet
        if (!_tmpWidgetRelatedPointArray.isEmpty())
	  _painter->drawPolyline(_tmpWidgetRelatedPointArray);
      }

      _painter->end();
    }
  }
}

// When resizing the widget:
//  - user selected regions need to be updated
void SmageImgSelectionWidget::resizeEvent(QResizeEvent *event)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgSelectionWidget::resizeEvent()\n");
#endif

  inherited::resizeEvent(event);

  for (unsigned int k=0; k<_tmpImgRelatedPointArray.size(); k++)
    _tmpWidgetRelatedPointArray.setPoint(k, _tmpImgRelatedPointArray.point(k) + _imgStartPos);
}

// Mouse press event has only effect, if the image is not coded with data;
//  otherwise, start position of outline is set
void SmageImgSelectionWidget::mousePressEvent(QMouseEvent *event)
{
  if (!_firstMousePress)
    _firstMousePress = TRUE;

  if (_allowRegionSelection)
  {
    // Clear temporary outline point arrays
    _pointVec.clear();
    _tmpImgRelatedPointArray = QPointArray();
    _tmpWidgetRelatedPointArray = QPointArray();
      
    update();

    getImgLikePos(event->pos(), _rubberStart);
    _rubberEnd = _rubberStart;
    _pointVec.push_back(_rubberStart);

    _rubberOn = FALSE;
    _regionSelected = FALSE;
    _mouseHasBeenMoved = FALSE;

    if (_shouldIncRegionID)
      incrementRegionID();

    _shouldIncRegionID = FALSE;

    if (_regionSelectedCBFunc != NULL)
      (*_regionSelectedCBFunc)(_regionSelectedUserData, _regionSelected);
  }
}

// When mouse button is released, outline is drawn regarding to type
void SmageImgSelectionWidget::mouseReleaseEvent(QMouseEvent *)
{
  if (_allowRegionSelection)
  {
    if (_selectionType == SELECT_RECT)
    {
      if (_rubberEnd.x()-_rubberStart.x() != 0 && _rubberEnd.y()-_rubberStart.y() != 0 &&
	  _regionSelected == FALSE)
      {
        _regionSelected = TRUE;

        _pointVec.push_back(QPoint(_rubberEnd.x(), _rubberStart.y()));
        _pointVec.push_back(_rubberEnd);
        _pointVec.push_back(QPoint(_rubberStart.x(), _rubberEnd.y()));
      }
    }
    else if (_selectionType == SELECT_POLY)
    { 
      if ((_rubberEnd != _rubberStart) || _mouseHasBeenMoved)
      {
        // Close the outline
        _painter->begin(this);
        
        _painter->setRasterOp(NotROP);
        _painter->setPen(*_pen);
        _painter->moveTo(_rubberEnd);
        _painter->lineTo(_rubberStart);
        
        _painter->end();
    
        _regionSelected = TRUE;
      }
    }

    if (_regionSelectedCBFunc != NULL)
      (*_regionSelectedCBFunc)(_regionSelectedUserData, _regionSelected);


    // "_tmpWidgetRelatedPointArray" holds the outline point array so that
    // outline can be redrawn due to refreshing etc.
    _tmpWidgetRelatedPointArray = QPointArray(_pointVec.size()+1);
    _tmpImgRelatedPointArray = QPointArray(_pointVec.size()+1);
    for (unsigned int i=0; i<_pointVec.size(); i++)
    {
      _tmpImgRelatedPointArray.setPoint(i, _pointVec[i] - _imgStartPos);
      _tmpWidgetRelatedPointArray.setPoint(i, _pointVec[i]);
    }

    if (_pointVec.size() > 0)
      {
	_tmpWidgetRelatedPointArray.setPoint(_pointVec.size(), _pointVec[0]);
	_tmpImgRelatedPointArray.setPoint(_pointVec.size(), _pointVec[0] - _imgStartPos);
      }
  }
}

// Mouse button must be pressed ... outline is drawn when mouse is moved
void SmageImgSelectionWidget::mouseMoveEvent(QMouseEvent *event)
{
  if (_allowRegionSelection && _firstMousePress)
  {
    _mouseHasBeenMoved = TRUE;

    if (_selectionType == SELECT_RECT)
      drawRect(event);
    else if (_selectionType == SELECT_POLY)
      drawPoly(event);
  }
}

// Draw rectangle
void SmageImgSelectionWidget::drawRect(const QMouseEvent *event)
{
  // Drawing a rubber band

  _painter->begin(this);

  _painter->setRasterOp(NotROP);
  _painter->setPen(*_pen);
  _pen->setStyle(Qt::DotLine);

  if (_rubberOn)
    _painter->drawRect(_rubberStart.x(), _rubberStart.y(),
                       _rubberEnd.x()-_rubberStart.x(), _rubberEnd.y()-_rubberStart.y());

  getImgLikePos(event->pos(), _rubberEnd);
  _painter->drawRect(_rubberStart.x(), _rubberStart.y(),
         _rubberEnd.x()-_rubberStart.x(), _rubberEnd.y()-_rubberStart.y());

  _painter->end();

  _rubberOn = TRUE;
}

// Draw polyline
void SmageImgSelectionWidget::drawPoly(const QMouseEvent *event)
{
  _painter->begin(this);

  _painter->setPen(*_pen);
  _painter->moveTo(_rubberEnd);
  _pen->setStyle(Qt::SolidLine);

  getImgLikePos(event->pos(), _rubberEnd);
  _painter->lineTo(_rubberEnd);

  _painter->end();

  _pointVec.push_back(_rubberEnd);
  _rubberOn = TRUE;
}

// Before closing the widget ...
//  - ask if coded image should be generated, if not yet done
void SmageImgSelectionWidget::closeEvent(QCloseEvent *event)
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgSelectionWidget::closeEvent()\n");
#endif

  inherited::closeEvent(event);
}

// Reset the image widgets data (in particular, regions that have been 
//  selected are deleted)
void SmageImgSelectionWidget::resetImgWidget()
{
#ifdef THE_S_PRINT_DEBUG
  _TheSmageLogFile.lprintf("SmageImgSelectionWidget::resetImgWidget()\n");
#endif

  _pointVec.clear();
  
  _tmpImgRelatedPointArray = QPointArray();
  _tmpWidgetRelatedPointArray = QPointArray();

  _regionSelected = FALSE;
  _sceneChanged = TRUE;
  repaint(FALSE);
}
