// project header
#include "SmageTextEditor.h"
#include "../smageSharedFiles/SmageViewerWidget.h"

// qt header
#include <qapplication.h>
#include <qbuttongroup.h>
#include <qcolordialog.h>
#include <qcombobox.h>
#include <qfontdatabase.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qtextedit.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qvbox.h>


#include "./pixmaps/textbold.xpm"
#include "./pixmaps/textitalic.xpm"
#include "./pixmaps/textleft.xpm"
#include "./pixmaps/textcenter.xpm"
#include "./pixmaps/textright.xpm"
#include "./pixmaps/transparentBackground.xpm"


// c/c++ header
#include <assert.h>
#include <iostream>

using namespace std;

SmageTextEditor::SmageTextEditor(QWidget *parent, const char* name, 
				 const QString &text, const QFont &font)
  : QMainWindow(parent, name, WStyle_StaysOnTop)
{    
  _prevUserData = NULL;
  _prevCBFunc = NULL;
  _okUserData = NULL;
  _okCBFunc = NULL;
  _cancelUserData = NULL;
  _cancelCBFunc = NULL;

  _textEditor = NULL;

  _lineWidth = 1;
  _backgroundCol = QColor(255,255,255);
  _outlineCol = QColor(0,0,0);
  _backgroundAlpha = 111;

  _vBoxWidget = new QVBox(this);
  setCentralWidget(_vBoxWidget);

  initTextEditor(text, font);
  initTextToolBar();
  initBackgroundToolBar();

  initButtonArea();

  if (parent != 0)
    setGeometry(parent->pos().x() + 20, parent->pos().y() + 20, 550, 400);
  else
    setGeometry(300, 200, 550, 400);
}


SmageTextEditor::~SmageTextEditor()
{
}

void SmageTextEditor::setBackgroundColor(const QColor &color)
{
  if (_textEditor != NULL)
    {
      _textEditor->setPaper(QBrush(color));

      QPixmap pix(16, 16);
      pix.fill(color);
      _backgroundColorButton->setIconSet(pix);
    }
}

void SmageTextEditor::initTextEditor(const QString &text, const QFont &font)
{
  _textEditor = new QTextEdit(_vBoxWidget);
  _textEditor->setPaper(QBrush(_backgroundCol));
  _textEditor->setText(text);

  _textEditor->selectAll();
  _textEditor->setCurrentFont(font);
  _textEditor->setAlignment(Qt::AlignCenter);
  _textEditor->selectAll(FALSE);

  connect(_textEditor, SIGNAL(currentFontChanged(const QFont &)), this, SLOT(fontChanged(const QFont &)));
  connect(_textEditor, SIGNAL(currentAlignmentChanged(int)), this, SLOT(alignmentChanged(int)));
  connect(_textEditor, SIGNAL(currentColorChanged(const QColor &)), 
	  this, SLOT(colorChanged(const QColor &)));
}


void SmageTextEditor::initTextToolBar()
{
  QToolBar *toolBar = new QToolBar(this);
  toolBar->setLabel(tr("Text properties"));

  _comboFont = new QComboBox(TRUE, toolBar);
  QFontDatabase fontDataBase;
  _comboFont->insertStringList(fontDataBase.families());
  connect(_comboFont, SIGNAL(activated(const QString &)), this, SLOT(textFamily(const QString &)));
  _comboFont->lineEdit()->setText(QApplication::font().family());

  _comboSize = new QComboBox(TRUE, toolBar);
  QValueList<int> sizes = fontDataBase.standardSizes();
  QValueList<int>::Iterator it = sizes.begin();
  for (; it != sizes.end(); ++it)
    _comboSize->insertItem(QString::number(*it));
  connect(_comboSize, SIGNAL(activated(const QString &)), this, SLOT(textSize(const QString &)));
  _comboSize->lineEdit()->setText(QString::number(QApplication::font().pointSize()));

  _textBoldButton = new QToolButton(QPixmap(bold_xpm) , tr("Bold"),
				    QString::null, this, SLOT(textBold()), toolBar, "textBold"); 
  _textBoldButton->setToggleButton(TRUE);
  _textItalicButton = new QToolButton(QPixmap(italic_xpm), tr("Italic"),
				      QString::null, this, SLOT(textItalic()), toolBar, "textItalic");
  _textItalicButton->setToggleButton(TRUE);
  /*
  _textUnderlineButton = new QToolButton(QPixmap(under_xpm), tr("Underlined"),
					 QString::null, this, SLOT(textUnderline()), toolBar, "textUnderline");
  _textUnderlineButton->setToggleButton(TRUE);
  */

  QButtonGroup *toggleButtonGroup = new QButtonGroup();    
  toggleButtonGroup->setExclusive(TRUE);

  _alignLeftButton = new QToolButton(QPixmap(left_xpm), tr("Left align text"),
				     QString::null, this, SLOT(textLeft()), toolBar, "textLeft"); 
  _alignLeftButton->setToggleButton(TRUE);
  _alignLeftButton->setOn(FALSE);
  toggleButtonGroup->insert(_alignLeftButton);
  _alignCenterButton = new QToolButton(QPixmap(center_xpm), tr("Center align text"),
				      QString::null, this, SLOT(textCenter()), toolBar, "textCenter");
  _alignCenterButton->setToggleButton(TRUE);
  _alignCenterButton->setOn(TRUE);
  toggleButtonGroup->insert(_alignCenterButton);
  _alignRightButton = new QToolButton(QPixmap(right_xpm), tr("Right align text"),
				      QString::null, this, SLOT(textRight()), toolBar, "textRight");
  _alignRightButton->setToggleButton(TRUE);
  toggleButtonGroup->insert(_alignRightButton);

  QPixmap pix(16, 16);
  pix.fill(black);
  _textColorButton = new QToolButton(pix, tr("Change text color"), 
				     QString::null, this, SLOT(textColor()), toolBar, "textColor");
  QPixmap pix1(16, 16);
  pix1.fill(_backgroundCol);
  _backgroundColorButton = new QToolButton(pix1, tr("Change background color"), QString::null, this, 
					   SLOT(backgroundColor()), toolBar, "backgroundColor");
					   		   
  _backgroundTransparentButton = new QToolButton(QPixmap(transparentBackground_xpm), tr("Transparent background"), QString::null, this, 
						 SLOT(transparentBackground()), toolBar, "transparentBackground");
  _backgroundTransparentButton->setToggleButton( TRUE ); 
}

void SmageTextEditor::initBackgroundToolBar()
{
  QToolBar *toolBar = new QToolBar(this);
  toolBar->setLabel(tr("Background properties"));

  _comboWidth = new QComboBox(TRUE,toolBar);
  _comboWidth->insertItem(QString::number(0));
  _comboWidth->insertItem(tr("Contour width: 1"));
  for (uint i=2; i<12; i++)
    _comboWidth->insertItem(QString::number(i));
  connect(_comboWidth, SIGNAL(activated(const QString &)), this, SLOT(outlineWidth(const QString &)));
  _comboWidth->setCurrentItem( 1 );
  _comboWidth->setEditable(FALSE);
  
  QPixmap pix2(16, 16);
  pix2.fill(_outlineCol);
  _outlineColorButton = new QToolButton(pix2, tr("Change  contour color"), QString::null, this, 
					SLOT(outlineColor()), toolBar, "outlineColor");
}


void SmageTextEditor::initButtonArea()
{  
  QHBox *hBox = new QHBox(_vBoxWidget);
  hBox->setSpacing(10);
  QPushButton *button;
  QWidget *spacer;
  spacer = new QWidget(hBox);
  button = new QPushButton(tr("Preview"), hBox);
  connect(button, SIGNAL(clicked()), this, SLOT(preview()));
  button = new QPushButton(QObject::tr("OK"), hBox);
  connect(button, SIGNAL(clicked()), this, SLOT(ok()));
  button = new QPushButton(QObject::tr("Cancel"), hBox);
  connect(button, SIGNAL(clicked()), this, SLOT(cancel()));
  spacer = new QWidget(hBox);
}


void SmageTextEditor::textFamily(const QString &font)
{
  _textEditor->selectAll();
  _textEditor->setFamily(font);
  _textEditor->selectAll(FALSE);
  _textEditor->viewport()->setFocus();
}


void SmageTextEditor::textSize( const QString &size)
{
  _textEditor->selectAll();
  _textEditor->setPointSize(size.toInt());
  _textEditor->selectAll(FALSE);
  _textEditor->viewport()->setFocus();
}


void SmageTextEditor::textBold()
{
  _textEditor->selectAll();
  _textEditor->setBold(_textBoldButton->isOn());
  _textEditor->selectAll(FALSE);
}

/*
void SmageTextEditor::textUnderline()
{
  _textEditor->selectAll();
  _textEditor->setUnderline(_textUnderlineButton->isOn());
  _textEditor->selectAll(FALSE);
}
*/

void SmageTextEditor::textItalic()
{
  _textEditor->selectAll();
  _textEditor->setItalic(_textItalicButton->isOn());
  _textEditor->selectAll(FALSE);
}

void SmageTextEditor::textLeft()
{
  _textEditor->selectAll();
  _textEditor->setAlignment(AlignLeft);
  _textEditor->selectAll(FALSE);
}

void SmageTextEditor::textRight()
{
  _textEditor->selectAll();
  _textEditor->setAlignment(AlignRight);
  _textEditor->selectAll(FALSE);
}

void SmageTextEditor::textCenter()
{  
  _textEditor->selectAll();
  _textEditor->setAlignment(AlignHCenter);
  _textEditor->selectAll(FALSE);
}

void SmageTextEditor::textColor()
{
  QColor col = QColorDialog::getColor(_textEditor->color(), this);
  if (!col.isValid())
    return;
  _textEditor->selectAll();
  _textEditor->setColor(col);
  _textEditor->selectAll(FALSE);

  QPixmap pix(16, 16);
  pix.fill(col);
  _textColorButton->setIconSet(pix);
}

void SmageTextEditor::fontChanged(const QFont &font)
{
  _comboFont->lineEdit()->setText(font.family());
  _comboSize->lineEdit()->setText(QString::number(font.pointSize()));
  
  _textBoldButton->setOn(font.bold());
  _textItalicButton->setOn(font.italic());
  //_textUnderlineButton->setOn(font.underline());
}

void SmageTextEditor::alignmentChanged(int alignment)
{    
  if ((alignment == AlignAuto) || (alignment & AlignLeft))
    _alignLeftButton->setOn(TRUE);
  else if ((alignment & AlignHCenter))
    _alignCenterButton->setOn(TRUE);
  else if ((alignment & AlignRight))
    _alignRightButton->setOn(TRUE);
}

void SmageTextEditor::colorChanged(const QColor &color)
{
  QPixmap pix(16, 16);
  pix.fill(color);
  _textColorButton->setIconSet(pix);
}


void SmageTextEditor::backgroundColor()
{
  QColor col = QColorDialog::getColor(_backgroundCol, this);
  if (!col.isValid())
    return;

  _backgroundCol = col;
  setBackgroundColor(col);

  QPixmap pix(16, 16);
  pix.fill(col);
  _backgroundColorButton->setIconSet(pix);
}

void SmageTextEditor::transparentBackground()
{
  if ( _backgroundAlpha == 0 ) _backgroundAlpha = 255;
  else _backgroundAlpha = 0;
}

void SmageTextEditor::outlineColor()
{
  QColor col = QColorDialog::getColor(_outlineCol, this);
  if (!col.isValid())
    return;

  _outlineCol = col;  

  QPixmap pix(16, 16);
  pix.fill(col);
  _outlineColorButton->setIconSet(pix);
}


void SmageTextEditor::outlineWidth( const QString &width)
{
  if (width == tr("Contour width: 1"))
    _lineWidth = 1;

  else
    _lineWidth = width.toInt();
}


void SmageTextEditor::preview()
{
  if (_prevUserData != NULL && _prevCBFunc != NULL)
    {
      TheS_BubbleData data;
      data.textData.bubbleText = _textEditor->text();
      data.textData.bubbleTextFont = _textEditor->currentFont();
      data.textData.bubbleTextCol = _textEditor->color();
      data.textData.bubbleTextFlags |= _textEditor->alignment();
      data.shapeData.bubbleBackCol = _textEditor->paper().color();
      data.shapeData.bubbleOutlineCol = _outlineCol;
      data.shapeData.bubbleOutlineWidth = _lineWidth;
      data.shapeData.bubbleBackgroundAlpha = _backgroundAlpha;

      (*_prevCBFunc)(_prevUserData, data);
    }
}

void SmageTextEditor::ok()
{
  if (_okUserData != NULL && _okCBFunc != NULL)
    {
      TheS_BubbleData data;
      data.textData.bubbleText = _textEditor->text();
      data.textData.bubbleTextFont = _textEditor->currentFont();
      data.textData.bubbleTextCol = _textEditor->color();
      data.textData.bubbleTextFlags |= _textEditor->alignment();
      data.shapeData.bubbleBackCol = _textEditor->paper().color();
      data.shapeData.bubbleOutlineCol = _outlineCol;
      data.shapeData.bubbleOutlineWidth = _lineWidth;
      data.shapeData.bubbleBackgroundAlpha = _backgroundAlpha;

      (*_okCBFunc)(_okUserData, data);
    }

  hide();
}

void SmageTextEditor::assignNewProps(const TheS_BubbleData &props)
{
  _textEditor->setText(props.textData.bubbleText);

  _textEditor->selectAll();
  _textEditor->setCurrentFont(props.textData.bubbleTextFont);
  _textEditor->setColor(props.textData.bubbleTextCol);
  _textEditor->setAlignment(props.textData.bubbleTextFlags);
  _textEditor->selectAll(FALSE);

  _lineWidth = props.shapeData.bubbleOutlineWidth;
  _outlineCol = props.shapeData.bubbleOutlineCol;
  _backgroundCol = props.shapeData.bubbleBackCol;
  _backgroundAlpha = props.shapeData.bubbleBackgroundAlpha;

  setBackgroundColor(_backgroundCol);
  colorChanged(props.textData.bubbleTextCol);

  QPixmap pix(16, 16);
  pix.fill(_outlineCol);
  _outlineColorButton->setIconSet(pix);

  _comboWidth->setCurrentItem( _lineWidth );
  
  if ( _backgroundAlpha == 0 ) _backgroundTransparentButton->setOn( TRUE);
  else _backgroundTransparentButton->setOn( FALSE);
}

void SmageTextEditor::cancel()
{
  if (_cancelUserData != NULL && _cancelCBFunc != NULL)
    (*_cancelCBFunc)(_cancelUserData);

  hide();
}
